# Steps that relates to media handling.

require "io/wait"
require "open3"

require_relative "../../lib/media/create-update-cd"
require_relative "../../lib/media/mount"

# Ensure there won’t be dangling temporary mount things created by below step.
at_exit do
  if ENV["MOUNT_DEBUG"]
    puts "Skipping exit cleanup due env var MOUNT_DEBUG"
    exit 1
  end

  @testmounts_path ||= "/tmp/testmounts"
  next unless Dir.exist? @testmounts_path

  mounts = File.readlines "/proc/mounts"
  mounts.keep_if do |line|
     line.include? "#{@testmounts_path}/"
  end

  mounts.size.times do
    mounts.delete_if do |line|
      splitted = line.split " "
      path = splitted[1]
      `umount -R #{path}`
      Dir.rmdir path if $?.success?
      $?.success?
    end
  end

  unless mounts.empty?
    raise "Failed to unmount mounts under #{@testmounts_path}. Following mounts still exists: #{mounts}"
  end

  FileUtils.rm_rf @testmounts_path
end

After("@mounts") do
  if ENV["MOUNT_DEBUG"]
    log "Skipping unmounts due env var MOUNT_DEBUG"
    next
  end

  next if @mount.nil?

  `umount #{@squashfs_path}` if system("mountpoint -q #{@squashfs_path}")
  @mount.unmount fail: true
  sleep 0.1

  Dir.rmdir @mount.fs_mount_point
  Dir.rmdir @mount.squash_mount_point
  Dir.rmdir @mount.mirror_mount_point
  Dir.rmdir @mount.mount_point
  Dir.rmdir "#{@testmounts_path}/media" # TODO: I’m not exactly sure why this directory gets created.
  Dir.rmdir "#{@testmounts_path}/mnt"
  Dir.rmdir "#{@testmounts_path}/temp" if Dir.exist? "#{@testmounts_path}/temp"
  FileUtils.rm_r "#{@testmounts_path}/upperdir" if Dir.exist? "#{@testmounts_path}/upperdir"
  FileUtils.rm_r "#{@testmounts_path}/workdir" if Dir.exist? "#{@testmounts_path}/workdir"
  Dir.rmdir @testmounts_path
end

def generate_fai_cd image_name, distribution, medium_type

  case @distribution_name
  when "Debian 10"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dir"
  when "OpenSUSE 15.2"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dirs/opensuse"
  else
    raise "Not implemented distro: #{@distribution_name}"
  end

  ARGV.clear
  ARGV << "--no-logging"
  #ARGV << "--verbose"
  #ARGV << "--verbose" << "--debug"
  ARGV << "medium"
  ARGV << "--fai-config-dir" << fai_config_dir
  ARGV << "--squash-compression" << "lz4" # For speedup.
  ARGV << "--target-distribution" << distribution
  ARGV << "--filename" << "/tmp/#{image_name}"
  ARGV << "--mirror-dir" << @fai_mirror_path
  ARGV << "--target-site" << "testsite"
  ARGV << "--target-host" << "opensuse-testhost"
  (ARGV << "--install") if medium_type == :install
  (ARGV << "--update") if medium_type == :update
  ARGV << "--force"
  ARGV.flatten!

  log "Command: “ft #{ARGV.join(" ")}”"

  ft = FT.new
  ft.options
  ft.perform

  @medium_object = ft.command_object
end

def chroot_exec command, timeout:
  #log "chroot executing #{command}"

  unless @chroot_wait_thr.alive?
    expect(@chroot_wait_thr.alive?).to be_truthy, "chroot process is not running any more, status: #{@chroot_wait_thr.status}, output: #{@chroot_out.read.inspect}"
  end

  @chroot_in.puts command
  @chroot_in.puts %{echo "result: $?"}
  @chroot_in.wait_writable(timeout)
  @chroot_out.wait_readable(timeout)

  result = "invalid"
  loop do
    output = @chroot_out.readpartial 10000
    if output.start_with? "result: "
      result = output.strip
      break
    end

    #log output
  end

  result
end

After do
  if @host_config_tempdir
    FileUtils.remove_entry_secure @host_config_tempdir
  end
end

Given('I have host configuration with following data:') do |table|
  # table is a Cucumber::MultilineArgument::DataTable
  data = table.rows_hash

  @host_config_tempdir = Dir.mktmpdir("fai-tests-media_creation-")

  Dir.chdir @host_config_tempdir do
    FileUtils.mkdir "sites"
    FileUtils.mkdir "sites/#{data["target_site"]}"

    content = {
      data["target_host"] => {
        "image_name" => data["image_name"],
        "distribution" => data["distribution"],
        "distribution_version" => data["dist_version"]
      }
    }

    yaml = YAML.dump content
    File.open "sites/#{data["target_site"]}/targets.yaml", "w" do |f|
      f.puts yaml
    end
  end
end

Given('I have {fai_medium_type} medium with following data:') do |fai_medium_type, table|
  # table is a Cucumber::MultilineArgument::DataTable
  data = table.rows_hash
  image_name = data["image_name"]
  distribution = data["distribution"]

  unless File.exist? "/tmp/#{image_name}"
    generate_fai_cd image_name, distribution, fai_medium_type
  end
end

When('I create {fai_medium_type} medium with following data:') do |medium_type, table|
  # table is a Cucumber::MultilineArgument::DataTable¨
  data = table.rows_hash
  image_name = data["image_name"]
  distribution = data["distribution"]
  distribution_version = data["dist_version"]

  generate_fai_cd image_name, distribution, medium_type
end

When('I create {fai_medium_type} medium for site {string} and host {string}') do |fai_medium_type, site, host|
  # TODO: I probably should either modify generate_fai_cd to use opts array (so I can pass host config to there) or
  # just kind of copy that here and modify to the needs.
  # TODO: But the bigger thing is that IIRC there is no way to pass host configurations dir as a cli arg. That needs to be implemented for this.

  pending # Write code here that turns the phrase above into concrete actions
end

When('I mount squashfs filesystem inside medium {string}') do |medium|
  @testmounts_path = "/tmp/testmounts"

  # First ensure there is no dangling files.
  expect(File.directory? @testmounts_path).to be_falsey, "There is already temp mount directory. Previous test probably has " +
                                                          "failed and at_exit hook did not work. Please clean up the directory manually"

  @mount = Media::Mount.new @testmounts_path
  #mount.send(:set_opts, { verbose: true })
  #mount.send(:set_opts, { verbose: true, debug: true })
  @mount.mount_cd_file "/tmp/#{medium}"

  @squashfs_path = "#{@testmounts_path}/temp"
  Dir.mkdir @squashfs_path
  Dir.mkdir "#{@testmounts_path}/upperdir"
  Dir.mkdir "#{@testmounts_path}/workdir"
  `mount -t overlay -o lowerdir="#{@mount.fs_mount_point}",upperdir=#{@testmounts_path}/upperdir,workdir=#{@testmounts_path}/workdir overlay "#{@squashfs_path}"`
  expect($?.success?).to be_truthy, "Failed to mount squashfs"
end

When('I chroot to mounted filesystem') do
  @chroot_in, @chroot_out, @chroot_wait_thr = Open3.popen2e("chroot #{@squashfs_path} bash")

  @chroot_in.wait_writable(4)

  # Getting output handling working better.
  unless @chroot_wait_thr.alive?
    expect(@chroot_wait_thr.alive?).to be_truthy, "chroot failed. status: #{@chroot_wait_thr.status}, output: #{@chroot_out.read}"
  end

  at_exit do
    @chroot_in&.close
    @chroot_dout&.close
  end
end

When('I use only fai_cd.repo') do
  Dir.glob("#{@squashfs_path}/etc/zypp/repos.d/*.repo").each do |file|
    if File.basename(file) != "fai_cd.repo"
      File.delete file
    end
  end
end

When('I refresh package lists') do
  result = chroot_exec "zypper --non-interactive refresh --force", timeout: 4
  expect(result).to be == "result: 0", "zypper refresh failed, result: #{result.inspect}"
end

When('I install package {string}') do |package|
  result = chroot_exec "zypper --non-interactive install #{package}", timeout: 8
  expect(result).to be == "result: 0", "zypper install #{package} failed, result: #{result.inspect}"
end

Then('I should have medium {string}') do |filename|
  expect(File).to exist("/tmp/#{filename}")
end

Then('medium {string} should contain file {string}') do |medium, file|
  expect(@command_object.send :command_available?, "lsar").to be_truthy, "unar needs to be installed in your system to run this test"

  content = @command_object.send :execvalue, "lsar /tmp/#{medium}"
  lines = content.split "\n"

  expect(lines).to include(file)
end

Then('medium {string} should contain directory {string}') do |medium, directory|
  log "This test fails because the target dir is inside squashfs, so that needs to be mounted here."
  expect(true).to be_falsey

  expect(@command_object.send :command_available?, "lsar").to be_truthy, "unar needs to be installed in your system to run this test"

  content = @command_object.send :execvalue, "lsar /tmp/#{medium}"
  lines = content.split "\n"

  expect(lines).to include("#{file}/")
end

Then('I should close chroot session') do
  @chroot_in.puts "exit"
  @chroot_in.close
  @chroot_out.close

  # Wait for bash to close itself.
  @chroot_wait_thr.join
  sleep 0.1

  if ENV["MOUNT_DEBUG"]
    log "Skipping unmounts due env var MOUNT_DEBUG"
  else
    `umount #{@squashfs_path}`
  end
end

Then('mounted squashfs filesystem should contain directory {string}') do |directory|
  expect(File.directory? "#{@mount.fs_mount_point}/#{directory}").to be_truthy
end

Then('I should have package {string} installed') do |package|
  result = chroot_exec %{rpm --quiet -q "#{package}"}, timeout: 4
  expect(result).to be == "result: 0", "#{package} is not installed in the chroot, #{result.inspect}"
end
