# nfsroot specific steps.

require_relative "../../lib/ft"

def generate_nfsroot fai_config_dir, features
  ARGV.clear
  ARGV << "--no-logging"
  #ARGV << "--verbose"
  #ARGV << "--verbose" << "--debug"
  ARGV << "nfsroot"
  ARGV << "--fai-config-dir" << fai_config_dir
  (ARGV << "--no-dracut") unless features["dracut"]
  (ARGV << "--base-archive" << features["base_archive"]) if features["base_archive"]
  ARGV << "--base-compress-algo" << "lz4"
  ARGV << "--force"
  ARGV.flatten!

  log "Command: “ft #{ARGV.join(" ")}”"

  ft = FT.new
  ft.no_root_check = true
  ft.options
  ft.perform

  @command_object = ft.command_object

  # Reset current working dir to make directory resolution work properly.
  Dir.chdir "#{__dir__}/../.."
end

def nfsroot_feature_table table
  hash = table.rows_hash
  hash.map do |key, value|
    value = true if value == "true"
    value = false if value == "false"
    [ key, value ]
  end.to_h
end

Given('I have existing nfsroot with following attributes:') do |table|
  features = nfsroot_feature_table(table)

  case @distribution_name
  when "Debian 10"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dir"
  when "OpenSUSE 15.2"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dirs/opensuse"
  else
    raise "Not implemented distro: #{@distribution_name}"
  end

  # Construct nfsroot object.

  ARGV.clear
  ARGV << "--no-logging"
  #ARGV << "--verbose"
  #ARGV << "--verbose" << "--debug"
  ARGV << "nfsroot"
  ARGV << "--fai-config-dir" << fai_config_dir
  ARGV << "--no-dracut" # For speedup.
  ARGV << "--base-compress-algo" << "lz4"
  ARGV << "--force"
  ARGV.flatten!

  #puts "Command: “ft #{ARGV.join(" ")}”"

  # Construct nfsroot object.

  ft = FT.new
  ft.no_root_check = true
  ft.options

  @command_object = ft.command_object

  config = @command_object.instance_variable_get :@config
  expect(config).to be

  initrd_version = Proc.new do
    path = Pathname.new "#{config.nfsroot}/boot/vmlinuz"
    if path.exist?
      path = path.realpath
      vmlinuz = path.basename.to_s
      splitted = vmlinuz.split("-", 2)
      splitted[1]
    else
      nil
    end
  end

  # Check if we have correct nfsroot.

  distro_found = false
  version_found = false
  already_generated = false
  loop do
    if File.exist? "#{config.nfsroot}/etc/os-release"
      File.foreach("#{config.nfsroot}/etc/os-release") do |line|
        line.strip!
        distro_found = true if line == %{ID="#{@distribution_id}"} or line == %{ID=#{@distribution_id}}
        version_found = true if line == %{VERSION_ID="#{@distribution_version}"}
      end
    end

    if not distro_found
      raise "Already generated nfsroot, and distro test is failing..." if already_generated
      log "Existing nfsroot’s distribution does not match requested nfsroot. Creating a new one."
      generate_nfsroot fai_config_dir, features
      already_generated = true
      next
    end

    if not version_found
      raise "Already generated nfsroot, and distro version test is failing..." if already_generated
      log "Existing nfsroot’s distribution version does not match requested nfsroot. Creating a new one."
      generate_nfsroot fai_config_dir, features
      already_generated = true
      next
    end

    if features["dracut"]
      kernel_version = initrd_version.call
      unless File.file? "#{config.nfsroot}/boot/initrd-#{kernel_version}"
      #if Dir.glob("#{config.nfsroot}/boot/initrd*").size <= 0
        raise "Already generated nfsroot, and dracut version test is failing..." if already_generated
        log "Existing nfsroot does not have dracut. Creating a new one."
        generate_nfsroot fai_config_dir, features
        already_generated = true
        next
      end
    end

    break
  end

  if features["dracut"]
    kernel_version = initrd_version.call
    expect(File.file? "#{config.nfsroot}/boot/initrd-#{kernel_version}").to be_truthy,
        "initrd is not present, probably dracut is not installed."
    #expect(Dir.glob("#{config.nfsroot}/boot/initrd*").size).to be > 0
  end
end

When('I generate nfsroot') do |table|
  features = nfsroot_feature_table(table)

  # Ensure we are at top of the project, even if something changes current working dir earlier.
  Dir.chdir "#{__dir__}/../.."

  case @distribution_name
  when "Debian 10"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dir"
  when "OpenSUSE 15.2"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dirs/opensuse"
  else
    raise "Not implemented distro: #{@distribution_name}"
  end

  generate_nfsroot fai_config_dir, features
end

When('I generate nfsroot from base archive') do |table|
  expect(File.file?($BASE_ARCHIVE_PATH)).to be_truthy

  features = nfsroot_feature_table(table)
  features["base_archive"] = $BASE_ARCHIVE_PATH

  # Ensure we are at top of the project, even if something changes current working dir earlier.
  Dir.chdir "#{__dir__}/../.."

  case @distribution_name
  when "Debian 10"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dir"
  when "OpenSUSE 15.2"
    fai_config_dir = "#{__dir__}/../../test/fai_config_dirs/opensuse"
  else
    raise "Not implemented distro: #{@distribution_name}"
  end

  generate_nfsroot fai_config_dir, features
end

Then('I should have nfsroot with file {string}') do |path|
  config = @command_object.instance_variable_get :@config
  expect(config).to be

  expect(config.nfsroot).to_not be_empty
  expect(File).to exist(config.nfsroot)
  expect(File).to exist("#{config.nfsroot}/#{path}")
end

Then('I save generated base archive') do
  config = @command_object.instance_variable_get :@config
  expect(config).to be

  expect(config.nfsroot).to_not be_empty

  expect(@command_object.send :command_available?, "lz4").to be_truthy, "lz4 needs to be installed in your system to run this test"
  @command_object.send :execute, "lz4 -q -f #{config.nfsroot}/var/tmp/base.tar #{$BASE_ARCHIVE_PATH}"
  expect(File.file?($BASE_ARCHIVE_PATH)).to be_truthy
end
