
require "tempfile"

require_relative "base"
require_relative "../common"

module FaiToolkit::Base::CreateBaseMethods
  include Common

  def setarch arch
    arch == "i386" ? "linux32" : ""
  end

  def check
    execute "mknod #{@tmpdir}/test-dev-null c 1 3"
    unless exec_success?
      puts "Cannot create device files on #{@tmpdir}, aborting."
      puts "Perhaps this directory is mounted with option nodev."
      FileUtils.rm_rf @tmpdir
      exit 1
    end

    testfile = "#{@tmpdir}/test-dev-null"
    File.open testfile, "w" do |file|
      file.write "test"
    end

    unless File.exist? testfile
      puts "Cannot create device files on #{@tmpdir}, aborting."
      puts "Perhaps this directory is mounted with option nodev."
      FileUtils.rm_rf @tmpdir
      exit 1
    end

    FileUtils.rm_f testfile
  end

  def mkpost_centos
    raise "MIRROR_CENTOS missing." if @create_base_image_config["MIRROR_CENTOS"].empty?

    postfile = "#{@tmpdir}/post"
    File.open postfile, "w" do |file|
      file.write <<~EOF
      #!/bin/sh
      mkdir -p #{@tmpdir}/etc/yum.repos.d/orig
      cp -p #{@tmpdir}/etc/yum.repos.d/*.repo #{@tmpdir}/etc/yum.repos.d/orig
      perl -pi -e 's,mirrorlist=,#mirrorlist=,; s,#baseurl=http://mirror.centos.org,baseurl=#{@create_base_image_config["MIRROR_CENTOS"]},;' #{@tmpdir}/etc/yum.repos.d/CentOS-Base.repo
      EOF
    end

    FileUtils.chmod 555, postfile
  end

  def mkpost_slc version
    raise "MIRROR_SLC missing" if @create_base_image_config["MIRROR_SLC"].empty?

    postfile = "#{@tmpdir}/post"
    File.open postfile, "w" do |file|
      file.write <<~EOF
      #! /bin/sh
      mkdir -p #{@tmpdir}/etc/yum.repos.d/orig
      cp -p #{@tmpdir}/etc/yum.repos.d/*.repo #{@tmpdir}/etc/yum.repos.d/orig
      perl -pi -e 's,baseurl=http://linuxsoft.cern.ch,baseurl=#{@create_base_image_config["MIRROR_SLC"]},;' #{@tmpdir}/etc/yum.repos.d/slc#{version}-os.repo
      perl -pi -e 's,baseurl=http://linuxsoft.cern.ch,baseurl=#{@create_base_image_config["MIRROR_SLC"]},;' #{@tmpdir}/etc/yum.repos.d/slc#{version}-updates.repo
      EOF
    end

    FileUtils.chmod 555, postfile
  end

  def cleanup_deb
    execute "chroot #{@tmpdir} apt-get clean"
    FileUtils.rm_f [ "#{@tmpdir}/etc/hostname", "#{@tmpdir}/etc/resolv.conf",
          "#{@tmpdir}/var/lib/apt/lists/*_*", "#{@tmpdir}/usr/bin/qemu-*-static",
          "#{@tmpdir}/etc/udev/rules.d/70-persistent-net.rules" ]
    File.truncate "#{@tmpdir}/etc/machine-id", 0
  end

  def cleanup_rinse

    # check if chroot works
    verb "Installed packages in chroot:"
    if @@_verbose
      puts execvalue "chroot #{@tmpdir} rpm -qa|sort"
      print "CHROOT rpm -qa: "
      puts execvalue "chroot #{@tmpdir} rpm -qa|wc -l"
    end

    FileUtils.rm_f [ "#{@tmpdir}/etc/resolv.conf", "#{@tmpdir}/post" ]
    if Dir.exist? "#{@tmpdir}/etc/yum.repos.d/orig"
        execute "mv #{@tmpdir}/etc/yum.repos.d/orig/* #{@tmpdir}/etc/yum.repos.d/"
        FileUtils.rm_rf "#{@tmpdir}/etc/yum.repos.d/orig"
    end
  end

  def tarit filename
    if @compress
      result_file = "#{@output_directory}/#{filename}.#{@ext}"
      execute "tar #{@tar_attributes} --numeric-owner --one-file-system -C #{@tmpdir} -cf - . | #{@zip} > #{result_file}"
    else
      execute "cp -a --one-file-system --target-directory #{@output_directory} #{@tmpdir}"
      dir_name = File.basename @tmpdir
      execute "mv #{@output_directory}/#{dir_name}/* #{@output_directory}"
      FileUtils.rmdir "#{@output_directory}/#{dir_name}"
    end
  end

  def opensuse distribution, arch, version
    check
    l32 = setarch arch
    rinse_command do |rinse|
      execute "#{l32} #{rinse} --directory #{@tmpdir} --distribution opensuse-#{version} --arch #{arch}"
    end
    cleanup_rinse
    tarit distribution
  end

  def centos distribution, arch, version
    domain = execvalue "domainname"

    check
    l32 = setarch arch
    mkpost_centos
    rinse_command do |rinse|
      execute "#{l32} rinse --directory #{@tmpdir} --distribution centos-#{version} --arch #{arch}"
    end
    execute "domainname domain" # workaround for #613377
    cleanup_rinse
    tarit distribution
  end

  def slc distribution, arch, version

    check
    l32 = setarch arch
    mkpost_slc version
    rinse_command do |rinse|
      execute "#{l32} rinse --directory #{@tmpdir} --distribution slc-#{version} --arch #{arch}"
    end
    cleanup_rinse
    tarit distribution
  end

  def debgeneric distribution, mirror, arch

    check

    unless arch.empty?
      execute "qemu-debootstrap --arch #{arch} #{exc} #{inc} #{distribution} #{@tmpdir} #{mirror}"
      # If this gets implemented, this target needs to be handled.
      target = "#{target}_#{arch}"
    else
      arch = "i386"
      arch = "amd64" if distribution.include? "64"
      execute "debootstrap --arch #{arch} #{exc} #{inc} #{distribution} #{@tmpdir} #{mirror}"
    end
    cleanup_deb
    tarit distribution
  end

  def list_distributions
    puts <<EOF
Following targets can be passed for --distribution:

    OPENSUSE_15_1_32  OPENSUSE_15_1_64
    OPENSUSE_15_2_32  OPENSUSE_15_2_64

    CENTOS5_32        CENTOS5_64
    CENTOS6_32        CENTOS6_64
    CENTOS7_32        CENTOS7_64
                      CENTOS8_64
    SLC5_32           SLC5_64
    SLC6_32           SLC6_64
                      SLC7_64
    TRUSTY32          TRUSTY64
    XENIAL32          XENIAL64
                      BIONIC64
    SQUEEZE32         SQUEEZE64
    WHEEZY32          WHEEZY64
    JESSIE32          JESSIE64
    STRETCH32         STRETCH64
    BUSTER32          BUSTER64
    SID32             SID64

    Also something like “--distribution opensuse --distribution-version 15.2” can be used, for convenience.
EOF
  end

  private

  def rinse_command
    packages_args = ""

    file = Tempfile.create "ft-create-base-rinse-packages-"
    if @added_packages and @added_packages.size > 0
      file.puts(@added_packages.join("\n"))
      file.close

      packages_args = "--add-pkg-list #{file.path}"
    end

    before_post = ""
    if File.exist? "#{@tmpdir}/post"
      before_post = "--before-post-install #{@tmpdir}/post"
    end

    command = "rinse --verbose"
    command = "#{command} #{packages_args} #{before_post}" unless packages_args.empty?

    yield command
  ensure
    file.close
    File.unlink file
  end
end