require "fileutils"
require "tmpdir"
require "psych"

require "optimist"

require_relative "base"
require_relative "../options"
require_relative "create-base-methods"

# Main class.
class FaiToolkit::Base::CreateBase
  include Options
  include FaiToolkit::Base::CreateBaseMethods

  def options parser
    common_options parser: parser do |p|
      #p.opt :fai_config_dir,    "FAI’s main config files.",
      #    type: :string,      default: "/etc/fai",          short: "c"
      p.opt :list_distributions,    "Show list of supported distributions",
                                                            short: "l"
      p.opt :foreign_arch,          "Build for given foreign architecture. Only used for Debian style bases.",
          type: :string,                                    short: "F"
      p.opt :distribution,          "Distribution to build the image from",
          type: :string,                                    short: "D"
      p.opt :distribution_version,  "Version of the distribution, if a slug was not used, as listed in --list-distributions",
          type: :string,                                    short: "V"
      p.opt :preserve_tmpdir,       "If given, preserves temporary files after script has finished.",
          type: :boolean,     default: false,               short: "p"
      p.opt :fai_config_space,      "Path to FAI’s config space.",
          type: :string,      default: "/srv/fai/config",   short: "s"
      p.opt :compress,              "Compress resulting base. Usually this is wanted.",
          type: :boolean,     default: true,                short: :none
      p.opt :output_directory,      "Directory where result will be saved to. Defaults to current working directory",
          type: :string
      p.opt :add_package,           "Install additional package to target installation.",
          type: :string,                                    short: "a", multi: true
    end
  end

  def check_options
    options = @options

    parse_common_options options

    #@config = FaiConfig.new @fai_config_dir
    #@config.load_fai_conf

    if options[:list_distributions]
      list_distributions
      exit 1
    end

    if options[:distribution].nil? or options[:distribution].size < 1
      error "--distribution must be given."
    end
    @distribution = options[:distribution]
    @distribution_version = options[:distribution_version]

    unless Dir.exist? options[:fai_config_space]
      error "FAI’s config space is used for base image configuration, but config space is missing. Please provide correct config space or create a new one."
    end

    @create_base_image_config_file = "#{options[:fai_config_space]}/basefiles/create_base_image.yaml"
    unless File.exist? @create_base_image_config_file
      error "FAI’s config space is used for base image configuration, but configuration file #{@create_base_image_config_file} is missing. Please add correct config file to config path. For now, you need to investigate this script how to do that."
    end
    lines = File.readlines(@create_base_image_config_file).join("\n")
    @create_base_image_config = Psych.safe_load lines

    @compress = options[:compress]

    @output_directory = options[:output_directory]
    if @output_directory.nil? and !@compress
      error "--output-directory must be specified when using --no-compress."
    end
    @output_directory = Dir.pwd if @output_directory.nil?

    FileUtils.mkdir @output_directory unless Dir.exist? @output_directory

    unless @compress
      if Dir["#{@output_directory}/*"].size > 0
        error "There is already content in #{@output_directory}. Please remove it manually if you intend to do this operation."
      end
    end

    @added_packages = options[:add_package]

    @foreign_arch = options[:foreign_arch]

    set_opts options
  end

  def perform
    startup_checks
    real_perform
  end

  private

  def real_perform
    # These could be used to use other compression algos.
    @ext = "tar.xz"
    @zip = "xz -8"
    @tar_attributes = "--xattrs --selinux --acls"

    @tmpdir = Dir.mktmpdir "create-base-image_"
    FileUtils.chmod 755, @tmpdir

    # also the name of the output file
    handle_by_distribution @distribution, @distribution_version

    # cleanup
    if not opts[:preserve_tmpdir] == true
      FileUtils.rm_rf @tmpdir
    end
  end

  def startup_checks
    # Nya.
  end

  def handle_by_distribution distribution, version

    if distribution == "opensuse"
      unless version
        raise "Basic distribution name “#{distribution}” was given, but --distribution-version was not used."
      end

      opensuse distribution, "amd64", version
      return
    end

    # NOTE: This list needs to be kept in sync with list_distributions()
    case distribution
    when "OPENSUSE_15_1_32"
      opensuse distribution, "i386", "15.1"
    when "OPENSUSE_15_1_64"
      opensuse distribution, "amd64", "15.1"
    when "OPENSUSE_15_2_64"
      opensuse distribution, "amd64", "15.2"
    when "CENTOS5_32"
      centos distribution, "i386", 5
    when "CENTOS5_64"
      centos distribution, "amd64", 5
    when "CENTOS6_32"
      centos distribution, "i386", 6
    when "CENTOS6_64"
      centos distribution, "amd64", 6
    when "CENTOS7_32"
      centos distribution, "i386", 7
    when "CENTOS7_64"
      centos distribution, "amd64", 7
    when "CENTOS8_64"
      centos distribution, "amd64", 8
    when "SLC5_32"
      slc distribution, "i386", 5
    when "SLC5_64"
      slc distribution, "amd64", 5
    when "SLC6_32"
      slc distribution, "i386", 6
    when "SLC6_64"
      slc distribution, "amd64", 6
    when "SLC7_64"
      slc distribution, "amd64", 7
    when /TRUSTY*|XENIAL*|BIONIC*/
      debgeneric distribution, @create_base_image_config["MIRROR_UBUNTU"]
    when /SQUEEZE*|WHEEZY*|JESSIE*|STRETCH*|BUSTER*|SID*/
      debgeneric distribution, @create_base_image_config["MIRROR_DEBIAN"], @foreign_arch
    else
      puts "Unknown distribution: #{distribution}. Aborting."
      exit 1
    end

  end
end