require_relative "base"

require_relative "../common"

module FaiToolkit::Base::NfsrootBaseMethods
  include Common

  # Creates a temporary base archive that is supplied with generated nfsroot.
  def create_base
    raise "Can't find debootstrap command. Aborting." unless command_available? "debootstrap"
    call_debootstrap
    execute "#{rootcmd} apt-get clean"
    execute %{rm -f "#{@config.nfsroot}/etc/resolv.conf" "#{@config.nfsroot}/etc/hostname" "#{@config.nfsroot}/etc/udev/rules.d/70-persistent-net.rules"}
    verb "Creating base.tar.xz"
    execute %{tar --xattrs --acls --one-file-system -C "#{@config.nfsroot}" -cf "#{@config.nfsroot}/var/tmp/base.tar" --exclude etc/machine-id --exclude var/tmp/base.tar --exclude 'var/lib/apt/lists/*_*' .}
    execute %{nice xz -q "#{@config.nfsroot}/var/tmp/base.tar"}, wait: false
  end

  # put contents of ramdisk back to original location on harddisk
  def umount_ramdisk path
    disk = execvalue %{readlink -f "#{path}"}
    ram = "#{disk}-in-ramdisk"

    # test, if a ramdisk is mounted to this directory
    if `mount | grep -q " #{disk} type tmpfs"` && !$?.success?
        trace "Tried to unmount nonexisting ramdisk #{disk}"
        return
    end

    # Overly cautious check to be sure I don’t accidentally delete root.
    raise "Totally invalid disk var." if disk.strip.size < 3

    # There may be a leftover ramdisk from a buggy run.
    if Dir.exist? ram
      # Sometimes there are zombie programs that still are running over the nfsroot.
      # Probably there is better place to kill them than here.
      execute "kill -TERM $(lsof -t /srv/fai/nfsroot)", fail: false, log: false

      execute %{umount "#{ram}"}, log: false
      FileUtils.rmdir ram
    end

    Dir.mkdir ram
    # move current ramdisk to a new location
    execute %{mount --bind "#{disk}" "#{ram}"}, log: false
    execute %{umount "#{disk}"}, log: false

    FileUtils.rm_rf Dir.glob("#{disk}/*")
    FileUtils.rm_rf Dir.glob("#{disk}/.??*")

    # copy ramdisk contents to disk
    FileUtils.cp_r "#{ram}/.", disk, preserve: true

    execute %{umount "#{ram}"}, log: false
    debug "Ramdisk on #{disk} umounted"
    FileUtils.rmdir ram
  end

  def umount_dirs
    debug "Performing cleanup..."

    Dir.chdir "/"
    execute %{umount "#{@config.nfsroot}/proc"}, log: false if Dir.exist? "#{@config.nfsroot}/proc/self"
    execute %{umount "#{@config.nfsroot}/sys"}, log: false if Dir.exist? "#{@config.nfsroot}/sys/class"
    raise "/proc still mounted inside the nfsroot." if Dir.exist? "#{@config.nfsroot}/proc/self"
    execute %{umount "#{@config.nfsroot}/dev/pts"}, log: false, fail: false

    umount_ramdisk "#{@config.nfsroot}/var/lib/dpkg"
    umount_ramdisk "#{@config.nfsroot}/var/cache"

    if !@config.fai_debmirror.nil? and !@config.fai_debmirror.empty?
      if Dir.exist? "#{@config.nfsroot}/#{@config.mirror_mountpoint}"
        execute %{umount "#{@config.nfsroot}/#{@config.mirror_mountpoint}"}, log: false
      end
      FileUtils.rmdir "#{@config.nfsroot}/#{@config.mirror_mountpoint}"
    end

    # show directories still mounted on nfsroot
    mounted = execvalue %{mount | grep " on #{@config.nfsroot} " || true}
    puts "[WARNING]: Still mounted: #{mounted}" unless mounted.empty?
  end

  # Reads target and repository url from /etc/fai/apt/sources.list.
  # Compared to ft nfsroot, this avoids having to pass extra data just for base creation.
  def resolve_sources
    path = "#{@fai_config_dir}/apt/sources.list"
    raise "#{path} does not exist" unless File.exist? path

    lines = File.readlines path

    # By convention, first real repo is main repo. Do some extra checks anyway to be sure.
    suite = nil
    repourl = nil
    lines.each do |line|
      next unless line[0..2] == "deb"
      splitted = line.split " "
      suite = splitted[2]
      repourl = splitted[1]

      break unless suite.include? "/"
      suite = nil
      repourl = nil
    end

    raise "Failed to parse #{path}" if suite.nil? or repourl.nil?

    @repository_suite = suite
    @repository_url = repourl
  end

  def call_debootstrap
    # check if NFSROOT directory is mounted with bad options
    fs = execvalue %{df -P "#{@config.nfsroot}" | tail -1 | awk '{print $6}'}
    if execvalue(%{mount | grep \"on "#{fs}" \" |  awk '{print $6}'}).match?(/nosuid|nodev/)
      raise "NFSROOT directory #{@config.nfsroot} is mounted using nosuid or nodev. Aborting"
    end

    debootstrap_version = execvalue("debootstrap --version").split(" ")[1]
    verb "Creating base system using debootstrap version #{debootstrap_version}"

    # Check if we need cross architecture debootstrap
    matches = @config.fai_debootstrap_opts.match /.*--arch[=[:space:]]([^[:space:]]*)/
    targetarch = matches && matches[1] || ""
    hostarch1 = execvalue "dpkg --print-architecture"
    hostarch2 = execvalue "dpkg --print-foreign-architectures"
    hostarch2 = hostarch1 if hostarch2.empty?

    debootstrap = "debootstrap"

    # Cross arch qemu.
    unless targetarch.empty?
      if targetarch != hostarch1 and targetarch != hostarch2
        debootstrap = "qemu-debootstrap"
        unless command_available? debootstrap
          error "qemu-debootstrap not found. Please install the package qemu-user-static."
        end
      end
    end

    # suite target mirror
    # Calling debootstrap --exclude=wget --include=ca-certificates,gnupg buster /srv/fai/nfsroot http://ryza.smar.fi:3142/debian/debian

    verbose = @@_verbose ? "--verbose" : ""
    command = %{#{debootstrap} --variant=minbase #{verbose} #{@config.fai_debootstrap_opts} "#{@repository_suite}" "#{@config.nfsroot}" "#{@repository_url}"}
    verb "Calling #{command}"
    execute command, fail: false
    unless exec_success?
      puts "ERROR: #{debootstrap} did not complete successfully."
      puts "This is mostly caused by a broken mirror."
      puts "Please check your mirror or use an official mirror."
      puts "Exit code was: #{$?.exitstatus}"
      exit 10
    end

    unless File.exist? "#{@config.nfsroot}/usr/bin/apt-get"
      puts "No apt-get executable available inside the NFSROOT."
      puts "Maybe #{debootstrap} did not finish successfully. Aborting."
      exit 11
    end
  end
end
