
require "optimist"
require "psych"
require "socket"

require_relative "../options"
require_relative "../fai-config"
require_relative "configuration"

# Main class.
#
# Source configuration dir is expected to follow structured layout.
# TODO: this needs to be documented properly.

# TODO: identifier can be changed to site.

# TODO: This script probably should first check if there is updates in git for host configurations.

# TODO: implement --dry-run.
class FaiToolkit::Configuration::SyncConfiguration
  include Options

  def options parser
    common_options parser: parser do |p|
      p.opt :to_target_host,        "Sync configuration from host configurations to target host",
          type: :boolean,   short: :none
      p.opt :from_target_host,      "Sync configuration from target host to host configurations",
          type: :boolean,   short: :none
      p.opt :fai_config_dir,        "FAI’s main config files.",
          type: :string,    short: "C",                   default: "/etc/fai"
      p.opt :confdir,               "Path to directory where configuration will be read from",
          type: :string,    short: "c"
      p.opt :target_site,           "Identifier that describes location where the host is",
          type: :string,    short: "s", required: true
      # TODO: implement this target_area. This probably can be done so that config structure can work with and without it.
      p.opt :target_area,           "Identifier that describes area where the host is",
          type: :string,    short: "a"
      p.opt :target_host,           "Name of the target host",
          type: :string,    short: "H", required: true
    end
  end

  def check_options
    options = @options

    set_opts options

    @target_host = options[:target_host]
    @target_site = options[:target_site]
    @configuration_dir = options[:confdir]

    @config = FaiConfig.new options[:fai_config_dir]
    @config.load_fai_conf
    @config.load_target_config @target_site, @target_host

    if @configuration_dir.nil?
      @configuration_dir = @config.host_configurations
      if @configuration_dir.nil?
        raise "--confdir must be given or it needs to be specified with HOST_CONFIGURATIONS= at fai.conf"
      end

      unless Dir.exist? @configuration_dir
        raise "Given configuration directory “#{@configuration_dir}” does not exist."
      end
    end

    @to_target_host = options[:to_target_host]
    @from_target_host = options[:from_target_host]

    if @to_target_host and @from_target_host
      raise "Only one of --to-target-host and --from-target-host can be specified."
    end

    if !@to_target_host and !@from_target_host
      raise "Either --to-target-host or --from-target-host must be specified."
    end
  end

  def perform
    startup_checks
    real_perform
  end

  private

  # By default, configuration is transferred under /etc to a folder which name comes from package name in hosts.yaml.
  def real_perform
    if @to_target_host
      @config.target.configuration.each do |config_dir|
        local_config_path = "#{@configuration_dir}/sites/#{@target_site}/#{@target_host}/#{config_dir}/"
        remote_config_path = "/#{config_dir}"
        execute %{rsync -avz --max-delete=100 --max-size=3mb --delete -e "ssh -p #{@remote_ssh_port}" #{local_config_path} root@#{@remote_ip}:#{remote_config_path}}
        debug "Sync completed."
      end
    end

    if @from_target_host
      @config.target.configuration.each do |config_dir|
        local_config_path = "#{@configuration_dir}/sites/#{@target_site}/#{@target_host}/#{config_dir}"
        remote_config_path = "/#{config_dir}/"
        execute %{rsync -avz --max-delete=100 --max-size=3mb --delete -e "ssh -p #{@remote_ssh_port}" root@#{@remote_ip}:#{remote_config_path} #{local_config_path} }
        debug "Sync completed."
      end
    end
  end

  def startup_checks
    config = Psych.safe_load(File.readlines("#{@configuration_dir}/hosts.yaml").join)
    raise "hosts.yaml in given configuration directory is malformed." if config["sites"].nil?
    sites = config["sites"]
    raise "Given target site #{@target_site} is missing from hosts.yaml" if sites[@target_site].nil?
    site = sites[@target_site]
    raise "Given target #{@target_host} is missing from hosts.yaml" if site[@target_host].nil?
    host = site[@target_host]

    @remote_ip = host["remote_ip"]
    @remote_ssh_port = host["remote_ssh_port"]

    if @remote_ip.nil? or @remote_ssh_port.nil?
      raise "remote_ip and remote_ssh_port must be specified in hosts.yaml."
    end

    begin
      Socket.tcp(@remote_ip, @remote_ssh_port, connect_timeout: 3) {}
    rescue Errno::ETIMEDOUT => e
      puts.STDERR "Connecting to host #{@remote_ip}:#{@remote_ssh_port} failed in 3 seconds. Please check if there is a connection problem."
      raise e
    end

    raise "Configuration is not specified for given host in hosts.yaml" if host["configuration"].nil?
    @configuration = host["configuration"]

    host_dir = "#{@configuration_dir}/sites/#{@target_host}"
    Dir.mkdir host_dir unless Dir.exist? host_dir
  end
end