
require "optimist"

require_relative "../checks"
require_relative "../fai-config"
require_relative "../options"
require_relative "../nfsroot/ft_nfsroot"

require_relative "create-fai-mirror"
require_relative "methods"
require_relative "mirror-debian"
require_relative "mirror-opensuse"
require_relative "mirror-suse"

# NOTE: Generating the mirror is a bit annoying, since what is going to included is dependent on classes.
# And since classes can be distribution agnostic, I think there should be some system to automatically resolve necessary dependencies.
# For now, classes that should be included should be specified as argument to this command. Though that’s not yet implemented..
class CreateFaiMirror::FaiMirror
  include FaiToolkit::Checks
  include Options
  include CreateFaiMirror::Methods
  include Nfsroot::FtNfsroot

  def perform
    startup_checks
    real_perform
  end

  def options
    argv_copy = ARGV.dup

    options = Optimist::options do
      opt :fai_config_dir,        "FAI’s system configuration directory",
          type: :string,    default: "/etc/fai",            short: "c"
      opt :fai_config_space,      "FAI’s config space. Defaults to nfsroot.conf’s FAI_CONFIGDIR, or /srv/fai/config",
          type: :string,                                    short: "s"
      opt :mirror_location,       "Directory where mirror will be generated to",
          type: :string,    default: "/srv/fai/mirror",     short: "m"
      opt :force,                 "Overwrite existing mirror",
          type: :boolean,   default: false,                 short: "f"
      opt :target_distribution,   "Target distribution. Supported values: debian, suse, opensuse",
          type: :string,                                    short: "t"
      opt :target_distribution_version, "Version of the target distribution. Currently only used by *SUSE. For example 15.2.",
          type: :string,    default: nil,                   short: "D"
      opt :exclude_class,         "Excludes given class if it otherwise would be present",
          type: :string, multi: true,                       short: "x"
      #opt :set_variables,         "Reads and sets variables from class/*.var if used in package_config/<classes>",                            short: "V"
      opt :nfsroot,               "Override FAI nfsroot location from FAI config dir’s nfsroot.conf",
          type: :string,                                    short: "n"
      opt :run_inside_nfsroot,    "Run inside specified FAI nfsroot. This is a path to FAI config, for example /etc/fai.",
          type: :string,                                    short: "r"
      opt :update,                "Update instead of creating new mirror",                  type: :boolean,   default: false,                 short: "u"
      opt :maxpackages,           "How many packages will be processed with single apt-get install command",
          type: :integer,   default: 1,                     short: "M"
      opt :add_packages_from_base, "Add packages to mirror that are defined in $FAI_CONFIG_DIR/NFSROOT.",
                            default: false,                 short: "b"
      opt :target_site,            "Build a mirror for a specific site, defined at host configurations.",
          type: :string,                                    short: "S"
      opt :debug,                 "Show debug output",                                                                                        short: "d"
      opt :verbose,               "Show verbose output",                                                                                      short: "v"
    end

    set_opts options

    unless Dir.exist? options[:fai_config_dir]
      raise "#{options[:fai_config_dir]} does not exist. Please create it or specify alternative path with --fai-config-dir."
    end
    @fai_config_dir = options[:fai_config_dir]

    @config = FaiConfig.new @fai_config_dir
    @config.load_fai_conf
    @config.load_nfsroot_conf

    @fai_config_space = options[:fai_config_space]
    @fai_config_space = @config.config_space if @fai_config_space.nil?

    unless Dir.exist? "#{@fai_config_space}/class"
      path = @fai_config_space
      error "FAI config space (#{path}) seems to be missing or invalid. Create a valid config space or specify path to one with --fai-config-space. " +
          "Config space can also be configured in FAI config’s nfsroot.conf using using FAI_CONFIGDIR=."
    end

    verb "Using config space #{@fai_config_space}"

    # Check if we should take included classes from host configurations.

    @target_site = options[:target_site]
    @included_classes = nil
    if @target_site
      targets = @config.load_site_config @target_site
      @included_classes = []
      targets.each do |slug, target|
        if slug == "default" and target.include? "include_in_mirror"
          @included_classes.concat target["include_in_mirror"]
          next
        end

        @included_classes.concat target["classes"]
      end

      @included_classes.uniq!
    end

    # First check if we need to hop into a chroot.
    unless options[:run_inside_nfsroot].nil?
      run_inside_nfsroot argv_copy, options
      exit
    end

    # Then parse options.

    if options[:nfsroot].nil?
      set_nfsroot @config.nfsroot

      # Not sure what uses this, but it’s exported in original script.
      ENV["NFSROOT"] = @config.nfsroot
    else
      if Dir.exist? options[:nfsroot]
        set_nfsroot options[:nfsroot]

        # Not sure what uses this, but it’s exported in original script.
        ENV["NFSROOT"] = options[:nfsroot]
      else
        raise "#{options[:nfsroot]} does not exist. Please create it or specify alternative path with --nfsroot."
      end
    end

    @add_packages_from_base = options[:add_packages_from_base]

    @target_distribution = options[:target_distribution]
    check_target_distribution options[:target_distribution] do |distro|
      case @target_distribution
      when "debian"
        raise "apt-get is missing. Are you sure you are on Debian?" unless command_available? "apt-get"
        @installer = CreateFaiMirror::MirrorDebian.new options
      when "suse"
        raise "zypper is missing. Are you sure you are on SUSE?" unless command_available? "zypper"
        @installer = CreateFaiMirror::MirrorSuse.new options
      when "opensuse"
        raise "zypper is missing. Are you sure you are on OpenSUSE?" unless command_available? "zypper"
        @installer = CreateFaiMirror::MirrorOpenSuse.new options
      else
        raise "Invalid target distribution: #{@target_distribution.inspect}"
      end
    end

    check_target_distribution_version options[:target_distribution_version]

    @installer.config = @config
    @installer.fai_config_dir = @fai_config_dir
    @installer.target_distribution_version = @target_distribution_version
    @installer.global_options = @global_options

    @force = options[:force]
    @update = options[:update]
    if @force and @update
      raise "Only --force or --update can be given."
    end

    if options[:mirror_location].size < 3
      raise "Please specify proper mirror location. Root is not allowed for safety."
    end
    if Dir.exist? options[:mirror_location]
      if @force
        FileUtils.rm_rf "#{options[:mirror_location]}/#{@target_distribution}"
      elsif @update
        # Updating, so there is nothing else to do.
      else
        raise "Mirror #{options[:mirror_location]} already exists. If you want to redo it, use --force. If you want to update it, use --update."
      end
    else
      if @update
        raise "For --update, mirror dir #{options[:mirror_location]} needs to exist."
      else
        FileUtils.mkdir_p options[:mirror_location]
      end
    end
    @mirror_location = options[:mirror_location]

    @excluded_classes = options[:exclude_class]

    @maxpackages = options[:maxpackages]
  end

  private

  def real_perform

    @installer._pre_real_perform @fai_config_dir, @mirror_location

    verb "Getting package information"
    @installer.refresh_repositories

    # In fai-mirror, this is executed if no -B or -b flags are given. Our flow is different.
    if @add_packages_from_base
      @installer.add_base_packages
    end

    @installer.download_packages_for_classes @classes, @fai_config_space

    umount_dirs

    # TODO: implement?
    #[ @add -eq 0 ] && delete_base_packages

    @installer._post_real_perform @mirror_directory

    verb "create-fai-mirror finished."
    if opts[:debug]
      debugdf = execvalue "du -sh --exclude aptcache --exclude db #{@mirror_directory}"
      debug "Mirror location: #{@mirror_directory.inspect}"
      debug "Mirror size: #{debugdf}"
    end
    cleandirs
  end

  def run_inside_nfsroot argv_copy, options
    nfsroot_fai_config_dir = options[:run_inside_nfsroot]

    raise "When using --run-inside-nfsroot, root is required (to construct nfsroot)." unless Process.uid == 0

    verb ""
    verb "[IMPORTANT] Using chroot nfsroot for mirroring; with a separate version of this script inside chroot."
    verb ""
    sleep 1.8

    unless Dir.exist? "#{nfsroot_fai_config_dir}"
      raise "Given FAI config dir #{nfsroot_fai_config_dir} does not exist. Please create your config dir for the mirror nfsroot."
    end

    nfsroot_config = FaiConfig.new nfsroot_fai_config_dir
    nfsroot_config.load_fai_conf
    nfsroot_config.load_nfsroot_conf

    if @config.host_configurations and not nfsroot_config.host_configurations
      error "When main FAI config dir has host configurations, --run-inside-nfsroot’s config dir " +
          "also needs to declare host configurations path."
    end


    vdflags = ""
    vdflags << "-d" if options[:debug]
    vdflags << " -v" if options[:verbose]

    unless Dir.exist? nfsroot_config.nfsroot
      if Dir.exist? nfsroot_config.nfsroot
        raise "FAI nfsroot #{nfsroot_config.nfsroot} seems to already exist, but it’s not valid nfsroot. Please move it away if it is unwanted or use another path."
      end

      run_nfsroot nfsroot_fai_config_dir, dracut: false
    else
      begin
        run_nfsroot nfsroot_fai_config_dir, dracut: false, refresh: true
      rescue Exception => e
        puts e.full_message
        exit 1
      end
    end

    verb "nfsroot checked."
    verb ""

    # Sync FAI config space to the chroot.
    cleaned = @fai_config_space.delete_suffix "/"
    nfsroot_config_space_path = "#{nfsroot_config.nfsroot}/#{nfsroot_config.config_space}/"
    dir = File.dirname nfsroot_config_space_path
    FileUtils.mkdir_p dir unless Dir.exist? dir
    execute %{rsync -a --delete #{cleaned}/ #{nfsroot_config_space_path}}

    # Sync FAI config dir to the chroot.
    cleaned = nfsroot_fai_config_dir.delete_suffix "/"
    temp_fai_config_dir = "#{nfsroot_config.nfsroot}/#{cleaned}"
    FileUtils.mkdir_p temp_fai_config_dir unless Dir.exist? temp_fai_config_dir
    execute %{rsync -a --delete #{cleaned}/ #{temp_fai_config_dir}/}

    # Sync FAI host configurations to the chroot.
    if @config.host_configurations
      execute %{rsync -a --delete #{@config.host_configurations}/ #{nfsroot_config.nfsroot}/#{nfsroot_config.host_configurations}/}
    end

    # Instruct to use correct FAI config
    args = []
    set_correct_config_space = false
    remove_fai_config_dir_arg = false
    argv_copy.each do |arg|

      if remove_fai_config_dir_arg
        arg = ""
        remove_fai_config_dir_arg = false
      end

      if arg == "-c" or arg == "--fai-config-dir"
        arg = ""
        remove_fai_config_dir_arg = true
      end

      if set_correct_config_space
        arg = ""
        set_correct_config_space = false
      end

      # If custom config space is given, remove the argument as it has been copied to default location inside nfsroot.
      if arg  == "-s" or arg == "--fai-config-space"
        set_correct_config_space = true
        arg = ""
      end

      if arg == "-r" or arg == "--run-inside-nfsroot"
        arg = "--fai-config-dir"
      end

      args << arg
    end
    args = args.join(" ").strip

    # TODO: This kind of contradicts whatever set_correct_config_space is trying to do, but let’s try...
    args += " --fai-config-space #{@config.config_space}"

    execute %{chroot #{nfsroot_config.nfsroot} create-fai-mirror #{args} --nfsroot /}
  end

  def startup_checks
    at_exit do
      umount_dirs
    end

    # This should trick install_packages not to use chroot when we are
    # already in chroot, by --run-inside-nfsroot.
    ENV["FAI_ROOT"] = "/" if opts[:nfsroot] == "/"

    @installer.startup_checks

    @mirror_directory = "#{@mirror_location}/#{@target_distribution}"
    FileUtils.mkdir @mirror_directory unless opts[:update]

    "package_config/ missing in #{@fai_config_space}" unless Dir.exist? "#{@fai_config_space}/package_config"

    # Defaults to 1.
    ENV["MAXPACKAGES"] = @maxpackages.to_s

    set_classes

    # TODO: implementing this is not easy, since it requires some conversion from bash to ruby.
    # Probably those files could be read and parsed manually, but not doing that now.
=begin
    if opts[:set_variables]
      variable_files = mk_class_var_list
      set -a
      for f in $flist; do
        echo Sourcing variable definitions in $f
        source $flist
      end
      set +a
    end
=end

    @installer._post_initialize @mirror_directory
  end
end