require "fileutils"

require_relative "mirror-impl"

class CreateFaiMirror::MirrorDebian < CreateFaiMirror::MirrorImpl

  def _post_initialize mirror_directory
    # holds the package cache data
    @aptcache = "#{mirror_directory}/aptcache"

    @archivedir = "#{@aptcache}/var/cache/apt/archives"
    statefile = "#{@aptcache}/statefile"

    # also used in install_packages.conf
    @aptoptions = <<~EOF
      -o Aptitude::Log=/dev/null
      -o Aptitude::CmdLine::Ignore-Trust-Violations=yes
      -o APT::Get::AllowUnauthenticated=true
      -o Acquire::AllowInsecureRepositories=true
      -o DPkg::force-conflicts::=yes
      -o Dir::State=#{@aptcache}/var/lib/apt
      -o Dir::Log=#{@aptcache}/var/log/apt
      -o Dir::State::extended_states=#{@aptcache}/var/lib/apt/lists/extended_states
      -o Dir::State::status=#{statefile}
      -o Dir::Cache=#{@aptcache}/var/cache/apt
      -o Dir::State=#{@aptcache}/var/cache/apt
      -o Dir::Cache::Archives=#{@archivedir}
      -o Dir::Etc=#{@aptcache}/etc/apt/
      -o Dir::State::Lists=#{@aptcache}/var/lib/apt/lists/
    EOF
    @aptoptions.gsub! "\n", " "

    # Add architecture.
    if !@arch.nil? and !@arch.empty?
      @aptoptions += " -o APT::Architecture=#{@arch}"
    end

    # Used by install_packages.
    ENV["aptoptions"] = @aptoptions
    ENV["FAI_ARCHIVE_DIR"] = @archivedir

    # we only need some empty dirs
    FileUtils.mkdir_p [ "#{@archivedir}/partial", "#{@aptcache}/etc/apt/preferences.d", "#{@aptcache}/var/lib/apt/lists/partial", "#{@aptcache}/var/log/apt" ]
    File.open statefile, "w" do |f|
      # Empties statefile.
    end
  end

  # Downloads packages that are included in the mirror.
  def download_packages package_list
    package_list = package_list.join(" ") if package_list.respond_to? :join
    execute "apt-get -d #{@aptoptions} -y --fix-missing install #{package_list}", direct_output: true
  end

  def refresh_repositories
    execute "apt-get #{@aptoptions} update", log: false
  end

  def startup_checks
    raise "reprepro not found. Please install the package." unless command_available? "reprepro"
  end

  # Add packages from base.tar.xz and additional packages in nfsroot.
  def add_base_packages
    # arch dependent packages defined in fai-make-nfsroot
    verb "Adding packages of #{@fai_config_dir}/NFSROOT."
    if File.exist? "#{nfsroot}/var/tmp/packages.nfsroot"
        plist = File.readlines("#{nfsroot}/var/tmp/packages.nfsroot")[0].split(" ")
        download_packages plist
    else
        STDERR.puts "WARNING: #{nfsroot}/var/tmp/packages.nfsroot does not exist."
        STDERR.puts "Can't add those packages. Maybe the nfsroot is not yet created."
        exit 1
    end

    use_temporary_sources_list

    base_packages_file = "#{nfsroot}/var/tmp/base-pkgs.lis"
    verb "Adding packages from #{base_packages_file}"
    if File.exist? base_packages_file
        plist = File.readlines base_packages_file
        plist.map! do |package|
          package.strip
        end
        refresh_repositories
        download_packages plist
    else
        STDERR.puts "WARNING: #{base_packages_file} does not exist."
        STDERR.puts "Can't add those packages. There is some problem with your nfsroot."
        exit 1
    end

    restore_sources_list
  end

  def download_packages_for_classes classes, fai_config_space
    verb "Downloading packages for classes: #{classes}"

    vflag = ""
    vflag << " -v" if @opts[:verbose]

    command = %{FAI="#{fai_config_space}" install_packages -N -d #{vflag}}
    command = %{FAI_ROOT="/" #{command}} if ENV["FAI_ROOT"].nil?

    execute command, direct_output: true, fail: false
    unless exec_success?
        raise "ERROR when downloading packages. Your mirror probably is broken."
    end
  end

  # Hooks to make it visually easier to understand the flow.

  def _pre_real_perform fai_config_dir, mirror_location
    # if we are using nfs mounts for Debian mirror, this may fail here, since inside a chroot environment different dir are used
    # if sources.list includes file AND @config.fai_debmirror is defined we Mhave to mount
    # otherwise mounting is not needed. call task_mirror
    if !@config.fai_debmirror.nil? and !@config.fai_debmirror.empty?
        FileUtils.mkdir_p @mirror_location
        execute "mount -r #{@config.fai_debmirror} #{@mirror_location}"
    end

    sources_list_dir = "#{fai_config_dir}/apt"
    if @opts[:target_site] and Dir.exist? "#{fai_config_dir}/apt/mirror_lists/#{@opts[:target_site].downcase}"
      sources_list_dir = "#{fai_config_dir}/apt/mirror_lists/#{@opts[:target_site].downcase}"
    end
    debug "Using sources.list from #{sources_list_dir}"

    # TODO: use -p to preserve sources.list
    execute "sed -e 's/file:/copy:/' #{sources_list_dir}/sources.list > #{@aptcache}/etc/apt/sources.list"
    unless Dir.glob("#{sources_list_dir}/sources.list.d/*").empty?
      execute "sed -e 's/file:/copy:/' #{sources_list_dir}/sources.list.d/* >> #{@aptcache}/etc/apt/sources.list"
    end

    FileUtils.cp(Dir.glob("#{fai_config_dir}/apt/preferences.d/*"), "#{@aptcache}/etc/apt/preferences.d/")

    # aptpref comes from cli arg on original script. Not implemented.
    #if File.exist? @aptpref
    #  FileUtils.cp @aptpref, "#{@aptcache}/etc/apt/preferences"
    #end

    # use the trusted keys from the host
    execute "cp -a /etc/apt/trusted.gpg.d #{@aptcache}/etc/apt"
  end

  def _post_real_perform mirror_directory
    # Default to host architecture.
    if @arch.nil? or @arch.empty?
      @arch = execvalue "dpkg  --print-architecture"
      @arch = "#{@arch} #{execvalue "dpkg --print-foreign-architectures"}"
    end

    # create mirror directory structure
    verb "Calling reprepro"
    FileUtils.mkdir "#{mirror_directory}/conf" unless Dir.exist? "#{mirror_directory}/conf"
    File.open "#{mirror_directory}/conf/distributions", "w" do |f|
      f.write <<~EOF
      # generate config file for reprepro
      Codename: cskoeln
      Architectures: #{@arch}
      Components: main non-free contrib
      DebIndices: Packages Release . .xz
      Description: Package reprository created by fai-mirror
      Label: fai-project.org
      Origin: fai-mirror
      EOF
    end

    # check if backports are used
    bponame = execvalue "grep -E '^deb ' #{@aptcache}/etc/apt/sources.list | grep -P -o '\S+-backports'"

    # add entry for backports
    unless bponame.empty?
      File.open "#{mirror_directory}/conf/distributions", "a" do |f|
        f.write <<~EOF

        Codename: #{bponame}
        Architectures: #{@arch}
        Components: main non-free contrib
        DebIndices: Packages Release . .xz
        Description: Package reprository created by fai-mirror
        Label: fai-project.org
        Origin: fai-mirror
        EOF
      end

      # maybe using reprepro pulls it's possible to move instead of copy the packages
      execvalue = "ls #{@archivedir}/*~bpo[0-9]*+[0-9-]*_*.deb"
      if exec_success? and !execvalue.empty?
        execute "reprepro -b #{mirror_directory} includedeb #{bponame} #{@archivedir}/*~bpo[0-9]*+[0-9-]*_*.deb"
        FileUtils.rm_f Dir.glob("#{@archivedir}/*~bpo[0-9]*+[0-9-]*_*.deb")
      end
    end

    # Allow package without the .deb suffix. At least they must start with a char or number and contain an underscore or hyphen.
    execute "reprepro --ignore=extension -b #{mirror_directory} includedeb cskoeln #{@archivedir}/[0-9a-zA-z]*[_-]*"
    FileUtils.rm_r Dir.glob("#{@archivedir}/[0-9a-zA-z]*[_-]*")
  end

  # Quasi-hooks

  def use_temporary_sources_list
    # Save original sources.list
    FileUtils.mv "#{@aptcache}/etc/apt/sources.list", @aptcache

    # Construct FAI_DEBOOTSTRAP string from sources.list

    #FAI_DEBOOTSTRAP="buster http://ryza.smar.fi:3142/debian/debian"
    # Let’s assume first entry is main source.
    config_dir_sources_list = "#{@opts[:fai_config_dir]}/apt/sources.list"
    lines = File.readlines config_dir_sources_list
    repo_url = nil
    repo_target = nil
    lines.each do |line|
      line.strip!
      if line[0..3] == "deb "
        splitted = line.split " "
        next if splitted.size < 3

        repo_url = splitted[1]
        repo_target = splitted[2]
        break
      end
    end

    raise "Failed to parse main repo from #{config_dir_sources_list}." if repo_target.nil?
    fai_debootstrap = "#{repo_target} #{repo_url}"

    # now use different sources.list for debootstrap packages
    execute %{echo "#{fai_debootstrap}" | awk '{printf "deb %s %s main\\n",$2,$1}' | sed -e 's/file:/copy:/' > #{@aptcache}/etc/apt/sources.list}
  end

  def restore_sources_list
    FileUtils.mv "#{@aptcache}/sources.list", "#{@aptcache}/etc/apt/sources.list"
    refresh_repositories
  end

  def try_base_package_delete p, arch
    if Dir.glob("#{@archivedir}/#{p}_*#{arch}.deb").size > 0
      debug "deleting package #{p} #{arch}"
      FileUtils.rm Dir.glob("#{@archivedir}/#{p}_*#{arch}.deb")
    end
  end
end