
require_relative "fai-toolkit"
require_relative "output"

# Methods to execute processes.
module FaiToolkit::Execution
  include Output

  class CommandError < StandardError
  end

  class ExecutionFailed < CommandError
  end

  private

  # Check if requested program can be found.
  #
  # Should work cross-platform and in restricted environents pretty well.
  def command_available? command
    exts = ENV["PATHEXT"] ? ENV["PATHEXT"].split(';') : ['']
    ENV["PATH"].split(File::PATH_SEPARATOR).each do |path|
      exts.each do |ext|
        exe = File.join(path, "#{command}#{ext}")
        return true if File.executable?(exe) && !File.directory?(exe)
      end
    end
    false
  end

  # Returns path to given executable.
  def which command, fail: true
    exts = ENV["PATHEXT"] ? ENV["PATHEXT"].split(';') : ['']
    ENV["PATH"].split(File::PATH_SEPARATOR).each do |path|
      exts.each do |ext|
        exe = File.join(path, "#{command}#{ext}")
        return exe if File.executable?(exe) && !File.directory?(exe)
      end
    end
    error "Failed to find #{command} from PATH." if fail
    nil
  end

  # Returns pid of executed command if wait is false.
  # Returns command output if wait is true.
  #
  # fail == If true, an error is raised in case the command returns failure code.
  def execute command, fail: true, wait: true, log: true, direct_output: false, env: {}
    # The debug message contents will always be evaluated, so can just do it like this.
    log_command = command.respond_to?(:join) && command.join(" ") || command
    if @dry_run
      puts "Would execute “#{log_command.force_encoding('UTF-8')}” at #{caller_locations(1,1)[0]}"
      return
    end
    debug "Executing “#{log_command.force_encoding('UTF-8')}” at #{caller_locations(1,1)[0]}" if log == true

    if command.respond_to? :map!
      command.map! do |argument|
        argument.nil? && "" || argument
      end
    end

    r, w = IO.pipe
    if direct_output
      pipe_w = w
      w = STDOUT
    end
    # If command is string, splat operator (the *) does not do anything. If array, it expand the arguments.
    # This causes spawning work correctly, as that’s how spawn() expects to have the arguments.
    pid = spawn(env, *command, err: [:child, :out], out: w)
    w = pipe_w if direct_output
    w.close
    unless wait
      r.close
      return pid
    end
    buffer = String.new capacity: 1024
    output = ""
    while not r.eof?
      r.read_nonblock 1024, buffer
      if log
        print buffer if @@_verbose
        @log_file&.write buffer
      end
      output << buffer
    end
    r.close

    # This is instant return, but allows us to have $? set.
    Process.wait pid

    raise ExecutionFailed, "Command execution failed.", caller if fail == true and !$?.success?

    # read_nonblock() uses read(), which always reads as ASCII-8BIT:
    #   In the case of an integer length, the resulting string is always in ASCII-8BIT encoding.
    output.force_encoding("UTF-8").strip
  end

  # Same as ::execute, but uses log: false to avoid unnecessary output when we’re just getting a value for internal needs.
  #
  # Defaults to fail: false, since when one just wants to get the value, there is not that much need to unconditionally die if getting bad exit code.
  def execvalue *args, fail: false
    if @dry_run
      puts "Would execvalue “#{args[0].force_encoding('UTF-8')}” at #{caller_locations(1,1)[0]}"
      return
    end
    trace "Execvaluing “#{args[0].force_encoding('UTF-8')}” at #{caller_locations(1,1)[0]}"
    execute *args, fail: fail, log: false
  end

  def exec_success?
    $?.success?
  end
end
