
require "pathname"
require "psych"

require_relative "common"

# Convenience class to load FAI config files.
class FaiConfig
  include Common

  attr_reader :nfsroot

  # FAI config space.
  #
  # This shouldn’t be modified by hand, but in some cases it’s cleanest way, so that configuration can be used wherever required.
  #
  # Defined by FAI_CONFIGDIR at nfsroot.conf.
  #
  # @see #config_space=
  attr_reader :config_space

  attr_reader :nfsroot_etc_hosts

  # This is just for mimicking how official scripts works.
  attr_reader :fai_rootpw, :fai_debootstrap_opts

  # From fai.conf
  attr_reader :fai_config_src, :loguser, :fai_logproto, :fai_debmirror, :fai_monitor_port
  attr_reader :mirror_mountpoint

  # @deprecated in favor of :config_space, as two duplicate values don’t make much sense.
  attr_reader :nfsroot_config_space

  # Host-specific configuration directory, containing for example IP addresses of the hosts
  # and other miscellaneous information that does not fit under FAI config space.
  #
  # @fai_toolkit_specific
  attr_reader :host_configurations

  # Base dir of where mirrors are generated to and read from, in respective scripts.
  #
  # @fai_toolkit_specific
  attr_reader :fai_mirror_dir

  # Site specific target host’s configuration.
  #
  # Contains a FaiConfig::Target struct.
  attr_reader :target

  # Distribution an nfsroot is created for.
  #
  # Identificators will be specified better later, but probably this should be something like
  # “debian”, “opensuse“ or “suse”.
  attr_reader :nfsroot_distribution

  # Version of te distribution an nfsroot is created for.
  #
  # Freeform string that for example for OpenSUSE should be like “15.2”.
  attr_reader :nfsroot_distribution_version

  # Compression algorithm dracut will use to compress initrd image.
  #
  # Defined in nfsroot.conf.
  attr_reader :dracut_compress

  # Array of additional packages that should be installed by rinse, when creating a base archive.
  #
  # Probably not used for Debian.
  attr_reader :rinse_additional_packages

  Target = Struct.new(:image_name,
                      :grub_config,
                      :classes,
                      :connection,
                      :configuration,
                      :distribution,
                      :distribution_version,
                      :hostname)

  def initialize fai_config_dir
    @fai_config_dir = fai_config_dir

    @mirror_mountpoint = "/media/mirror"
    @nfsroot_config_space = "/var/lib/fai/config"
  end

  def config_space= path
    ensure_clean_state path if path
    @config_space = convert_to_absolute_path(path)
  end

  def fai_mirror distribution: nil
    return nil if @fai_mirror_dir.nil?

    if distribution.nil?
      if @target&.distribution
        distribution = @target.distribution
      end
    end

    if distribution.nil?
      raise "distribution must be given or it needs to be defined in host configurations."
    end

    "#{@fai_mirror_dir}/#{distribution}"
  end

  def load_nfsroot_conf
    parse_config_file "#{@fai_config_dir}/nfsroot.conf" do |variable, value|
      #puts variable.inspect
      #puts value.inspect

      case variable
      when "NFSROOT"
        @nfsroot = slow_resolve_symlinks(value)
      when "FAI_CONFIGDIR"
        # Upstream FAI project is using this strange CONFIGDIR name for config space in many variables...
        self.config_space=(value)
      when "FAI_ROOTPW"
        # This is just for mimicking how official scripts works.
        @fai_rootpw = value
      when "FAI_DEBOOTSTRAP_OPTS"
        # This is just for mimicking how official scripts works.
        @fai_debootstrap_opts = value
      when "NFSROOT_ETC_HOSTS"
        @nfsroot_etc_hosts = value
      when "DISTRIBUTION"
        @nfsroot_distribution = value
      when "DISTRIBUTION_VERSION"
        @nfsroot_distribution_version = value
      when "DRACUT_COMPRESS"
        @dracut_compress = value
      when "RINSE_ADDITIONAL_PACKAGES"
        unless value.nil?
          value = value.split " "
        end
        @rinse_additional_packages = value
      end
    end
  end

  def load_fai_conf

    parse_config_file "#{@fai_config_dir}/fai.conf" do |variable, value|

      # There is many variables that could be used (see man fai.conf):
      # - FAI_CONFIG_SRC
      # - LOGUSER
      # - FAI_LOGPROTO
      # - FAI_DEBMIRROR
      # - FAI_MONITOR_PORT

      # There is also some undocumented variables:
      # - FAI
      # - MNTPOINT

      case variable
      when "FAI_CONFIG_SRC"
        ensure_clean_state value if value
        @fai_config_src = value
      when "LOGUSER"
        @loguser = value
      when "FAI_LOGPROTO"
        @fai_logproto = value
      when "FAI_DEBMIRROR"
        @fai_debmirror = value
      when "FAI_MONITOR_PORT"
        @fai_monitor_port = value
      when "MNTPOINT"
        @mirror_mountpoint = value
      when "FAI"
        ensure_clean_state value if value
        @nfsroot_config_space = value
      when "HOST_CONFIGURATIONS"
        if value
          value = convert_to_absolute_path(value)
          ensure_clean_state value
        end
        @host_configurations = value
      when "FAI_MIRROR_DIR"
        @fai_mirror_dir = value
      else
        puts "WARNING: Unrecognized variable in fai.conf: #{variable}"
      end
    end
  end

  # Returns absolute path to grub config file in host configurations.
  def host_grub_path grub_config, default
    grub_config = default["grub_config"] if grub_config.nil?

    grub_config = "#{@host_configurations}/grub/#{grub_config}"
    File.exist?(grub_config) && grub_config || nil
  end

  def load_site_config target_site
    if @host_configurations.nil?
      raise "HOST_CONFIGURATIONS variable needs to be set in fai.conf in order to load site specific configuration."
    end

    site_path = "#{@host_configurations}/sites/#{target_site}"
    unless Dir.exist? site_path
      raise "Site directory #{site_path} does not exist. Please fix your host configurations directory."
    end

    targets_file = "#{site_path}/targets.yaml"

    unless File.exist? targets_file
      raise "Targets file #{targets_file} is missing. Please fix your host configurations directory."
    end

    Psych.safe_load(File.open(targets_file))
  end

  def load_target_config target_site, target_host
    targets = load_site_config target_site

    default = nil
    target = nil
    targets.each do |host, configuration|
      if default.nil? and host == "default"
        default = configuration
        next
      end
      if target.nil? and host == target_host
        target = configuration
        next
      end
    end

    if target.nil?
      STDERR.puts "Target host #{target_host} on site #{target_site} can’t be found in host configuration."
      exit FaiToolkit::TARGET_HOST_NOT_FOUND
    end

    grub_config = host_grub_path target["grub_config"], default
    if grub_config.nil?
      targets_file = "#{@host_configurations}/sites/#{target_site}/targets.yaml"
      raise "GRUB config file specified in #{targets_file} for host #{target_host} does not exist."
    end

    connection = nil
    if target["connection"]
      connection = {}
      connection["remote_ip"] = target["connection"]["remote_ip"]
      connection["remote_ssh_port"] = target["connection"]["remote_ssh_port"]
    end

    configuration = nil
    if target["configuration"]
      # Dummy to ensure we’re getting an array.
      configuration = target["configuration"].to_a
    end

    hostname = target["hostname"]
    if hostname.nil?
      # Default to target’s slug.
      hostname = target_host
    end

    @target = Target.new
    @target.image_name =              target["image_name"]
    @target.grub_config =             grub_config
    @target.classes =                 target["classes"]
    @target.connection =              connection
    @target.configuration =           configuration
    @target.distribution =            target["distribution"]
    @target.distribution_version =    target["distribution_version"]
    @target.hostname =                hostname
  end

  def unload_target_config
    @target = nil
  end

  private

  def parse_config_file configfile
    variableregex = /^(\w+)=(.*)$/
    valueregex = /^([^"^'.].*)(#.*[^"^'])$/

    File.readlines(configfile).each do |line|
      line = line.strip
      matches = line.match variableregex

      next unless matches

      variable = matches[1]

      valuedata = matches[2].strip

      # Remove potential comments
      if valuedata.match? valueregex
        valuedata.sub! valueregex, '\1'
      end

      value = valuedata.strip

      # Strip quotes
      value.delete_suffix!('"')
      value.delete_prefix!('"')
      value.delete_suffix!("'")
      value.delete_prefix!("'")

      yield variable, value
    end
  end

  def convert_to_absolute_path relative_path
    File.absolute_path relative_path
  end

  # Recursively resolves symlinks, so that paths that does not yet exist can get their symlinks resolved too.
  #
  # Usually you don’t want to use this method.
  def slow_resolve_symlinks input_path
    path = Pathname.new(input_path)
    path.ascend do |part|
      begin
        resolved = part.realpath.to_s
        input_path = input_path.to_s
        input_path.sub! part.to_s, resolved
        break
      rescue Errno::ENOENT
        # If path is not accessible, conversion can’t work, so let’s try parent directory.
        next
      end
    end

    input_path
  end
end