
require "optimist"

require_relative "common"
require_relative "output"
require_relative "fai-config"
require_relative "version"

# Main class.
class FT
  include Common
  include Output

  # Bypass root checks.
  #
  # This is probably mainly useful for testing.
  attr_accessor :no_root_check

  # Can be used to get object created from subcommand.
  attr_reader :command_object

  COMMAND_MAP ||= {
    "base"                => "Create a base image to use with FAI operations.",
    "fai-reprepro-mirror" => "Create a FAI mirror with custom reprepro configuration.",
    "medium"              => "Create a FAI install/update medium.",
    "nfsroot"             => "Manage a minimal chroot installation used for various FAI operations",
    "nfsroot-base"        => "Create a base image how nfsroot would create it. “ft base” is probably what you want to use.",
    "packages"            => "Install or download packages specified in FAI config space.",
    "qemu"                => "Perform operations inside QEMU",
    "sync-configuration"  => "Update configuration on target host"
  }

  def options
    @subopts, @command = nil, nil
    @global_options = Optimist::options do
      version "ft-#{FaiToolkit::VERSION}"

      banner "Usage:"
      banner "  ft [global options] [<command> [suboptions]]\n \n"
      banner "Global options:"
      opt :verbose,           "Show verbose output",
          short: "v"
      opt :debug,             "Show debug output (useful when debugging a problem)",
          short: "d"
      opt :trace,             "Show trace output (noisiest, probably not useful for most of time)",
          short: :none
      opt :logging,            "--no-logging can be used to fully disable logging to /var/log/fai.",
          short: :none,       default: true
      opt :version,           "Print version and exit",
          short: "V"
      opt :help,              "Show help message",
          short: "h"

      educate_on_error

      stop_on COMMAND_MAP.keys
      #stop_on_unknown

      banner "\nft supports following commands:"
      COMMAND_MAP.each { |cmd, desc| banner format("  %-20s %s", cmd, desc) }
    end
    if ARGV.empty?
      Optimist::die "No arguments given"
    end

    @input_command = ARGV.shift
    @command = @input_command.gsub "-", "_"
    command_method_name = "opt_#{@command}"

    if not self.class.private_instance_methods.include?(command_method_name.to_sym)
      Optimist::die "ft does not recognize given command “#{@input_command}”"
    end

    # dispatch to command handling method
    self.send(command_method_name)
  end

  def perform
    @command_object.perform
  end

  private

  def opt_qemu
    #require_root # TODO: necessary?

    require_relative "qemu/qemu"
    @command_object = Qemu::Qemu.new
    parse_command_arguments
  end

  def opt_nfsroot
    require_root # For the chroot.

    require_relative "nfsroot/create-nfsroot"
    @command_object = Nfsroot::CreateNfsroot.new
    parse_command_arguments
  end

  def opt_medium
    require_root

    require_relative "media/create-update-cd"
    @command_object = Media::CreateUpdateCd.new
    parse_command_arguments
  end

  def opt_packages
    #require_root # For the chroot.
    @run_by_default = true

    require_relative "packaging/packages"
    @command_object = Packaging::Packages.new
    parse_command_arguments
  end

  def opt_fai_reprepro_mirror
    require_root # For the chroot.

    require_relative "mirror/fai-reprepro-mirror"
    @command_object = Mirror::FaiRepreproMirror.new
    parse_command_arguments
  end

  def opt_nfsroot_base
    require_root

    require_relative "base/nfsroot-base"
    @command_object = FaiToolkit::Base::NfsrootBase.new
    parse_command_arguments
  end

  def opt_base
    require_root

    result = require_relative "base/create-base"
    @command_object = FaiToolkit::Base::CreateBase.new
    parse_command_arguments
  end

  def opt_sync_configuration
    require_relative "configuration/sync-configuration"
    @command_object = FaiToolkit::Configuration::SyncConfiguration.new
    parse_command_arguments
  end

  def parse_command_arguments
    # Necessary to have this here to avoid variable missing inside banner.
    command = @input_command

    @command_object.parse_global_options @global_options

    parser = Optimist::Parser.new
    parser.banner "options for #{command}"

    # This causes command specific arguments to be declared.
    @command_object.options parser

    parser.educate_on_error

    options = Optimist::with_standard_exception_handling parser do
      args = ARGV.clone

      result = parser.parse ARGV

      if not @run_by_default
        raise Optimist::HelpNeeded if args.empty?
      end

      result
    end

    #if ARGV.empty?
    #  Optimist::die "No arguments given"
    #end

    @command_object.options = options
    @command_object.check_options
  end

  def require_root
    return if @no_root_check
    raise "“ft #{@input_command}” must be run as root." unless Process.uid == 0
  end

end
