
require "fileutils"
require "tempfile"
require "tmpdir"

require "optimist"

require_relative "../checks"
require_relative "../options"
require_relative "../fai-config"
require_relative "update-cd-methods"

# TODO: how base files should be supplied?
# Now, they are inside config space’s basefiles dir, but this isn’t really a good way; probably a base file should be explicitly pointed out, so those that are wanted can be bundled, not all of them.

# TODO: mirror dir could also be defined in host configurations. That would allow target specific mirrors.

# Main class.
class Media::CreateUpdateCd
  include FaiToolkit::Checks
  include Options
  include Media::UpdateCdMethods

  LIVEOS_PATH = "squashfs-root/LiveOS/"

  def initialize
    # TODO: These should be configurable somehow. Maybe have a configuration file?
    @hidedirs = [
      "/usr/share/locale",
      "/usr/share/doc",
      "/var/lib/apt",
      "/var/lib/aptitude",
      "/var/cache/debconf",
      "/var/cache/apt",
      "/var/cache/man",
      "/usr/share/man",
      "/var/lib/dpkg/info",
      "/media/mirror/aptcache",
      "/boot" # There shouldn’t be a need to have kernel and initrd, as they are inside the squashfs already.
    ]

  end

  def perform
    startup_checks
    real_perform

    # Triple-remove to be sure they actually gets removed...
    FileUtils.remove_entry @tempdir if Dir.exist? @tempdir
    FileUtils.remove_entry @scratch if Dir.exist? @scratch
  end

  def options parser
    options = common_options parser: parser, log_file: "create-update-cd.log" do |p|
      p.opt :fai_config_dir,    "FAI’s system configuration directory",
          type: :string,    short: "c",   default: "/etc/fai"
      p.opt :fai_config_space,  "FAI config space to bundle in generated install medium. Defaults to FAI_CONFIGDIR in /etc/nfsroot.conf. You should use nfsroot.conf instead of this flag.",
          type: :string,    short: :none
      p.opt :mirror_dir,        "Location of fai-mirror",
          type: :string,    short: "m"
      p.opt :grub_config,       "Path to grub.cfg for this image",
          type: :string,    short: "g"
      p.opt :filename,          "Filename to generate image to. For example image.iso",
          type: :string,    short: "n"
      p.opt :initrd,            "Use a custom initrd instead of one from NFSROOT",
          type: :string,    short: "i"
      p.opt :target_distribution, "Target distribution. Currently support values are “debian” and “suse”.",
          type: :string,    short: "t"
      p.opt :media_label,         "Name/label of the CD",
          type: :string,    short: "l"
      p.opt :target_site,         "Optionally specify a site that can be used to define default values for various settings",
          type: :string,    short: "s"
      p.opt :target_host,         "Optionally specify target host. Required when --target-site is used.",
          type: :string,    short: "H"
      p.opt :add_class,           "Add a class to be set in the resulting image.",
          type: :string,    short: "C", multi: true
      p.opt :install,             "Specified if resulting image should be of install type.",
          type: :boolean,   short: :none
      p.opt :update,              "Tells that generated CD should only be used for updates.",
          type: :boolean,   short: :none
      # TODO: After SUSE gets zstd compression, it should be set as default here.
      p.opt :squash_compression,  "Method to use to compress squashfs image. See mksquasfs -comp for available compressions.",
          type: :string,    short: "S",   default: "xz"
      p.opt :hide_var_tmp,        "Don’t include nfsroot’s /var/tmp in created medium. /var/tmp is used for base archives, so this can save some space.",
          type: :boolean,   short: "T",   default: false
      p.opt :force,               "Allow overriding existing image",
          type: :boolean,   short: "f"
    end
  end

  def check_options
    options = @options

    parse_common_options options

    force = options[:force]

    @install_type = options[:install]
    @update_type = options[:update]

    @config = FaiConfig.new @fai_config_dir
    @config.load_nfsroot_conf
    @config.load_fai_conf

    @fai_config_space = options[:fai_config_space]
    @fai_config_space ||= @config.config_space
    unless @fai_config_space
      raise "FAI_CONFIGDIR must be set to a  FAI config space in nfsroot.conf."
    end

    @target_site = options[:target_site]
    @target_host = options[:target_host]

    if @target_site and not @target_host and not @update_type
      error "--target-site without --target-host is only supported for --update media. Fresh installs would not be possible without user intervention."
    end

    if @target_site
      if @target_host
        verb "Looking for host configuration for site #{@target_site}, host #{@target_host}"
        @config.load_target_config @target_site, @target_host
      else
        verb "Creating medium for all hosts at site #{@target_site}"
        @site_config = @config.load_site_config @target_site
      end
    end

    check_target_distribution options[:target_distribution] do |distro|
      case distro
      when "debian"
        # Platform specific checks can be done here.
      when "suse", "opensuse"
        # Platform specific checks can be done here.
      else
        raise "Not supported target distribution: #{distro.inspect}"
      end
    end

    @mirrordir = check_fai_mirror options

    check_grub_config options

    filename = options[:filename]
    @image_name = check_filename filename, force

    raise "Given --initrd file is not valid." if not options[:initrd].nil? and not File.exist? options[:initrd]
    @initrd = options[:initrd]

    @cd_label = options[:media_label]
    @cd_label = "FAI_CD" if @cd_label.nil?

    if @install_type and @update_type
      raise "Only one of --install and --update can be used. Leave both out to create unspecified image."
    end

    @added_classes = options[:add_class]
    if @target_site
      if @config.target&.classes
        @added_classes.concat @config.target.classes
      else
        # Have target specific added classes.
        @site_classes = {}
        @site_config.each do |host, config|
          if config["classes"]
            @site_classes[host] = config["classes"]
          end
        end if @site_config
      end
    end

    @squash_compression = options[:squash_compression]

    @hide_var_tmp = options[:hide_var_tmp]

    set_opts options
  end

  private

  def check_fai_mirror options
    if @target_site
      if options[:mirror_dir].nil?
        fai_mirror = @config.fai_mirror
      end
    end

    fai_mirror = options[:mirror_dir] if fai_mirror.nil?

    raise "--mirror-dir is required, or FAI_MIRROR_DIR needs to be defined at fai.conf." if fai_mirror.nil?
    raise "--mirror-dir (or FAI_MIRROR_DIR) needs to be a valid directory. Given dir: #{fai_mirror}" unless Dir.exist? fai_mirror

    fai_mirror
  end

  def check_grub_config options
    if options[:grub_config]
      error "--grub-config needs to be a valid file." unless File.exist? options[:grub_config]
      @grub_config = options[:grub_config]
      return
    end

    # For update medium, have a specially crafted grub config that launches update process.
    update_medium = Proc.new do
      if options[:update]
        @grub_config_temp = Tempfile.new('create-update-cd-updategrub')
        @grub_config_temp.write generate_update_grub_config
        @grub_config = @grub_config_temp.path
        return
      end
    end

    if not @target_site
      update_medium.call
      error "--grub-config is required."
    end

    if @target_host
      if @config.target.grub_config
        @grub_config = @config.target.grub_config
      else
        update_medium.call
        error "Please specify grub config file in host configurations or with --grub-config."
      end
    else
      unless @site_config["default"].include? "grub_config"
        update_medium.call
        error "When using --target-site without --target-host, grub_config needs to be defined at default config block."
      end
      # TODO: robably this logic can be undone, as this can now fall back to common grub config?
      @grub_config = @config.host_grub_path @site_config["default"]["grub_config"], @site_config["default"]
      if @grub_config.nil?
        error "Grub config file #{@site_config["default"]["grub_config"]} defined in host configurations does not exist. Fix your host configurations."
      end
    end
  end

  def generate_update_grub_config
    content = <<~EOF
    set default="Run update"
    set timeout=4
    set timeout_style=countdown

    # For rare case something is using a real CD.
    insmod iso9660

    # make sure we can access partitions
    insmod part_msdos
    insmod part_gpt
    insmod all_video
    insmod gfxterm

    set color_normal=white/black
    set color_highlight=red/black
    set menu_color_normal=white/black
    set menu_color_highlight=black/yellow

    menuentry "Run update" {
        search --set=root --file /FAI-UPDATE-CD
        linux   /boot/vmlinuz root="LABEL=Root" ip=dhcp systemd.unit=perform-system-update.target
        initrd  /boot/initrd.img
    }
    EOF



  end

  def check_filename filename, force
    if filename.nil?
      if not @target_site
        error "--filename is required when not using --target-site."
      end

      if @target_host
        raise "Invalid filename" if @config.target.image_name.empty?
        filename = "#{@config.target.image_name}-#{Date.today.to_s}.iso"
      else
        filename = "#{@target_site.downcase}-#{Date.today.to_s}.iso"
      end

      if @update_type
        filename = "update-#{filename}"
      end
    end

    if File.exist? filename
      unless force
        raise "--filename needs not to exist before. Remove the file if you wish to overwrite current image."
      else
        FileUtils.rm filename
      end
    end

    filename
  end

  def hidedirs
    @hidedirs
  end

  def liveos_path
    LIVEOS_PATH
  end

  def real_perform
    create_ext3fs_image
    create_squashfs_image

    create_grub2_image

    #provide_memtest_boot_option

    # Not sure this is actually good place for the file, but since some old code may require it, it’s here still.
    add_extra_files @tempdir

    mkiso @cd_label
  end

  def startup_checks
    check_programs

    if @target_distribution == "debian"
      unless File.directory? "#{@mirrordir}/dists"
        error "Given mirror directory does not contain dists/, which a Debian mirror requires. " +
              "Have you generated a mirror to the given path?"
      end
    end

    if @hide_var_tmp
      @hidedirs << "/var/tmp"
    end

    @tempdir = Dir.mktmpdir "create-update-cd-"
    debug "Got tmp: #{@tempdir}"
    @scratch = Dir.mktmpdir "scratch-"

    at_exit do
      cleanup_liveos_mounts

      # Double-remove to be sure they actually gets removed.
      FileUtils.remove_entry @tempdir if Dir.exist? @tempdir
      FileUtils.remove_entry @scratch if Dir.exist? @scratch
    end

    # Don’t show an exception when pressing ctrl+c to interrupt execution.
    trap "SIGINT" do
      #startcolor = "\x1b[38;2;255;255;255m"
      startcolor = color(255, 255, 255, bold: true)
      puts "#{startcolor}Aborting FAI install media creation...#{stopcolor}"
      exit 100
    end

    files = Dir.glob("#{@config.nfsroot}/boot/vmlinuz-*").sort_by do |f|
      File.mtime f
    end
    if files.last.nil?
      debug "nfsroot location: #{@config.nfsroot}, files var content: #{files.inspect}"
      error "Were not able to find initrd from nfsroot. Maybe you’ve passed --no-dracut during nfsroot creation?"
    end
    @kernelversion = files.last.split("-", 2)[1]

    if @target_distribution == "debian"
      faiversion = execvalue("dpkg --root=#{@config.nfsroot} -l fai-client|awk '/fai-client/ {print $3}'")
    elsif ["suse", "opensuse"].include? @target_distribution
      faiversion = execvalue %{rpm --root=#{@config.nfsroot} -q --queryformat="%{VERSION}" fai-client}
    end


    date_str = execvalue("date '+%Y %b %d - %R'")
    @isoversion = sprintf "FAI %-6s build #{date_str}", "#{faiversion}"

    @aname = "Fully Automatic Installation, #{@isoversion}"

    # Hide some kernel modules that are unlikely to be used in an install cd.
    # TODO: The GPU drivers probably should be behind a flag.
    @hidedirs << "/usr/lib/modules/#{@kernelversion}/kernel/drivers/gpu"
    @hidedirs << "/usr/lib/modules/#{@kernelversion}/kernel/drivers/media"
    @hidedirs << "/usr/lib/modules/#{@kernelversion}/kernel/drivers/staging"
    @hidedirs << "/usr/lib/modules/#{@kernelversion}/kernel/sound"
  end
end
