
require "fileutils"
require "json"

require_relative "media"
require_relative "../common"

class Media::Mount
  include Common

  attr_reader :mount_point
  attr_reader :squash_mount_point
  attr_reader :fs_mount_point
  attr_reader :mirror_mount_point

  def initialize mount_root = ""
    @mount_point = "#{mount_root}/mnt/cdrom"
    @squash_mount_point = "#{mount_root}/mnt/squash"
    @fs_mount_point = "#{mount_root}/mnt/filesystem"
    @mirror_mount_point = "#{mount_root}/media/mirror"
  end

  # Mounts a FAI-style CD so that it can be used for updating.
  def mount_cd
    cd_devices = find_fai_cd

    Dir.mkdir @mount_point unless Dir.exist? @mount_point
    Dir.mkdir @squash_mount_point unless Dir.exist? @squash_mount_point
    Dir.mkdir @fs_mount_point unless Dir.exist? @fs_mount_point

    update_cd_found = false
    cd_devices.each do |cd_device|
      dev_path = "/dev/#{cd_device["name"]}"
      execute "mount #{dev_path} #{@mount_point}"
      if File.exist? "#{@mount_point}/LiveOS/squashfs.img"
        update_cd_found = true
        break
      end
    end

    raise "FATAL: None of attached rom devices were an update cd. Please check the media you have inserted." unless update_cd_found
    mount_squashfs
  end

  # Mounts a disk image (e.g. test.iso).
  def mount_cd_file medium
    FileUtils.mkdir_p @mount_point unless Dir.exist? @mount_point
    FileUtils.mkdir_p @squash_mount_point unless Dir.exist? @squash_mount_point
    FileUtils.mkdir_p @fs_mount_point unless Dir.exist? @fs_mount_point
    FileUtils.mkdir_p @mirror_mount_point unless Dir.exist? @mirror_mount_point

    execute "mount #{medium} #{@mount_point}"
    mount_squashfs
  end

  # Unmounts all mounted file systems.
  def unmount fail: false
    # The syncs are said to ensure mounts have done whatever they are meant to be.
    execute %{umount "#{@mirror_mount_point}"}, fail: fail
    execute %{umount "#{@fs_mount_point}"}, fail: fail
    begin
      execute %{umount "#{@squash_mount_point}"}, fail: fail
    rescue Common::ExecutionFailed
      # There is a bug that may cause mount to fail. Let’s sleep and retry to see if there just was something slow.
      sleep 0.2
      execute %{umount "#{@squash_mount_point}"}, fail: fail
    end
    execute %{umount "#{@mount_point}"}, fail: fail
  end

  private

  def mount_squashfs
    execute %{mount "#{@mount_point}/LiveOS/squashfs.img" "#{@squash_mount_point}"}
    execute %{mount "#{@squash_mount_point}/LiveOS/ext3fs.img" "#{@fs_mount_point}"}
    execute %{mount --bind "#{@fs_mount_point}/media/mirror" #{@mirror_mount_point}}
  end

  def find_fai_cd
    json = `lsblk -a -J -o name,size,rm,ro,mountpoint,fstype,type,label`
    data = JSON.parse json

    cd_devices = []

    data["blockdevices"].each do |device|
      if device["type"] == "rom"
        cd_devices << device
      elsif device["label"] == "FAI_CD"
        cd_devices << device
      end
    end

    raise "FATAL: No suitable ROM devices found." if cd_devices.empty?

    cd_devices
  end
end
