
require_relative "media"
require_relative "../common"

module Media::UpdateCdMethods
  include Common

  def check_programs
    msg = ""
    msg+="xorriso not found. Please install package." unless command_available? "xorriso"
    msg+="\nmksquashfs not found. Please install the package squashfs-tools." unless command_available? "mksquashfs"
    msg+="\nmkfs.vfat not found. Please install the package dosfstools." unless command_available? "mkfs.vfat"
    msg+="\nmcopy not found. Please install the package mtools." unless command_available? "mcopy"

    unless msg.empty?
      puts msg
      exit 1
    end
  end

  def error_space status, message
    STDERR.puts "Not enough space left on image for the #{message}"
    STDERR.puts execvalue(%{df -h "#{@tempdir}/#{liveos_path}/mounted-ext3fs"})
    exit status.exitstatus
  end

  def customize_nfsroot
    FileUtils.mkdir_p "#{@tempdir}/empty"

    hidedirs.each do |dir|
      if Dir.exist? "#{@config.nfsroot}/#{dir}"
        debug "hiding #{dir}"
        execute(%{mount --bind "#{@tempdir}/empty" "#{@config.nfsroot}/#{dir}"})
      end
    end
  end

  def unhide_dirs
    hidedirs.each do |dir|
      if Dir.exist? "#{@config.nfsroot}/#{dir}"
        debug "disclosing #{dir}"
        execute(%{umount "#{@config.nfsroot}/#{dir}" 2>/dev/null}, fail: false, log: false)
      end
    end
  end

  # calculates the amount of space needed for the ext3fsimage
  def calculate_required_size
    if @target_distribution == "debian"
      size = execvalue(%[du -scLl "#{@config.nfsroot}" "#{@fai_config_space}" "#{@mirrordir}"/{dists,pool} 2> /dev/null | tail -n1 | awk '{print $1}']).to_i
    elsif ["suse", "opensuse"].include? @target_distribution
      size = execvalue(%[du -scLl "#{@config.nfsroot}" "#{@fai_config_space}" "#{@mirrordir}" 2> /dev/null | tail -n1 | awk '{print $1}']).to_i
    else
      raise "BUG: Unknown target distribution here #{@target_distribution}"
    end

    # not dividing by 1024 to get the exact, this allows us to get some additional space
    calculated_size = size / 850
    debug("Calculated size: #{calculated_size} MiB")
    calculated_size
  end

  def addmirror
    FileUtils.mkdir_p "#{@tempdir}/#{liveos_path}/mounted-ext3fs/media/mirror"

    verb "Copying the mirror to CD image"
    if @target_distribution == "debian"
      execute(%{cp -Pr "#{@mirrordir}"/{dists,pool} "#{@tempdir}/#{liveos_path}/mounted-ext3fs/media/mirror"})
    elsif ["suse", "opensuse"].include? @target_distribution
      execute(%{cp -Pr "#{@mirrordir}" "#{@tempdir}/#{liveos_path}/mounted-ext3fs/media/mirror"})
    end

    error_space $?, "mirror" unless exec_success?

    if @target_distribution == "debian"
      # create the sources.list for the CD
      execute <<~MULTILINE
      cat > "#{@tempdir}/#{liveos_path}/mounted-ext3fs/etc/apt/sources.list" <<EOF
      # mirror location for fai CD, file generated by fai-cd
      EOF
      MULTILINE

      dists = execvalue(%{find "#{@mirrordir}" -name "Packages*" | grep binary | sed 's/binary-.*//' | sed "s##{@mirrordir}/*dists/##" | xargs -r -n 1 dirname | sort | uniq})
      raise "No suitable Packages file found in mirror." if dists.empty?

      dists = dists.split " "
      dists.each do |dist|
        comp = execvalue(%{find "#{@mirrordir}/dists/#{dist}" -maxdepth 2 -type d -name "binary-*" | sed -e "s##{@mirrordir}/*dists/#{dist}/##" -e 's#/binary-.*##' | uniq | tr '\n' " "})
        execute(%{echo "deb [trusted=yes] file:/media/mirror #{dist} #{comp}" >> "#{@tempdir}/squashfs-root/LiveOS/mounted-ext3fs/etc/apt/sources.list"})
      end
    elsif ["suse", "opensuse"].include? @target_distribution
      # Difference between rpm-md and yast2: https://lists.opensuse.org/opensuse/2018-01/msg01097.html
      # TODO: maybe certificate installation could be handled somehow so it can be used gpgcheck=1 here? Not a big thing though, as the mirror ought to be curated.
      File.open "#{@tempdir}/#{liveos_path}/mounted-ext3fs/etc/zypp/repos.d/fai_cd.repo", "w" do |f|
        out = <<~EOF
        [FAI_CD_REPO]
        name=Repository in FAI CD
        enabled=1
        autorefresh=1
        baseurl=file:///media/mirror
        type=plaindir
        gpgcheck=0
        EOF
        f.write out
      end
    end
  end

  def create_ext3fs_image
    ext3fs_path = "#{@tempdir}/#{liveos_path}/mounted-ext3fs"

    # create the folder that will be used for the squashfs.img later
    # along with it, a temporary folder is created to mount the ext3fs.img
    FileUtils.mkdir_p ext3fs_path

    # hide some directories before copying the nfsroot
    customize_nfsroot
    count = calculate_required_size

    # creates an empty file
    verb "Creating empty ext3fs.img"
    execute(%{dd if=/dev/zero of="#{@tempdir}/#{liveos_path}/ext3fs.img" bs=1MB count="#{count}"})
    raise "dd failed. Maybe no enough free space in #{@tempdir}" unless exec_success?

    execute(%{mkfs.ext4 -m1 -q -O^has_journal -F -L FAI-NFSROOT "#{@tempdir}/#{liveos_path}/ext3fs.img"})

    # the resulting ext3fs.img is mounted
    execute(%{mount -o loop,noatime "#{@tempdir}/#{liveos_path}/ext3fs.img" "#{ext3fs_path}"})

    # Recalculate module information, as some module dirs may be hidden by @hidedirs.
    execute %{depmod --basedir="#{@config.nfsroot}" "#{@kernelversion}"}

    verb "Copying the nfsroot to CD image"
    debug "nfsroot size: #{execvalue %[du -sh0 '#{@config.nfsroot}']}"
    # copy the contents of the nfsroot into it
    execute(%{cp -Ppr "#{@config.nfsroot}/." "#{ext3fs_path}/"})
    error_space $?, "nfsroot" unless exec_success?
    FileUtils.rm_f "#{ext3fs_path}/etc/resolv.conf"

    # execute("find #{ext3fs_path}") # for debugging

    # copy config space into nfsroot-copy unless -d is given
    verb "Copying the config space to CD image"
    fs_fai_config_space = "#{ext3fs_path}/var/lib/fai/config/"
    FileUtils.mkdir_p fs_fai_config_space unless Dir.exist? fs_fai_config_space
    execute %{cp -Ppr "#{@fai_config_space}"/* "#{fs_fai_config_space}"}
    error_space $?, "config directory" unless exec_success?

    # add the partial mirror to the CD
    addmirror

    add_extra_files "#{ext3fs_path}/var/tmp"

    unhide_dirs

    # Revert to original module configuration.
    execute %{depmod --basedir="#{@config.nfsroot}" "#{@kernelversion}"}

    execute(%{umount "#{ext3fs_path}"})
    Dir.rmdir ext3fs_path
  end

  def create_squashfs_image

    # this is where the squashfs.img needs to be placed
    FileUtils.mkdir_p "#{@tempdir}/LiveOS"

    # create the squashfs.img
    output_path = "#{@tempdir}/LiveOS/squashfs.img"
    sqopt = %{-comp "#{@squash_compression}"}
    execute(%{mksquashfs #{@tempdir}/squashfs-root "#{output_path}" #{sqopt} -progress}, direct_output: true)

    File.delete "#{@tempdir}/#{liveos_path}/ext3fs.img"
    Dir.rmdir "#{@tempdir}/#{liveos_path}"
    Dir.rmdir "#{@tempdir}/squashfs-root"
  end

  def create_grub2_image
    FileUtils.mkdir_p "#{@tempdir}/boot"

    if @target_distribution == "debian"
      unless Dir.exist? "#{@config.nfsroot}/usr/lib/grub/x86_64-efi"
        raise "No grub-efi-amd64-bin installation found in NFSROOT. Aborting."
      end
    elsif ["suse", "opensuse"].include? @target_distribution
      unless Dir.exist? "#{@config.nfsroot}/usr/share/grub2/x86_64-efi"
        raise "No grub2-x86_64-efi installation found in NFSROOT. Aborting."
      end
    end

    nfsroot_grub_cfg = "#{@config.nfsroot}/tmp/grub.cfg"
    FileUtils.cp @grub_config, nfsroot_grub_cfg

    process_line = lambda do |line, r|
      return line unless @target_host
      return line if not line.match? /^\s*linux /
      return line if line.match? r

      if @config.target&.hostname.nil?
        verb "Used grub.cfg does not have hostname specified, and it is not set in host configurations either. Not defining hostname in grub.cfg."
        line
      else
        %{#{line.rstrip} hostname="#{@config.target.hostname}"}
      end
    end

    # Constructing regex from string to avoid problem in syntax highlighting.
    process_line_regex = Regexp.new(" hostname=[^\s]+")

    # Modify grub.cfg to contain hostname if one is specified and grub.cfg does not define one.
    lines = File.readlines(nfsroot_grub_cfg).map do |line|
      process_line.call line, process_line_regex
    end
    File.open nfsroot_grub_cfg, "w" do |f|
      f.puts lines
    end

    # insert date into grub menu
    execute(%{perl -pi -e "s/_VERSIONSTRING_/   #{@isoversion}     /" #{nfsroot_grub_cfg}})

    if @target_distribution == "debian"
      grub_mkstandalone = "grub-mkstandalone"
    elsif ["suse", "opensuse"].include? @target_distribution
      grub_mkstandalone = "grub2-mkstandalone"
    end

    execute(%{chroot "#{@config.nfsroot}" #{grub_mkstandalone} \
        --format=x86_64-efi \
        --output=/tmp/bootx64.efi \
        --locales="" --fonts="" \
        "boot/grub/grub.cfg=/tmp/grub.cfg"})
    FileUtils.move "#{@config.nfsroot}/tmp/bootx64.efi", @scratch

    verb "Creating efiboot.img"
    execute(%{dd if=/dev/zero of="#{@scratch}/efiboot.img" bs=1M count=3})
    execute(%{mkfs.vfat "#{@scratch}/efiboot.img"}, log: false)
    execute(%{mmd -i "#{@scratch}/efiboot.img" efi efi/boot})
    execute(%{mcopy -i "#{@scratch}/efiboot.img" "#{@scratch}/bootx64.efi" ::efi/boot/})

    if @target_distribution == "debian"
      @grub_pc_path = "/usr/lib/grub/i386-pc"
    elsif ["suse", "opensuse"].include? @target_distribution
      @grub_pc_path = "/usr/share/grub2/i386-pc"
    else
      raise "Using not implemented distribution #{@target_distribution}."
    end

    unless File.exist? "#{@config.nfsroot}#{@grub_pc_path}"
      raise "No grub-pc installation found in NFSROOT. Aborting."
    end

    execute(%{chroot "#{@config.nfsroot}" #{grub_mkstandalone} \
        --format=i386-pc \
        --output=/tmp/core.img \
        --locales="" --fonts="" \
        --install-modules="cat linux normal iso9660 biosdisk memdisk search ls echo test chain msdospart part_msdos part_gpt minicmd ext2 keystatus all_video font sleep gfxterm" \
        --modules="linux normal iso9660 biosdisk search" \
        "boot/grub/grub.cfg=/tmp/grub.cfg"})

    execute(%{cat "#{@config.nfsroot}#{@grub_pc_path}/cdboot.img" "#{@config.nfsroot}/tmp/core.img" > "#{@scratch}/bios.img"})
    FileUtils.rm "#{@config.nfsroot}/tmp/core.img"

    FileUtils.cp "#{@config.nfsroot}/boot/vmlinuz-#{@kernelversion}", "#{@tempdir}/boot/vmlinuz", preserve: true

    if @target_distribution == "debian"
      initrd_base = "initrd.img"
    elsif ["suse", "opensuse"].include? @target_distribution
      initrd_base = "initrd"
    end

    # If no specific initrd has been given, default to nfsroot’s initrd.
    @initrd ||= "#{@config.nfsroot}/boot/#{initrd_base}-#{@kernelversion}"
    FileUtils.cp @initrd, "#{@tempdir}/boot/#{initrd_base}", preserve: true
    FileUtils.cp "#{@config.nfsroot}/boot/config-#{@kernelversion}", "#{@tempdir}/boot/", preserve: true

    filelabel = "FAI-CD"
    filelabel = "FAI-INSTALL-CD" if @install_type
    filelabel = "FAI-UPDATE-CD" if @update_type
    FileUtils.touch "#{@tempdir}/#{filelabel}"
  end

  # Adds extra files that are not normally present.
  def add_extra_files directory
    # If some new classes should be added to the host, there is limited ways to perform it.
    # This method works with some initrd trickery, which adds these extra classes as kernel
    # cmdline parameters.
    if @added_classes
      verb "Adding following extra classes: #{@added_classes.join ", "}"
      File.open "#{directory}/new_classes", "w" do |f|
        @added_classes.each do |c|
          f.puts c
        end
      end
    end

    return unless @site_classes

    @site_classes.each do |host, classes|
      verb "Writing extra classes for #{host}: #{classes.join ", "}"

      Dir.mkdir "#{directory}/host_classes" unless Dir.exist? "#{directory}/host_classes"

      @config.load_target_config @target_site, host

      File.open "#{directory}/host_classes/#{@config.target.hostname}", "w" do |f|
        classes.each do |c|
          f.puts c
        end
      end
    end

    @config.unload_target_config
  end

  def mkiso cd_label
    raise "@grub_pc_path is not set." unless @grub_pc_path

    # NOTE: I currently have too old xorriso to have this argument.
    # -iso_mbr_part_type 00 \
    verb "Writing FAI CD-ROM image to #{@image_name}. This may need some time."
    execute(%{xorriso -report_about HINT -as mkisofs -iso-level 3 \
        -full-iso9660-filenames \
        -volid "#{cd_label}" -appid "#{@aname}" \
        -eltorito-boot boot/grub/bios.img  \
        -no-emul-boot -boot-load-size 4 -boot-info-table \
        --eltorito-catalog boot/grub/boot.cat \
        --grub2-boot-info \
        --grub2-mbr "#{@config.nfsroot}#{@grub_pc_path}/boot_hybrid.img" \
        -eltorito-alt-boot -e EFI/efiboot.img -no-emul-boot \
        -append_partition 2 0xef "#{@scratch}/efiboot.img" \
        -o "#{@image_name}" -graft-points \
        --sort-weight 0 / --sort-weight 1 /boot \
        "#{@tempdir}" \
        /boot/grub/bios.img="#{@scratch}/bios.img" \
        /EFI/efiboot.img="#{@scratch}/efiboot.img"}, log: false)

    temp = execvalue %{du -h "#{@image_name}"}
    puts "ISO image size and filename: #{temp}"

    unhide_dirs
  end

  def cleanup_liveos_mounts
    #puts "DEBUG"
    #sleep 10000

    execute(%{mountpoint -q "#{@tempdir}/#{liveos_path}/mounted-ext3fs" && umount "#{@tempdir}/#{liveos_path}/mounted-ext3fs"}, fail: false, log: false)
    unhide_dirs
    FileUtils.rm_rf [ @tempdir, @scratch ], secure: true
  end

end
