
require "fileutils"

require_relative "../options"
require_relative "../fai-config"
require_relative "../nfsroot/ft_nfsroot"

require_relative "mirror"

# Creates a Debian nfsroot and runs reprepro with given configuration.
class Mirror::FaiRepreproMirror
  include Options
  include Nfsroot::FtNfsroot

  def options parser
    parser.opt :reprepro_config_dir,        "Path to directory containing conf/distributions and other configuration files.",
        short: "r",   type: :string, required: true
    parser.opt :nfsroot_fai_config_dir,     "FAI config dir that is used to generate nfsroot for reprepro run.",
        short: "n",   type: :string, required: true
    parser.opt :recreate_nfsroot,           "Remove existing nfsroot and recreate it. Use with care.",
        short: :none,
    parser.opt :recreate_mirror,            "Create a new mirror instead of updating existing mirror.",
        short: "-m"
  end

  def check_options
    unless Dir.exist? "#{@options[:reprepro_config_dir]}/conf"
      error "Given --reprepro-config-dir does not contain conf dir."
    end
    @reprepro_config_dir = @options[:reprepro_config_dir]

    unless Dir.exist? "#{@options[:nfsroot_fai_config_dir]}"
      error "Given --nfsroot-fai-config-dir is not valid."
    end
    @nfsroot_fai_config_dir = @options[:nfsroot_fai_config_dir]

    @config = FaiConfig.new @nfsroot_fai_config_dir
    @config.load_nfsroot_conf

    set_nfsroot @config.nfsroot

    @recreate_nfsroot = @options[":recreate_nfsroot"]
    @recreate_mirror = @options[":recreate_mirror"]
  end

  def perform
    startup_checks
    real_perform
  end

  private

  def startup_checks
    if @recreate_nfsroot
      verb "Removing existing mirror nfsroot at #{@config.nfsroot}."
      FileUtils.rm_r @config.nfsroot
    end

    @mirror_name = File.basename @reprepro_config_dir
    @target_reprepro_config_dir = "/srv/fai/reprepro/#{@mirror_name}"

    if @recreate_mirror
      verb "Removing existing reprepro mirror at #{nfsroot}#{@target_reprepro_config_dir}."

      FileUtils.rm_r "#{nfsroot}#{@target_reprepro_config_dir}"
    end
  end

  def real_perform
    unless Dir.exist? @config.nfsroot
      run_nfsroot @nfsroot_fai_config_dir, dracut: false
    else
      run_nfsroot @nfsroot_fai_config_dir, dracut: false, refresh: true
    end

    FileUtils.mkdir_p "#{nfsroot}#{@target_reprepro_config_dir}"
    FileUtils.cp_r Dir.glob("#{@reprepro_config_dir}/*"), "#{nfsroot}#{@target_reprepro_config_dir}"

    #start_chroot do
    #  execute "cd #{@target_reprepro_config_dir}"
    #  execute "reprepro update"
    #end

    verboseflag = @@_verbose && "-v" || ""
    verboseflag = "-V" if @@_debug

    showpercent = @@_verbose && "--show-percent" || ""

    execute "#{rootcmd} reprepro #{verboseflag} #{showpercent} -b #{@target_reprepro_config_dir} update"
  end
end
