
require "fileutils"
require "optimist"

require_relative "../options"
require_relative "../fai-config"
require_relative "methods"
require_relative "scripts"
require_relative "nfsroot-debian"
require_relative "nfsroot-opensuse"
require_relative "nfsroot-suse"

# Main class.
class Nfsroot::CreateNfsroot
  include Options
  include Nfsroot::Methods
  include Nfsroot::Scripts

  def options parser
    options = common_options(parser: parser, log_file: "create-nfsroot.log") do |p|
      p.opt :fai_config_dir,    "FAI’s main config files.",
          type: :string,      default: "/etc/fai",    short: "c"
      p.opt :fai_config_space,  "Use specified config space instead of one from FAi config dir.",
          type: :string,                              short: :none
      p.opt :all_packages,      "Installs full set of packages to nfsroot.",
                              default: true,          short: "a"
      p.opt :refresh,           "Installs new packages from NFSROOT and updates all packages.",
                                                      short: "r"
      p.opt :force,             "Removes existing nfsroot before installing a new one.",
                              default: false,         short: "f"
      p.opt :base_archive,      "Use existing base archive instead of creating a new one. Base can also be set with an environment variable $NFSROOT_BASE",
          type: :string,                              short: "b"
      p.opt :nonfree,           "Also install packages with NONFREE class in NFSROOT file.",
                              default: true,          short: "N"
      p.opt :run_scripts,       "Runs scripts defined in FAI config space’s (determined from nfsroot.conf) nfsroot/scripts directory. Defaults to true.",
                              default: true,          short: :none
      p.opt :no_dracut,         "Do not run dracut related actions. This does not mean that dracut gets or does not get installed.",
          type: :boolean,     default: false,         short: :none
      p.opt :base_compress_algo, "Algorithm used to compress base archive. Currently only implemented for suse.",
          type: :string,      default: nil,           short: :none
    end
  end

  def check_options
    options = @options

    parse_common_options options

    @nonfree = options[:nonfree] ? "NONFREE" : ""
    @all_packages = options[:all_packages]
    @allow_run_scripts = options[:run_scripts]
    @base_compress_algo = options[:base_compress_algo]

    @base_archive = options[:base_archive]
    @base_archive = ENV["BASE_ARCHIVE"] if @base_archive.nil?
    if not @base_archive.nil? and not File.exist? @base_archive
      raise "Given #{@base_archive} does not exist."
    end

    verb "Using configuration files from #{@fai_config_dir}"

    @config = FaiConfig.new @fai_config_dir
    @config.load_fai_conf
    @config.load_nfsroot_conf
    set_nfsroot @config.nfsroot

    if options[:fai_config_space]
      @config.config_space = options[:fai_config_space]
    end
    ENV["FAI"] = @config.config_space

    raise "NFSROOT needs to be defined in nfsroot.conf" unless @config.nfsroot

    target_distribution = @config.nfsroot_distribution
    # Default to Debian to work better with old code.
    # TODO: This should be deprecated by some major version in future.
    target_distribution ||= "debian"
    case target_distribution
    when "debian"
      @impl = Nfsroot::NfsrootDebian.new options, @config
      raise "#{@fai_config_dir}/apt does not exist." unless Dir.exist? "#{@fai_config_dir}/apt"
    when "opensuse"
      @impl = Nfsroot::NfsrootOpensuse.new options, @config
      raise "#{@fai_config_dir}/zypper does not exist." unless Dir.exist? "#{@fai_config_dir}/zypper"
    when "suse"
      @impl = Nfsroot::NfsrootSuse.new options, @config
      raise "#{@fai_config_dir}/zypper does not exist." unless Dir.exist? "#{@fai_config_dir}/zypper"
    else
      error "Target distribution #{target_distribution} from nfsroot.conf is not supported."
    end

    set_opts options

    @impl.ft_output_flags = ft_output_flags
    @impl.base_compress_algo = @base_compress_algo
  end

  def perform
    startup_checks

    at_exit do
      umount_dirs
    end

    return perform_refresh if opts[:refresh]

    perform_fresh_install
  end

  private

  def startup_checks
    @deldir = @config.nfsroot.dup

    @impl.fai_config_dir = @fai_config_dir
    @impl.all_packages = @all_packages
    @impl.nonfree = @nonfree
    @impl.base_archive = @base_archive

    # Ensure scripts won’t be run if they don’t exist.
    if @allow_run_scripts and !Dir.exist?("#{@config.config_space}/nfsroot/scripts")
      debug "nfsroot scripts does not exist, so not running them."
      @allow_run_scripts = false
    end
    @impl.allow_run_scripts = @allow_run_scripts

    @impl.startup_checks
  end

  # Updates existing nfsroot.
  def perform_refresh
    verb "Upgrading nfsroot and installing new packages into the nfsroot."

    @impl.nfsroot_configuration
    @impl.perform_refresh
    @impl.copy_package_manager_configuration
    @impl.add_packages_to_nfsroot
    set_root_pw
    post_install_configuration

    # Not implemented feature.
    #setup_tftp

    run_scripts
  end

  # Installs a completely new nfsroot.
  def perform_fresh_install
    verb "Using configuration files from #{@fai_config_dir}"

    begin
      initialize_nfsroot_dir
    rescue ExecutionFailed
      # There is some bug that requires a cycle. Let’s just emulate that to see if it helps.
      umount_dirs
      initialize_nfsroot_dir
    end

    if opts[:force]
      FileUtils.rm_rf @config.nfsroot
    end

    # These are from Methods.
    @impl.solve_repository_info
    create_nfsroot
    @impl.copy_package_manager_configuration
    add_etc_hosts_entries
    @impl.handle_nfsroot_packages
    set_root_pw
    set_timezone
    post_install_configuration
    #setup_tftp
    setup_ssh
    run_scripts
    check_nfsroot
    write_variables

    if @generic_nfsroot
      # remove fai.conf which comes from package fai-client
      FileUtils.rm_f "#{@config.nfsroot}/etc/hosts"
      FileUtils.rm_f Dir.glob("#{@config.nfsroot}/etc/resolv.conf*")
      FileUtils.rm_f "#{@config.nfsroot}/etc/fai/fai.conf"
      FileUtils.rm_f Dir.glob("#{@config.nfsroot}/var/lib/apt/lists/*_*")

      execute %{ainsl -as "#{@config.nfsroot}/etc/hosts" "127.0.0.1 localhost"}
    end

    umount_dirs

    Process.waitall

    verb "fai-make-nfsroot finished."
    opts[:verbose] and puts "Log file written to /var/log/fai/fai-make-nfsroot.log"
  end
end