
require_relative "nfsroot"
require_relative "../common"
require_relative "../base/nfsroot-base-methods"

module Nfsroot::Methods
  include Common

  include FaiToolkit::Base::NfsrootBaseMethods

  def initialize_nfsroot_dir
    if Dir.exist? @config.nfsroot and !opts[:force]
      STDERR.puts "-----------------------------------------------------------------"
      STDERR.puts "Error: #{@config.nfsroot} exists already."
      STDERR.puts "Neither force nor update option present, exiting."
      STDERR.puts "-----------------------------------------------------------------"
      STDERR.puts "NOTE: Use --force to force overwriting an existing nfsroot."
      STDERR.puts "      Use --refresh to update/install packages defined in NFSROOT config."
      exit 1
    end

    verb "Creating FAI nfsroot in #{@config.nfsroot}"

    if Dir.exist? "#{@config.nfsroot}/proc"
      # Since the umount seems to be unfinished in some cases, let’s do it here too...
      umount_dirs

      execute %{umount "#{@config.nfsroot}/dev/pts" >&/dev/null}, fail: false
      execute %{umount "#{@config.nfsroot}/proc"}, fail: false if File.symlink? "#{@config.nfsroot}/proc/self"
      raise "/proc is still mounted inside the nfsroot." if File.symlink? "#{@config.nfsroot}/proc/self"
      execute %{umount "#{@config.nfsroot}/#{@config.mirror_mountpoint}" 2>/dev/null}, fail: false # it's safer to try to umount
      # remove old nfsroot using a background process
      execute %{rm -rf "#{@deldir}/../.will-now-be-deleted"}, fail: false
      execute %{mv "#{@deldir}" "#{@deldir}/../.will-now-be-deleted"}, fail: false
      unless exec_success?
        FileUtils.rm_rf @deldir
      else
        execute %{nice rm -rf "#{@deldir}/../.will-now-be-deleted"}, wait: false
      end
    end
  end

  def add_etc_hosts_entries
    verbose = @@_verbose && "-v" || ""

    # some default entries
    execute %{ainsl -as "#{@config.nfsroot}/etc/hosts" "127.0.0.1 localhost"}
    execute %{ainsl     "#{@config.nfsroot}/etc/hosts" "# Following entries are specific to your environment"}
    execute %{ainsl #{verbose} "#{@config.nfsroot}/etc/hosts" "#{@config.nfsroot_etc_hosts}"}

    # add entries for all network devices
    ips = execvalue("ip addr show up| grep -P -o '(?<=inet )[^/]+' | grep -v 127.0.0.1").split("\n")

    ips.each do |eth|
      line = execvalue "getent hosts #{eth}", fail: false
      if line.empty?
        warn "no hostname for #{eth} found, not adding to /etc/hosts."
        next
      end
      execute %{ainsl #{verbose} "#{@config.nfsroot}/etc/hosts" "#{line}"}
    end

    if File.exist? "/etc/resolv.conf"
      execute %{cp -Lp #{verbose} /etc/resolv.conf "#{@config.nfsroot}/etc/resolv.conf-installserver"}

      # this is needed during ft-nfsroot
      execute %{cp -Lp #{verbose} /etc/resolv.conf "#{@config.nfsroot}/etc/resolv.conf"}
    end
  end

  def set_root_pw
    # NOTE: @generic_nfsroot is not implemented for now.
    return if @generic_nfsroot # do nothing here if we generate a generic nfsroot

    # set root pw inside the nfsroot
    execute %{#{rootcmd} usermod -p "#{@config.fai_rootpw}" root}
  end

  def create_nfsroot
    FileUtils.mkdir_p "#{@config.nfsroot}"
    Dir.chdir @config.nfsroot

    @impl.create_nfsroot
    @impl.nfsroot_configuration

    FileUtils.mkdir_p "#{@config.nfsroot}/#{@config.config_space}"

    unless opts[:no_dracut]
      fslist = ""

      File.open "#{@config.nfsroot}/etc/dracut.conf.d/13-add-cd-support.conf", "w" do |f|
        if @all_packages
          f.puts 'add_dracutmodules+=" livenet "'
        else
          f.puts 'add_dracutmodules+=" dmsquash-live "'
        end

        fslist = "lockd ext4 xfs"

        # Always bundle iso9660, so reading from CD can work even if booted kernel
        # is different from the kernel inside installed system.
        # This is rarely used, but it only increases size of the initrd by ~80 kB, so
        # there is little reason not to include it.
        f.puts 'add_drivers+=" isofs "'
      end

      File.open "#{@config.nfsroot}/etc/dracut.conf.d/20-generic-image.conf", "w" do |f|
        f.puts 'hostonly="no"'
      end

      # We do not need all those filesystems inside the initrd.
      File.open "#{@config.nfsroot}/etc/dracut.conf.d/02-filesystems.conf", "w" do |f|
        f.puts %{filesystems="#{fslist}"}
      end

      compress_file = "#{@config.nfsroot}/etc/dracut.conf.d/80-compress.conf"
      if @config.dracut_compress
        File.open compress_file, "w" do |f|
          f.puts %{compress="#{@config.dracut_compress}"}
        end
      else
        FileUtils.rm compress_file if File.exist? compress_file
      end
    end
  end

  # set timezone in nfsroot
  def set_timezone
    timezonefile = "/etc/timezone"
    if File.exist? timezonefile
      execute %{cp -fH "#{timezonefile}" "#{@config.nfsroot}/etc/timezone"}
    else
      File.open timezonefile, "w" do |file|
        timezone = execvalue %{timedatectl status | grep "Time zone" | awk '{ print $3 }'}
        file.write timezone
      end
    end
    execute %{cp -a /etc/localtime "#{@config.nfsroot}/etc/localtime"}
  end

  def post_install_configuration
    FileUtils.ln_s "/proc/mounts", "#{@config.nfsroot}/etc/mtab" unless File.symlink? "#{@config.nfsroot}/etc/mtab"

    # Do not activate RAID arrays.
    if Dir.exist? "#{@config.nfsroot}/etc/mdadm"
      File.open "#{@config.nfsroot}/etc/mdadm/mdadm.conf", "w" do |f|
        f.puts "AUTO -all"
      end
    end

    if !File.exist? "#{@config.nfsroot}/root/.profile" or !File.readlines("#{@config.nfsroot}/root/.profile").grep(%r{/usr/lib/fai/subroutines}).any?
      File.open "#{@config.nfsroot}/root/.profile", "a" do |f|
        f.puts <<~EOF
          PATH=/usr/local/sbin:/usr/local/bin:/usr/lib/fai:/bin:/sbin:/usr/bin:/usr/sbin:
          export PATH
          . /usr/lib/fai/subroutines
          set -a
          . /tmp/fai/variables.log 2>/dev/null
        EOF
      end
    end

    @impl.post_install_configuration
  end

  def setup_ssh
    # nothing to do if no ssh is available in nfsroot
    return unless File.exist? "#{@config.nfsroot}/usr/bin/ssh"

    # enable root login
    execute %{sed -i -e '/PermitRootLogin/d' "#{@config.nfsroot}/etc/ssh/sshd_config"}
    execute %{ainsl "#{@config.nfsroot}/etc/ssh/sshd_config" 'PermitRootLogin yes'}

    FileUtils.mkdir_p "#{@config.nfsroot}/root/.ssh", mode: 700

    # TODO: better than env var?
    return unless ENV["LOGUSER"]

    loguserhome = execvalue %{cd ~#{ENV["LOGUSER"]} 2>/dev/null && pwd;true}
    return unless File.directory? "#{loguserhome}/.ssh"

    if File.file? "#{loguserhome}/.ssh/known_hosts"
      FileUtils.cp "#{loguserhome}/.ssh/known_hosts", "#{@config.nfsroot}/root/.ssh/known_hosts"
    end
    if File.file? "#{loguserhome}/.ssh/id_ed25519"
      FileUtils.cp Dir.glob("#{loguserhome}/.ssh/id_ed25519*"), "#{@config.nfsroot}/root/.ssh/", preserve: true
    end
    if File.file? "#{loguserhome}/.ssh/id_rsa"
      FileUtils.cp Dir.glob("#{loguserhome}/.ssh/id_rsa*"), "#{@config.nfsroot}/root/.ssh/", preserve: true
    end
  end

  # Simple tests to see if important things are available inside the nfsroot.
  def check_nfsroot
    if !opts[:no_dracut] and !File.executable? "#{@config.nfsroot}/usr/bin/dracut"
        raise "dracut was not installed into the nfsroot."
    end

    unless File.exist? "#{@config.nfsroot}/sbin/init"
        raise "/sbin/init is not present inside nfsroot. Please install a stub init. For Debian, that is sysvinit-core."
    end

    @impl.check_nfsroot
    @impl.check_required_packages

    unless opts[:no_dracut]
      files = Dir.glob "#{@config.nfsroot}/boot/initrd*"
      raise "No initrd installed." if files.empty?
      if opts[:logging]
        execute %{grep -E -q "^ERROR: |^E: Sub-process |^dpkg: error processing |^dpkg: dependency problems" "#{@log_file.path}"}, fail: false
        bad_exit if exec_success?
      end
    end

    nil
  end

  def write_variables
    varfile = "/var/log/fai/variables"
    fai_version = @impl.fai_version

    # remove old values from file
    execute %{sed -i -e '/FAI_VERSION/d' -e '/NFSROOT/d' "#{varfile}"} if File.exist? varfile and !File.zero? varfile
    execute %{echo "FAI_VERSION=#{fai_version}" >> "#{varfile}"}
    execute %{echo "NFSROOT=#{@config.nfsroot}" >> "#{varfile}"}
  end

  private

  def extract_base base_archive
    unless File.exist? base_archive
      raise "Base archive #{base_archive} could not be used."
    end

    Dir.chdir nfsroot do
      verb "Extracting base archive #{base_archive}: "

      case File.basename(base_archive).split(".", 2)[1]
      when "tar.gz", ".tgz"
        execute %{tar --xattrs --acls --xattrs-include=* -C "#{@config.nfsroot}" -zxpf "#{base_archive}"}
        FileUtils.cp base_archive, "#{@config.nfsroot}/var/tmp/base.tar.gz", preserve: true
      when "tar.xz", "txz"
        execute %{tar --xattrs --acls --xattrs-include=* -C "#{@config.nfsroot}" -Jxpf "#{base_archive}"}
        FileUtils.cp base_archive, "#{@config.nfsroot}/var/tmp/base.tar.xz", preserve: true
      end
    end
  end

  def bad_exit
    puts "An error occured during ft nfsroot."
    puts "Please fix the error or try ft nfsroot -v -d"
    exit 1
  end
end