require_relative "nfsroot-impl"
require_relative "../base/nfsroot-base-methods"

class Nfsroot::NfsrootDebian < Nfsroot::NfsrootImpl
  include FaiToolkit::Base::NfsrootBaseMethods

  def startup_checks
    # Required so that apt-get won’t try to run debconf.
    ENV["DEBIAN_FRONTEND"] = "noninteractive"
  end

  # Dummy implementation since nothing is needed here.
  def nfsroot_configuration
    # Nya.
  end

  def perform_refresh
    execute "#{rootcmd} apt-get update"
    execute "#{rootcmd} apt-get -y dist-upgrade"
  end

  def fai_version
    @fai_version ||= execvalue %{#{rootcmd} dpkg-query -W -f='${Version}' fai-client}
  end

  def copy_package_manager_configuration
    if Dir.exist? "#{@fai_config_dir}/apt/sources.list"
      # Use fcopy and classes to copy apt configuration files to the nfsroot.
      execute %{fcopy -s "#{@fai_config_dir}/apt" -t "#{@config.nfsroot}/etc/apt" sources.list}
    else
      # Copy all apt configuration files to the nfsroot.
      execute %{cp -aL "#{@fai_config_dir}/apt" "#{@config.nfsroot}/etc"}
    end
  end

  def add_packages_to_nfsroot
    ENV["FAI_ROOT"] = @config.nfsroot

    # Run hooks in pre-packages dir.
    run_scripts hook: "pre-packages"

    setclasses
    execute %{#{rootcmd} apt-cache dumpavail > "#{@config.nfsroot}/var/lib/dpkg/available"}, log: false
    execute %{env LC_ALL="C" install_packages -l "-p#{@fai_config_dir}" > "#{@config.nfsroot}/var/tmp/packages.nfsroot"}, log: false

    verb "Adding additional packages to #{@config.nfsroot}:"
    verb File.read("#{@config.nfsroot}/var/tmp/packages.nfsroot") if @@_verbose
    output = execute %{env LC_ALL="C" install_packages -v "-p#{@fai_config_dir}"}

    warnings = output.scan(/^WARNING: (.*)$/)
    warnings = warnings.first
    if warnings
      STDERR.puts
      STDERR.puts "When installing packages, following errors happened:"
      warnings.each do |warning|
        STDERR.puts"  - #{warning}"
      end
      STDERR.puts
      STDERR.puts "Please fix above errors and try again."
      exit 1
    end

    FileUtils.rm "#{@config.nfsroot}/var/lib/dpkg/available"
  end

  def solve_repository_info
    # resolve_sources is from nfsroot-base-methods.rb
    resolve_sources
  end

  # Called by Methods#create_nfsroot.
  def create_nfsroot
    if @base_archive
      extract_base @base_archive
    else
      create_base
    end

    FileUtils.touch ".THIS_IS_THE_FAI_NFSROOT"
    # save the list of all packages in the base.tar.gz
    execute %{#{rootcmd} dpkg --get-selections | awk '/install$/ {print $1}' > "#{@config.nfsroot}/var/tmp/base-pkgs.lis"}

    if !@config.fai_debmirror.nil? and !@config.fai_debmirror.empty?
      verb "Mounting #{@config.fai_debmirror} to #{@config.nfsroot}/#{@config.mirror_mountpoint}."

      FileUtils.mkdir_p "#{@config.nfsroot}/#{@config.mirror_mountpoint}"
      execute %{mount -o ro,noatime "#{@config.fai_debmirror}" "#{config.nfsroot}/#{@config.mirror_mountpoint}"}

      raise "Can't mount #{@config.fai_debmirror} to #{@config.nfsroot}/#{@config.mirror_mountpoint}." unless exec_success?
    end

    # liloconfig, dump needs this file
    FileUtils.touch "#{@config.nfsroot}/etc/fstab"

    # We need these option before installing the first package. So we
    # can't put this into fai-nfsroot /etc/apt/apt.conf.d/90fai
    File.open "#{@config.nfsroot}/etc/apt/apt.conf.d/10fai", "w" do |f|
      f.puts <<~EOF
        APT::Install-Recommends no;
        APT::Get::AllowUnauthenticated true;
        Acquire::AllowInsecureRepositories "true";
        Aptitude::CmdLine::Ignore-Trust-Violations yes;
        Acquire::Languages none;
        Acquire::Check-Valid-Until false;
      EOF
    end

    # Currently we do not want to use systemd, but CD booting.
    unless @opts[:no_dracut]
      # If dracut is not included in default set of packages, dracut.conf.d is missing here.
      # ft nfsroot will install dracut at later point if necessary.
      unless Dir.exist? "#{@config.nfsroot}/etc/dracut.conf.d"
        Dir.mkdir "#{@config.nfsroot}/etc/dracut.conf.d"
      end

      File.open "#{@config.nfsroot}/etc/dracut.conf.d/01-omit.conf", "w" do |f|
        f.puts 'omit_dracutmodules+=" btrfs crypt dash lvm resume usrmount modsign mdraid shutdown dracut-systemd systemd systemd-initrd systemd-networkd "'
      end

      FileUtils.mkdir "#{@config.nfsroot}/etc/dracut.conf.d" unless Dir.exist? "#{@config.nfsroot}/etc/dracut.conf.d"
      File.open "#{@config.nfsroot}/etc/dracut.conf.d/12-no-systemd.conf", "w" do |f|
        f.puts 'omit_dracutmodules+=" dracut-systemd systemd systemd-initrd systemd-networkd systemd-initrd "'
      end
    end
  end

  def handle_nfsroot_packages
    if ENV["APTPROXY"]
      File.open "#{@config.nfsroot}/etc/apt/apt.conf.d/02proxy", "w" do |f|
        f.puts %{Acquire::http::Proxy "#{ENV["APTPROXY"]}";}
      end
    end

    verb "Upgrading #{@config.nfsroot}"

    run_scripts hook: "pre-nfsroot-packages"

    upgrade_nfsroot_base
    add_packages_to_nfsroot
    create_initrd
  end

  def check_required_packages
    if @@_verbose
      verb "FAI packages and related packages inside the nfsroot:"
      verb execvalue(%{#{rootcmd} dpkg-query -W -f='${Package;-18} ${Version}\n' fai-client fai-nfsroot fai-setup-storage dracut-network dracut dracut-live dracut-squash 2>/dev/null})
    end

    # check if all important packages are installed
    n = execvalue("#{rootcmd} dpkg-query -l fai-client fai-nfsroot fai-setup-storage 2>/dev/null|grep -E ^ii |wc -l").to_i
    if n != 3
        puts "ERROR: Some essential packages are missing."
        bad_exit
    end
    unless @opts[:no_dracut]
      n = execvalue("#{rootcmd} dpkg-query -l dracut-network dracut 2>/dev/null|grep -E ^ii |wc -l").to_i
      if n != 2
          puts "ERROR: Some essential dracut packages are missing."
          bad_exit
      end
    end

    unless @opts[:no_dracut]
      files = Dir.glob "#{@config.nfsroot}/boot/initrd*"
      raise "No initrd installed." if files.empty?
      if @opts[:logging]
        execute %{grep -E -q "^ERROR: |^E: Sub-process |^dpkg: error processing |^dpkg: dependency problems" #{@log_file.path}}, fail: false
        bad_exit if exec_success?
      end
    end
  end

  def post_install_configuration
    # Set FAI as the init (pid 1) when booting to a live/install CD.
    FileUtils.rm "#{@config.nfsroot}/etc/init.d/rcS" if File.exist? "#{@config.nfsroot}/etc/init.d/rcS"
    FileUtils.ln_sf "/usr/sbin/fai", "#{@config.nfsroot}/etc/init.d/rcS"

    # This second rm of the same file is needed.
    FileUtils.rm_f "#{@config.nfsroot}/etc/udev/rules.d/70-persistent-net.rules"

    # remove Ubuntu-specific rules to auto-start volume groups, which confuses setup-storage
    FileUtils.rm_f "#{@config.nfsroot}/lib/udev/rules.d/85-lvm2.rules"

    # definition for loopback device
    File.open "#{@config.nfsroot}/etc/network/interfaces", "w" do |f|
      f.puts "iface lo inet loopback"
    end
  end

  def check_nfsroot
    execute %{readlink "#{@config.nfsroot}/sbin/init" | grep -q systemd}, fail: false
    if exec_success?
        raise "Do not use systemd inside the nfsroot."
    end
  end

  private

  def setclasses
    # Export all the information we have about the system
    iarch = execvalue "#{rootcmd} dpkg --print-architecture|tr /a-z/ /A-Z/"
    distro = execvalue %{. "#{@config.nfsroot}/etc/os-release"; echo ${ID} ${ID}_${VERSION_ID} | tr '[:lower:].' '[:upper:]_'}

    # do we want to install the full list of packages?
    c = "FULL" if @all_packages

    classes = "NFSROOT #{c} #{iarch} #{distro} #{@nonfree}"
    ENV["classes"] = classes
    verb "Classes are set to #{classes}"
  end

  def create_initrd
    return if @opts[:no_dracut]

    # remove file diversion, then create initrd
    FileUtils.rm "#{@config.nfsroot}/etc/kernel/postinst.d/dracut"
    execute "#{rootcmd} dpkg-divert --rename --remove /etc/kernel/postinst.d/dracut"
    execute "#{rootcmd} dpkg-reconfigure dracut"
  end

  def upgrade_nfsroot_base
    FileUtils.mkdir_p "#{@config.nfsroot}/dev/pts"
    execute %{mount -t proc   /proc  "#{@config.nfsroot}/proc"}
    execute %{mount -t sysfs  /sys   "#{@config.nfsroot}/sys"}
    execute %{mount -t devpts devpts "#{@config.nfsroot}/dev/pts"}
    execute %{/usr/lib/fai/mkramdisk "#{@config.nfsroot}/var/lib/dpkg" "#{@config.nfsroot}/var/cache"}

    # stop mdadm from calling mkconf
    FileUtils.mkdir "#{@config.nfsroot}/etc/mdadm"
    FileUtils.touch "#{@config.nfsroot}/etc/mdadm/mdadm.conf"

    # add any 3rd party repository keys that you may need installed into the NFSROOT
    debug "Copying apt keys"
    if Dir.exist? "#{@fai_config_dir}/apt/keys"
      for f in `find #{@fai_config_dir}/apt/keys -type f -name '*.asc' -o -name '*.gpg'`.split("\n")
        debug "Copying #{f} to #{@config.nfsroot}/etc/apt/trusted.gpg.d"
        FileUtils.cp f, "#{@config.nfsroot}/etc/apt/trusted.gpg.d"
      end
    end

    # if we have the keyring on the host, install the key of the fai-project.org package repository
    execute <<MULTILINE
if [ ! -f "#{@config.nfsroot}/etc/apt/trusted.gpg.d/fai-project.gpg" -a -n "$(apt-key list 2BF8D9FE074BCDE4 2>/dev/null)" ]; then
    HOME=/root apt-key export 2BF8D9FE074BCDE4 > "#{@config.nfsroot}/etc/apt/trusted.gpg.d/fai-project.asc"
fi
MULTILINE

    pkgs = @all_packages ? "nfs-common" : ""

    execute "#{rootcmd} apt-get update"

    # fai-nfsroot must be successfully installed before anything else
    # TODO: why fai-nfsroot must be first installed?
    execute "#{rootcmd} apt-get -yf --no-install-recommends install fai-nfsroot #{pkgs}"
    pkgs = ""

    unless @opts[:no_dracut]
      # check if we need more dracut packages (dracut >=049)
      execute "#{rootcmd} apt show dracut-live dracut-squash >/dev/null 2>/dev/null", log: false, fail: false
      pkgs = "dracut-live dracut-squash" if exec_success?

      # TODO: this should be later in pipeline, and also should be triggered for --refresh.
      # If using Dracut specific compression algorithm, ensure that is installed.
      if @config.dracut_compress
        if @config.dracut_compress == "lzo"
          pkgs += " lzop"
        elsif @config.dracut_compress == "xz"
          pkgs += " xz-utils"
        else
          pkgs += " #{@config.dracut_compress}"
        end
      end

      # disable generating the initrd
      fdivert "/etc/kernel/postinst.d/dracut"
      execute "#{rootcmd} ln -s /bin/true /etc/kernel/postinst.d/dracut"

      # fai-nfsroot must be successfully installed before anything else
      # TODO: why fai-nfsroot must be first installed?
      execute "#{rootcmd} apt-get -yf --no-install-recommends install dracut-network dracut-config-generic #{pkgs}"
    end

    execute "#{rootcmd} apt-get -y dist-upgrade"

    unless @opts[:no_dracut]
      # patch dracut. Do not fail if network is not available
      fn = execvalue %{find "#{@config.nfsroot}/usr/lib/dracut/modules.d/" -name net-genrules.sh}
      execute "sed -i -e 's/exit 1/exit 0/' #{fn}"
      fdivert "/usr/sbin/update-grub"
    end
  end

  # Calls dpkg-divert to mark a file not to be overridden when a package updates.
  def fdivert *args
    for item in args
      execute %{#{rootcmd} dpkg-divert --quiet --add --rename "#{item}"}
    end
  end
end
