require_relative "nfsroot-impl"
require_relative "../ft"

class Nfsroot::NfsrootSuse < Nfsroot::NfsrootImpl

  def startup_checks
    unless command_available? "rinse"
      error "rinse must be installed in order to create a SUSE nfsroot."
    end

    unless command_available? "firejail"
      error "firejail must be installed in order to create a SUSE nfsroot."
    end
    # TODO: I probably should create my own profile for this.
    rootcmd = %{firejail --noprofile --quiet --chroot=#{@@nfsroot} --shell=none env LC_ALL="C"}

    if not @config.nfsroot_distribution_version
      error "DISTRIBUTION_VERSION should be set in nfsroot.conf for SUSE nfsroot creation."
    end
  end

  def nfsroot_configuration
    # Only install required packages, no recommends.
    zypp_conf_path = "#{@config.nfsroot}/etc/zypp/zypp.conf"
    zyppconf = File.read zypp_conf_path
    if zyppconf.sub!("# solver.onlyRequires = false", "solver.onlyRequires = true")
      File.write zypp_conf_path zyppconf
    end
  end

  def perform_refresh
    execute "#{rootcmd} zypper --non-interactive refresh"
    execute "#{rootcmd} zypper --non-interactive dup"
  end

  def fai_version
    @fai_version ||= execvalue %{#{rootcmd} rpm -q --queryformat='%{VERSION}' fai-client}
  end

  def copy_package_manager_configuration
    # Copies files inside repos.d to target.
    FileUtils.cp_r("#{@fai_config_dir}/zypper/repos.d/.", "#{@config.nfsroot}/etc/zypp/repos.d", preserve: true)
  end

  def add_packages_to_nfsroot
    ENV["FAI_ROOT"] = @config.nfsroot

    # Run hooks in pre-packages dir.
    run_scripts hook: "pre-packages"

    setclasses

    # TODO: instead of install_packages, maybe better use my script for this here.
    output = execute %{env LC_ALL="C" install_packages -v "-p#{@fai_config_dir}"}
    verb "Adding additional packages to #{@config.nfsroot}:"
    verb output

    File.open "#{@config.nfsroot}/var/tmp/packages.nfsroot", "w" do |f|
      f.write output
    end

    warnings = output.scan(/^WARNING: (.*)$/)
    warnings = warnings.first

    # Some warnings doesn’t make on SUSE.
    warnings.delete_if do |warning|
      if warning.start_with? "Uppercase character found in package name in line"
        true
      else
        false
      end
    end if warnings

    if not warnings.nil? and not warnings.empty?
      STDERR.puts
      STDERR.puts "When installing packages, following errors happened:"
      warnings.each do |warning|
        STDERR.puts"  - #{warning}"
      end
      STDERR.puts
      STDERR.puts "Please fix above errors and try again."
      exit 1
    end

    #FileUtils.rm "#{@config.nfsroot}/var/lib/dpkg/available"
  end

  # This is for now only useful for Debian nfsroot creation.
  def solve_repository_info
    # Nya.
  end

  # Called by Methods#create_nfsroot.
  def create_nfsroot
    if @base_archive
      extract_base @base_archive
    else
      create_base
    end

    FileUtils.touch ".THIS_IS_THE_FAI_NFSROOT"

    # save the list of all packages in the base.tar.gz
    execute %{rpm -qa --queryformat "%{NAME}\n" > "#{@config.nfsroot}/var/tmp/base-pkgs.lis"}

    unless @opts[:no_dracut]
      File.open "#{@config.nfsroot}/etc/dracut.conf.d/01-omit.conf", "w" do |f|
        f.puts 'omit_dracutmodules+=" btrfs crypt dash lvm resume usrmount modsign mdraid shutdown "'
      end
    end
  end

  def handle_nfsroot_packages
    verb "Upgrading #{@config.nfsroot}"

    run_scripts hook: "pre-nfsroot-packages"

    upgrade_nfsroot_base
    add_packages_to_nfsroot
    create_initrd
  end

  # check if all important packages are installed
  def check_required_packages
    puts "FAI packages and related packages inside the nfsroot:"
    execute %{#{rootcmd} rpm -q --qf "%{NAME}-%{VERSION}\n" fai-client fai-nfsroot fai-setup-storage dracut 2>/dev/null}, fail: false

    execute("#{rootcmd} rpm --quiet -q fai-client fai-nfsroot fai-setup-storage", fail: false)
    error "Some of essential packages are missing from nfsroot: fai-client fai-nfsroot fai-setup-storage" unless exec_success?
    unless @opts[:no_dracut]
      execute("#{rootcmd} rpm --quiet -q dracut", fail: false)
      error "Dracut must included in nfsroot." unless exec_success?
    end
  end

  # Post-install customization to make a SUSE nfsroot work properly.
  def post_install_configuration

    # Install systemd services to launch fai.

    fai_target = <<~EOF
    [Unit]
    Description=FAI installation
    DefaultDependencies=no
    Requires=sysinit.target
    After=sysinit.target
    Conflicts=shutdown.target
    Before=shutdown.target

    Wants=fai.service
    EOF

    # FAI does some extra stuff when launching as an “init”. Let’s do those by hand here.
    fai_service = <<~EOF
    [Unit]
    Description=Runs fai operations

    DefaultDependencies=no
    After=getty@tty1.service
    Requires=local-fs.target sysinit.target
    After=local-fs.target sysinit.target
    Before=shutdown.target

    RefuseManualStart=true
    ConditionPathIsDirectory=/tmp

    [Service]
    Type=simple

    StandardError=tty
    StandardOutput=tty
    StandardInput=tty
    TTYPath=/dev/tty1
    TTYReset=yes
    TTYVHangup=yes

    # TODO: I wonder how this will affect boot cycle.
    RemainAfterExit=yes
    ExecStart=/usr/sbin/start_fai
    EOF

    fai_script = <<~EOF
    #!/bin/bash

    # Simple script to make FAI launched like intended.

    # Enforce fai to do init tasks.
    sed 's/export do_init_tasks=0/export do_init_tasks=1/' /usr/sbin/fai > /usr/sbin/fai_patched

    chmod 700 /usr/sbin/fai_patched

    /usr/sbin/fai_patched
    EOF

    File.open "#{@config.nfsroot}/etc/systemd/system/fai.target", "w" do |f|
      f.write fai_target
    end

    File.open "#{@config.nfsroot}/etc/systemd/system/fai.service", "w" do |f|
      f.write fai_service
    end

    File.open "#{@config.nfsroot}/usr/sbin/start_fai", "w" do |f|
      f.write fai_script
    end
    FileUtils.chmod 700, "#{@config.nfsroot}/usr/sbin/start_fai"

    # FAI’s check-cross-arch expects /bin/more to be real binary, so let’s make it so.
    # TODO: fix upstream to handle the symlink properly.
    if File.symlink? "#{@config.nfsroot}/bin/more"
      FileUtils.rm "#{@config.nfsroot}/bin/more"
      FileUtils.cp "#{@config.nfsroot}/usr/bin/more", "#{@config.nfsroot}/bin/more"
    end

    # Enable dhcp for all interfaces by default.
    content = <<~EOF
    [Match]
    Name=*

    [Network]
    DHCP=yes
    EOF
    File.open "#{@config.nfsroot}/etc/systemd/network/80-default-dhcp.network", "w" do |f|
      f.write content
    end
    execute "#{rootcmd} systemctl disable network.service"
    execute "#{rootcmd} systemctl enable systemd-networkd.service"

  end

  # NOTE: Debian version wants ensure systemd is not set as init. But does it really matter here?
  # It probably makes sense as not installing systemd saves some space though.
  def check_nfsroot
    # Nya.
  end

  private

  def setclasses
    # Export all the information we have about the system
    iarch = execvalue("uname -m").upcase
    distro = execvalue %{. "#{@config.nfsroot}/etc/os-release"; echo ${ID} ${ID}_${VERSION_ID} | tr '[:lower:].' '[:upper:]_'}

    # do we want to install the full list of packages?
    c = "FULL" if @all_packages

    classes = "NFSROOT #{c} #{iarch} #{distro} #{@nonfree}"
    ENV["classes"] = classes
    verb "Classes are set to #{classes}"
  end

  def create_initrd
    return if @opts[:no_dracut]
    # remove file diversion, then create initrd
    if File.exist? "#{@config.nfsroot}/etc/kernel/postinst.d/dracut"
      FileUtils.rm "#{@config.nfsroot}/etc/kernel/postinst.d/dracut"
    end
    #execute "#{rootcmd} dpkg-divert --rename --remove /etc/kernel/postinst.d/dracut"
    execute "#{rootcmd} mkinitrd"
  end

  # Creates a temporary base archive that is supplied with generated nfsroot.
  def create_base
    # Check if NFSROOT directory is mounted with bad options.
    fs = execvalue %{df -P "#{@config.nfsroot}" | tail -1 | awk '{print $6}'}
    if execvalue(%{mount | grep \"on "#{fs}" \" |  awk '{print $6}'}).match?(/nosuid|nodev/)
      raise "NFSROOT directory #{@config.nfsroot} is mounted using nosuid or nodev. Aborting"
    end

    # Actually create the base archive.
    ARGV << @ft_output_flags if @ft_output_flags
    ARGV << "base"
    ARGV << "--fai-config-space" << @config.config_space
    ARGV << "--distribution" << @config.nfsroot_distribution
    ARGV << "--distribution-version" << @config.nfsroot_distribution_version
    ARGV << "--no-compress"
    ARGV << "--output-directory" << @config.nfsroot
    if @config.rinse_additional_packages
      ARGV << (@config.rinse_additional_packages.map { |x| [ "--add-package", x ] })
    end
    ARGV.flatten!

    verb ""
    debugstr = @@_debug && "(at #{caller_locations(0,1)[0]})" || ""
    verb "Running “ft #{ARGV.join(" ")}” #{debugstr}"
    verb ""

    ft = FT.new
    ft.options
    ft.perform

    verb "Base created."
    verb ""

    execute "#{rootcmd} zypper clean"

    if File.exist? "#{@config.nfsroot}/etc/resolv.conf"
      # NOTE: I’m not sure this really does anything. Shouldn’t be too dangerous though.
      FileUtils.rm "#{@config.nfsroot}/etc/resolv.conf"
    end
    FileUtils.rm "#{@config.nfsroot}/etc/hostname"

    compressor = @base_compress_algo
    compressor ||= "xz"
    verb "Creating base.tar.#{compressor}"
    excludes = %{--exclude etc/machine-id --exclude var/tmp/base.tar}
    execute %{tar --xattrs --acls --one-file-system -C "#{@config.nfsroot}" -cf "#{@config.nfsroot}/var/tmp/base.tar" #{excludes} .}

    execute %{nice #{compressor} -q "#{@config.nfsroot}/var/tmp/base.tar"}, wait: false
  end

  def upgrade_nfsroot_base
    # NOTE: If I’m not reading this wrong, these mounts are not unmounted at all.
    # Only place these are unmounted is before start of new nfsroot creation.
    # Not sure if that’s good or bad, as this conveniently makes the nfsroot behave more nicely
    # when using it by hand.
    # TODO: Instead of that, maybe have something like ”ft chroot” instead?
    FileUtils.mkdir_p "#{@config.nfsroot}/dev/pts"
    execute %{mount -t proc   /proc  "#{@config.nfsroot}/proc"}
    execute %{mount -t sysfs  /sys   "#{@config.nfsroot}/sys"}
    execute %{mount -t devpts devpts "#{@config.nfsroot}/dev/pts"}

    # TODO: this really needs to be implemented, instead of that autoimport below.
    # add any 3rd party repository keys that you may need installed into the NFSROOT
=begin
    debug "Copying apt keys"
    if Dir.exist? "#{@fai_config_dir}/apt/keys"
      for f in `find #{@fai_config_dir}/apt/keys -type f -name '*.asc' -o -name '*.gpg'`.split("\n")
        debug "Copying #{f} to #{@config.nfsroot}/etc/apt/trusted.gpg.d"
        FileUtils.cp f, "#{@config.nfsroot}/etc/apt/trusted.gpg.d"
      end
    end
=end

    # if we have the keyring on the host, install the key of the fai-project.org package repository
    #execute <<MULTILINE
#if [ ! -f #{@config.nfsroot}/etc/apt/trusted.gpg.d/fai-project.gpg -a -n "$(apt-key list 2BF8D9FE074BCDE4 2>/dev/null)" ]; then
#    HOME=/root apt-key export 2BF8D9FE074BCDE4 > #{@config.nfsroot}/etc/apt/trusted.gpg.d/fai-project.asc
#fi
#MULTILINE

    # TODO: see “Copying apt keys” todo.
    execute "#{rootcmd} zypper --gpg-auto-import-keys --non-interactive refresh --force"

    # Forcibly update all existing packages.
    execute "#{rootcmd} zypper --non-interactive dup"
  end

end

