
require "fileutils"
require "socket"
require "tempfile"
require "optimist"

require_relative "../options"
require_relative "online-update"

# Main class.
class OnlineUpdate::PerformOnHost
  include Options

  def perform
    startup_checks

    cleanup

    @fai_mirror_path = "/var/tmp/fai_mirror"
    @packages_dir_path = "/var/tmp/packages"

    at_exit do
      # Comment out in order to debug leftover state of the script.
      #cleanup
    end

    # Transfer temporary sources.list to target.

    content = generate_sources_list
    tmpfile = Tempfile.new "perform-on-host"
    tmpfile.write content
    tmpfile.flush
    execute %{scp -P #{@ssh_port} #{tmpfile.path} root@#{@target_host}:/var/tmp/update.sources.list}
    # TODO: remove the tmpfile as it’š not needed any more.

    # Transfer packages to target.

    if @packages_dir
      #archive = "/tmp/packages.tar"
      #path = File.dirname @packages_dir
      #target_dir = File.basename @packages_dir
      #excludes = @include_debug ? '--exclude="*dbgsym*"' : ""
      #execute %{tar #{excludes} -cvf #{archive} -C #{path} #{target_dir}}
      #execute %{scp -P #{@ssh_port} #{archive} root@#{@target_host}:/var/tmp}

      #execute %{#{ssh_command} 7z x /var/tmp/packages.7z -o/var/tmp}
      #execute %{#{ssh_command} tar xf /var/tmp/packages.tar -C /var/tmp}

      # Ensure there is target dir on target host.
      execute %{#{ssh_command} mkdir -p #{@packages_dir_path}}

      # TODO: delete locally removed files from target.
      execute %{rsync -az -e "ssh -p #{@ssh_port}" #{@packages_dir}/* root@#{@target_host}:#{@packages_dir_path}}
    end

    if @fai_mirror
      # Ensure there is mirror dir on target host.
      execute %{#{ssh_command} mkdir -p #{@fai_mirror_path}}

      if @target_distribution == "debian"
        #execute %{scp -r -P #{@ssh_port} #{@fai_mirror}/{dists,pool} root@#{@target_host}:#{@fai_mirror_path}}
        execute %{rsync -az -e "ssh -p #{@ssh_port}" #{@fai_mirror}/{dists,pool} root@#{@target_host}:#{@fai_mirror_path}}
      elsif @target_distribution == "suse"
        execute %{rsync -az -e "ssh -p #{@ssh_port}" #{@fai_mirror}/* root@#{@target_host}:#{@fai_mirror_path}}
      end
    end

    # Transfer config space to target.
    config_space_archive = "config_space.tar.xz"
    config_space_path = File.dirname @config_space
    config_space_target_dir = File.basename @config_space
    execute %{tar --exclude="basefiles/*" --exclude=".git/*" -cf /tmp/#{config_space_archive} -C #{config_space_path} #{config_space_target_dir}}
    execute %{scp -P #{@ssh_port} /tmp/#{config_space_archive} root@#{@target_host}:/var/tmp}

    execute %{#{ssh_command} tar xf /var/tmp/#{config_space_archive} -C /var/tmp}

    # Renames the file to correct name, to avoid extra confusion.
    execute %{#{ssh_command} mv /var/tmp/#{config_space_target_dir} /var/tmp/fai_config}

    # Perform actual update.

    class_string = @add_classes.map do |c|
      "--add-class #{c}"
    end.join(" ")

    execute %{#{ssh_command} perform-online-update #{class_string} --sources-list /var/tmp/update.sources.list}
  end

  def options
    options = common_options do
      opt :packages_dir,          "Path to directory containing packages that will be used for update",       type: :string,    short: "D"
      opt :target_host,           "Host where update will be performed at",                                   type: :string,    short: "H"
      opt :ssh_port,              "Port where target host’s SSHD is listening at",                            type: :integer,   short: "P"
      opt :config_space,          "Path to FAI config space",                                                 type: :string,    short: "c", default: "/srv/fai/config"
      opt :fai_mirror,            "Path to FAI mirror, created by fai-mirror or create-fai-mirror",           type: :string,    short: "m"
      opt :target,                "Target distribution. Currently support values are “debian” and “suse”",    type: :string,    short: "t"
      opt :add_class,             "Add a class to target host. Can be given multiple times",                  type: :string, multi: true, short: "a"
      opt :include_debug,         "Include debug (dbgsym) packages when transferring packages dir to target", type: :boolean,   short: "e"
    end

    if !options[:packages_dir].nil? and !options[:fai_mirror].nil?
      raise "Only one of --fai-mirror and --packages-dir can be used."
    end

    unless Dir.exist? options[:config_space]
      raise "Given FAI config space (#{options[:sources_list]}) does not exist."
    end
    @config_space = options[:config_space]
    unless Dir.exist? "#{@config_space}/class"
      raise "Given FAI config space (#{options[:sources_list]}) is malformed."
    end

    if options[:packages_dir].nil?
      if options[:fai_mirror].nil? or !Dir.exist? options[:fai_mirror]
        raise "Given FAI mirror dir (#{options[:fai_mirror].inspect}) does not exist."
      end
      @fai_mirror = options[:fai_mirror]
    end

    if options[:fai_mirror].nil?
      if options[:packages_dir].nil? or !Dir.exist? options[:packages_dir]
        raise "Given path to packages dir “#{options[:sources_list]}” does not exist."
      end
      @packages_dir = options[:packages_dir]
    end

    @target_host = options[:target_host]
    @ssh_port = options[:ssh_port]
    @add_classes = options[:add_class]
    @include_debug = options[:include_debug]

    if @target_host.nil? or @ssh_port.nil?
      raise "--target-host and --ssh-port must be given"
    end

    begin
      Socket.tcp(@target_host, @ssh_port, connect_timeout: 3) {}
    rescue Errno::ETIMEDOUT => e
      puts.STDERR "Connecting to host #{@target_host}:#{@ssh_port} failed. Please check if there is a connection problem."
      raise e
    end

    @target_distribution = options[:target]
    case @target_distribution
    when "debian"
      # Platform specific checks can be done here.
    when "suse"
      puts "WARNING: This path has not been tested. Remove following exit locally if you want to test this for SUSE."
      exit 1
      # Platform specific checks can be done here.
    else
      raise "Invalid target distribution: #{@target_distribution.inspect}"
    end

    set_opts options
  end

  private

  def generate_sources_list
    if @fai_mirror
      # TODO: The mirror should be signed, even though the effect is limited. It still has its used.
      # After it’s been signed, trusted=yes can be removed.
      content = <<~EOF
      deb [trusted=yes] file:#{@fai_mirror_path}/ cskoeln main non-free contrib
      EOF
    elsif @packages_dir
      content = <<~EOF
      deb file:#{@packages_dir_path} ./
      EOF
    end

    content
  end

  def ssh_command
    %{ssh -p #{@ssh_port} root@#{@target_host}}
  end

  def startup_checks
    unless command_available? "ssh"
      raise "You need to have ssh (Secure SHell) available."
    end

    unless command_available? "scp"
      raise "You need to have scp (a utility to copy files over ssh) available."
    end

    unless command_available? "tar"
      raise "You need to have tar (archiver) to create a package archive."
    end
  end

  def cleanup
    execute %{#{ssh_command} rm -rf /var/tmp/fai_config /var/tmp/update.sources.list}, fail: false
    execute %{#{ssh_command} rm -rf /var/tmp/config_space.tar.xz /var/tmp/packages.tar}, fail: false

    # This could be used to clean up existing package data, but to save bandwidth, that shouldn’t be done at least on my development environments.
    #execute %{#{ssh_command} rm -rf #{@packages_dir_path}, fail: false
    #execute %{#{ssh_command} rm -rf #{@fai_mirror_path}, fail: false
  end
end
