require "fileutils"

# Gems.
require "optimist"
require "mimemagic"

require_relative "../fai-config"
require_relative "../common"
require_relative "packaging"
require_relative "qemu"
require_relative "ssh"

# Compress an image:
#   zstdmt --long -15 --zstd=hlog=26 disks/build_machine_image.qcow2

# TODO: Maybe there could be simpler check for raw images that checks that given file is ”raw image” by file command and
# then validate it’s raw image with qemu-img info?
class Packaging::BuildPackage
  include Common

  DISABLE_SSH_KEY_CHECK = "-o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null"

  def perform
    startup_checks
    real_perform
  end

  def options
    options = common_options do
      opt :source_path,       "Path to source code that needs to be built.",
          short: "s", type: :string
      opt :packages_path,     "Path where built packages as written to",
          short: "p", type: :string
      opt :build_machine,     "Path to build machine image. Given machine must be checkable with “qemu-img check”",
          short: "m", type: :string
      opt :hypervisor,        "Virtualization platform to use. Default is platform specific. To get list of available platforms, use --hypervisor help",
          short: "H", type: :string
      opt :target_project,    "Target OBS project. Defaults to source path’s innermost directory name.",
          short: "t", type: :string
      opt :target_distro,     "Target distribution for building",
          short: "D", type: :string
      opt :target_arch,       "Target architecture for building",
          short: "a", type: :string,    default: "x86_64"
    end

    set_opts options

    # TODO: some platform specific detection probably should be here.
    @hypervisor = options[:hypervisor]
    @hypervisor = "qemu" if @hypervisor.nil?
    if @hypervisor == "help"
      puts
      puts "Available hypervisors:"
      puts "  - qemu"
      puts
      exit 1
    end

    available_hypervisors = [
      "qemu"
    ]
    unless available_hypervisors.include? @hypervisor
      raise "Given hypervisor is not available. See --hypervisor help for available hypervisors."
    end

    if @hypervisor == "qemu"
      @impl = Packaging::Qemu.new options
    end

    if options[:build_machine].nil? or options[:build_machine].empty?
      raise "--build-machine must be given."
    end
    @build_machine = options[:build_machine]
    @impl.build_machine = @build_machine

    raise "#{@build_machine} is not valid file." unless File.exist? @build_machine

    if options[:packages_path].nil? or options[:packages_path].empty?
      raise "--packages-path must be given."
    end
    @packages_path = options[:packages_path]

    FileUtils.mkdir @packages_path unless Dir.exist? @packages_path

    if options[:source_path].nil? or options[:source_path].empty?
      raise "--source-path must be given."
    end
    @source_path = options[:source_path]

    raise "--source-path must be a valid directory" unless Dir.exist? @source_path

    @mac = generate_mac
    @impl.mac = @mac

    @tap_device = "buildtap0"
    @impl.tap_device = @tap_device
    @bridge_device = "virbr1"
    @impl.bridge_device = @bridge_device

    @target_distribution = options[:target_distro]
    raise "--target must be given. See relevant project in OBS for valid values." if @target_distribution.nil? or @target_distribution.empty?
    @target_architecture = options[:target_arch]
    raise "--target-arch must be given. See relevant project in OBS for valid values." if @target_architecture.nil? or @target_architecture.empty?

    # Default to directory where files to be built resides.
    @target_project = options[:target_project]
    @target_project = File.basename @source_path if @target_project.nil?
  end

  private

  def real_perform
    @impl.start_vm @build_machine

    # Wait for VM to boot up.
    sleep 3

    @impl.wait_for_connection

    # Allow user to log in as build user.
    ssh_key = resolve_ssh_key
    execute_on_remote "root", %{ainsl -a -v /home/build/.ssh/authorized_keys "#{ssh_key}"}

    remote_path = "cache/KC/#{@target_project}"

    # Transfer files to target.
    execute_on_remote "build", %{cd #{remote_path} ; rm -rf #{@target_project}}
    # rsync is missing from build_machine, and it doesn’t make much sense anyway.
    #execute %{rsync -e "ssh -l build" -avzh #{@source_path} build@#{@impl.vm_ip}:build}
    execute %{scp #{DISABLE_SSH_KEY_CHECK} -r #{@source_path} build@#{@impl.vm_ip}:#{remote_path}/#{@target_project}}

    execute_on_remote "build", %{cd #{remote_path} ; tar Jcf "#{@target_project}.tar.xz" "#{@target_project}"}

    execute_on_remote "build", %{cd #{remote_path} ; env BUILD_DIR=/usr/lib/obs-build osc build --trust-all-projects -o "#{@target_distribution}" "#{@target_architecture}"}

    if @target_distribution.include? "Debian"
      packagedir = "DEBS"
    else
      raise "Not implemented for other distributions than Debian so far."
    end
    execute %{scp #{DISABLE_SSH_KEY_CHECK} root@#{@impl.vm_ip}:/var/tmp/build-root/#{@target_distribution}-#{@target_architecture}/usr/src/packages/#{packagedir}/* #{@packages_path}}

    #puts "DEBUG"
    #sleep 100000

    @impl.close_hypervisors
  end

  def startup_checks

    # Set up mimemagic recognition bits.

    MimeMagic.add "application/x-zstd", {
      extensions: [ "zst" ],
      magic: [
        [0, "\050\265\057\375".force_encoding("ASCII-8BIT")] #ASCII-8BIT is what Ruby says to be binary data.
      ],
      comment: "Zstandard v0.8+"
    }

    # This probably matches some non-qcow2 files too.
    MimeMagic.add "x-application/qemu-image-format", {
      extensions: [ "qcow", "qcow2" ],
      magic: [
        [0, "\121\106\111\373"]
      ],
      comment: "Qemu image format."
    }

    # Decompress image as necessary.
    mime = MimeMagic.by_magic(File.open @build_machine)

    if mime.nil?
      raise "Were not able to recognize give file. Probably you’re using some new type of file that does not yet have support added to this script."
    end

    case mime.subtype
    when "x-zstd"
      verb "Extracting image..."
      FileUtils.rm_f "/tmp/build_machine" # Remove existing image if it exists.
      execute %{zstdmt -d #{@build_machine} -o /tmp/build_machine}
      @build_machine = "/tmp/build_machine"

      # Remove the extracted image
      at_exit do
        FileUtils.rm "/tmp/build_machine"
      end
    when "qemu-image-format"
      verb "Using qcow2 image directly. You can use compressed image to save some extra space."
    else
      raise "Unrecognized file type for #{@build_machine}. Currently handled qcow2 and zstd compressed qcow2 images."
    end

    # If image was decompressed, the path needs to be updated on worker class too.
    @impl.build_machine = @build_machine

    # Run platform specific checks.
    @impl.startup_checks

    # Set up networking.

    network_setup
    use_bridging
    at_exit do
      remove_tap_device
    end
  end

  def generate_mac
    # Deterministic macs can be generated with sample(random: Random.new(1)).
    #22-22-22-xx-xx-xx
    "22:22:22:" + 6.times.map { num = (0..15).to_a.sample; num.to_s(16) }.each_slice(2).to_a.map { |arr| arr.join('') }.join(":").upcase
  end

  def remove_tap_device
    # Only link delete is required, since it automatically deletes the tap device.
    #sudo ip tuntap del dev "#{@tap_device}" mode tap
    execute %{sudo ip link del dev "#{@tap_device}" master "#{@bridge_device}"}
  end

  def use_bridging
    # Use libvirt’s bridge device to provide connection.
    verb "Adding #{@bridge_device}"

    existing = execvalue %(ip -br link show #{@bridge_device} | awk '{ print $1 }')
    verb "Create tap device"
    execute "sudo ip tuntap add dev #{@tap_device} mode tap vnet_hdr"
    verb "Set tap device $tap_device’s master: $bridge_device"
    execute %{sudo ip link set dev "#{@tap_device}" master "#{@bridge_device}"} # set $bridge_device as the target bridge for $tap_device
    verb "Bring tap up"
    execute %{sudo ip link set dev "#{@tap_device}" up}
  end

  def network_setup
    # Ensure that apt-cacher-ng is usable inside VM.
    execute "sudo firewall-cmd --add-port=3142/tcp", fail: false, log: false
    @@port_originally_open = !exec_success?

    # Necessary to get traffic working with wlan interface.
    @@original_proxy_arp = execvalue "sudo sysctl net.ipv4.conf.all.proxy_arp -n"
    execute "sudo sysctl net.ipv4.conf.all.proxy_arp=1", log: false

    at_exit do
      execute("sudo firewall-cmd --add-port=3142/tcp", fail: false, log: false) unless @@port_originally_open
      execute "sudo sysctl net.ipv4.conf.all.proxy_arp=#{@@original_proxy_arp}", log: false
    end
  end

  # Just a dummy method to read one public key from ~/.ssh
  def resolve_ssh_key
    keys = Dir.glob File.expand_path("~/.ssh/id_*.pub")
    raise "Failed to resolve ssh public key. Probably an option should be added to manually specify one." if keys.empty?
    keys.sort!
    File.readlines(keys.first).first
  end

  def execute_on_remote user, command
    execute %{/usr/bin/ssh #{DISABLE_SSH_KEY_CHECK} -p 22 #{user}@#{@impl.vm_ip} '#{command}'}
  end
end