require "fileutils"

require_relative "packages-impl"

class Packaging::PackagesDebian < Packaging::PackagesImpl

  def initialize opts
    super opts

    # Map package name from the list (including +-/=) to real package names.
    @package_name_map = {}

    # Key is package names, value is whether we want to install it
    @to_be_installed_flags = {}

    # List of packages, but unknown packages removed.
    @existing_packages = []

    # Hash of all available packages
    @available_packages = {}

    @debsourcesdir = "/var/lib/fai/packages"
  end

  def startup_checks
    unless File.file? "#{@FAI_ROOT}/etc/os-release"
      error "#{@FAI_ROOT}/etc/os-release does not exist. Do you have proper nfsroot?"
    end

    unless File.foreach("#{@FAI_ROOT}/etc/os-release").any?{ |l| l['NAME="Debian GNU/Linux"'] }
      raise "Tried to install Debian packages, but #{@FAI_ROOT} is not a Debian."
    end

    # Reads packages from the available file, if it exists.
    create_debian_pkg_list if File.exist? "#{@FAI_ROOT}/var/lib/dpkg/available"
  end

  def download_commands pm_command_options, install_options
    quiet_apt = @@_verbose && "" || "-qq"
    devnull = @@_verbose && "" || ">/dev/null"

    aptoptions = ENV["aptoptions"]

    {
               "install" => "apt-get #{pm_command_options} #{quiet_apt} -d #{aptoptions} -y --fix-missing install #{install_options}",
         "install-norec" => "apt-get #{pm_command_options} #{quiet_apt} -d #{aptoptions} -y --fix-missing install --no-install-recommends #{install_options}",
              "taskinst" => "aptitude #{pm_command_options} -d #{aptoptions} -y install #{install_options} #{devnull}",
                   "apt" => "apt #{pm_command_options} -d #{aptoptions} -y install #{install_options} #{devnull}",
              "aptitude" => "aptitude #{pm_command_options} -R -d #{aptoptions} -y install #{install_options} #{devnull}",
            "aptitude-r" => "aptitude #{pm_command_options} -r -d #{aptoptions} -y install #{install_options} #{devnull}",
                "unpack" => "cd #{@download_directory}; aptitude download",
                  "cupt" => "cupt -R --download-only #{aptoptions} -y install #{install_options} #{devnull}",
                "cupt-r" => "cupt --download-only #{aptoptions} -y install #{install_options} #{devnull}"
    }
  end

  def create_package_list packages
    # NOTE from original install_packages: CURRENTLY no packages are removed from the list of known packages

    complete = packages.dup

    # on Debian system, clean list of packages
    if !@no_package_name_check and File.exist? "#{@FAI_ROOT}/var/lib/dpkg/available"
      known = clean_pkg_list complete
    else
      known = complete
    end

    writepackages known

    known
  end

  # Warning if uppercase letters are found (package are always lowercase).
  def readconfig_checks type, file, line
    warn "Uppercase character found in package name in line #{line}" if line =~ /[A-Z]/
    error "PACKAGES [..] line missing in #{file}" unless type
  end

  # Download debian source packages.
  def get_sources known
    return unless @download_sources

    verb "Trying to retrieve sources as specified via -s option for install_packages"
    error "FAI_DEBSOURCESDIR is not set, cannot retrieve sources." unless ENV["FAI_DEBSOURCESDIR"]
    if @dry_run
      puts "Skipping source download due dry-run."
      return
    end

    FileUtils.mkdir_p "#{@FAI_ROOT}/#{@debsourcesdir}"
    FileUtils.mkdir_p ENV['FAI_DEBSOURCESDIR']

    known.each do |package|
      execute(%{#{@rootcmd} sh -c "cd '#{@debsourcesdir}' ; apt-get --download-only source '#{package}'"})
      FileUtils.mv Dir.glob("#{@FAI_ROOT}/#{@debsourcesdir}/*"), ENV['FAI_DEBSOURCESDIR']
    end
  end

  def dselectupgrade command, commands_command, package_list

    list = ""
    package_list[command].each do |array|
      package = array.shift
      action = array.shift
      list += "#{package} #{action}\n"
    end

    tempfile = Tempfile.new("packages-debian-dselectupgrade-", "#{@FAI_ROOT}/tmp")
    if @dry_run
      puts "Would write #{list} to file #{tempfile.path}"
    else
      tempfile.write list
    end

    execute %{#{@rootcmd} dpkg --set-selections < "#{tempfile.path}"}
    execute "#{@rootcmd} #{commands_command}"
  end

  private

  # Build a hash of all known package names.
  def create_debian_pkg_list
    File.open "#{@FAI_ROOT}/var/lib/dpkg/available" do |f|
      line = f.readline
      next unless line.start_with? "Package: "
      line.delete_prefix! "Package: "

      @available_packages[line] = 1
    end
  end

  # clean up the list of packages, remove unknown packages,
  # "merge" things like: pkgA pkgA-
  def clean_pkg_list packages
    inlist = packages

    unknown = []
    final = [] # final list of packages

    # clean global hashes
    @package_name_map = {}
    @to_be_installed_flags = {}

    inlist.each do |in_package|
      pack = in_package

      # plain package name is known, this also matches g++, memtest86+, ...
      next if insert_pkg(in_package, pack, 1, "#{in_package} found")

      # handle packagename-
      if pack.sub! /-$/, ""
        next if insert_pkg(in_package, pack, 0, "#{pack} - removed")
      end

      # handle packagename+, where packagename itself does not include a +
      if pack.sub! /\+$/, ""
        next if insert_pkg(in_package, pack, 1, "#{pack} + added")
      end

      # remove /distribution or =version from package name
      if pack.sub! %r{[/=].+}, ""
        next if insert_pkg(in_package, pack, 1, "#{in_package} using #{pack} found")
      end

      # remove :arch from package name
      if pack.sub! /:\S+/, ""
        next if insert_pkg(in_package, pack, 1, "#{in_package} using #{pack} found")
      end

      # handle $varname in package name
      if pack.sub! /\$\w+/, ""
        next if insert_pkg(in_package, pack, 1, "#{in_package} using #{pack} found")
      end

      # handle ${varname} in package name
      if pack.sub! /\$\{\w+\}/, ""
        next if insert_pkg(in_package, pack, 1, "#{in_package} using #{pack} found")
      end

      if $_debug
        puts "Got unknown package: #{in_package.inspect}. Dying because --debug has been given."
        exit 1
      end

      # If no matches, the package is unknown.
      unknown << in_package
    end

    error "These unknown packages are removed from the installation list: #{unknown.join(" ")}" unless unknown.empty;

    debug "-" * 30
    debug "ORIG: #{inlist.join("\n")}"
    debug "INSTALL: #{@existing_packages.join("\n")}"

    # create new list of packages (incl. +-/=) that should be installed or removed
    @existing_packages.each do |package|

      # pkg names with at end - that should be not installed

      if (@to_be_installed_flags[@package_name_map[package]] == 0 && package.match?(/-$/))
        final << package
      end

      next if (@to_be_installed_flags[@package_name_map[package]] == 1 && package.match?(/-$/));

      # do not include package that are marked as do not install
      next unless @to_be_installed_flags[@package_name_map[package]]

      final << package
    end

    debug "DO INST: #{final.join("\n")}"

    final
  end

  # If package name is known, insert package into data structure and return success.
  #
  # Inserts original package name and modified name into @package_name_map.
  #
  # Saves status install/remove of each package in @to_be_installed_flags.
  def insert_pkg orig, pack, install, msg
    return false if not @available_packages[pack]

    debug msg

    @existing_packages << orig
    @package_name_map[orig] = pack
    @to_be_installed_flags[pack] = install

    true
  end
end