
# Conversion of FAI’s install_packages to work better cross-distribution.

require "fileutils"

require_relative "packaging"
require_relative "packages-methods"
require_relative "packages-debian"
require_relative "packages-suse"
require_relative "packages-opensuse"
require_relative "../checks"
require_relative "../options"
require_relative "../fai-config"

# Env vars that must be declared:
# - classes
# Env vars that can be optionally declared:
# - FAI
# - FAI_ROOT
# - INSTALL_COMMAND_OPTIONS
# - PM_COMMAND_OPTIONS
# - FAI_DEBSOURCESDIR (for --download-sources)
# - MAXPACKAGES (for --max-packages)
# - aptoptions (used for Debian’s download commands; use PM_COMMAND_OPTIONS instead)
# - (This list is probably incomplete)
class Packaging::Packages
  include Options
  include FaiToolkit::Checks
  include Packaging::PackagesMethods

  def initialize
    # Run without requiring command line arguments.
    @run_by_default = true

    # Order of the commands that are executed.
    @EXECUTION_ORDER = %w[ y2i y2r zypper zypper-pattern zypper-product zypper-rm yast rpmr urpmi urpme yumgroup yumi yumr dnfgroup dnfi dnfr smarti smartr hold taskrm taskinst clean-internal cupt cupt-r install install-norec aptitude aptitude-r unpack remove dselect-upgrade ]

    # apt commands shouldn’t be interactive.
    ENV["DEBIAN_FRONTEND"] = "noninteractive"

    # list of all known packages
    @known_packages = []

    # Maximum length of package list.
    #
    # Kind of means how many packages will be installed with a single install (e.g. zypper) command.
    @max_package_list_size = 99

    # As far as I’ve understood, this is list of packages specified in package_config files.
    @package_list = {}

    # Hash containing the types found in all loaded package_config files.
    #
    # That means, for example contains "zypper" => 1 if zypper is used to install packages per package_config files.
    @install_command_types = {}

    # This is pretty hacky, but the thing this is used is not really worth moving to packages-debian.rb.
    @hasdebian = false

    @preloadlist = []
    @preloadrmlist = []
  end

  def options parser
    options = common_options(parser: parser) do |p|
      p.opt :fai_config_dir,                "FAI’s main config files.",
          type: :string,                              short: "c"
      p.opt :list_packages,                 "List packages that would be installed/downloaded and exit",
                                                      short: "l"
      p.opt :list_packages_with_version,    "List packages with version number that would be installed/downloaded and exit",
                                                      short: "L"
      p.opt :list_install_targets,          "List install commands that can be used in package_config files. List varies per given distribution",
                                                      short: "H"
      p.opt :target_distribution,           "Target distribution. Supported values: debian, suse, opensuse",
          type: :string,                              short: "D"
      p.opt :download_only,                 "Only download packages, do not install them",
                                                      short: "o"
      p.opt :install_command_options,       "These options are passed to the install command executed.",
          type: :string,                              short: "i"
      p.opt :dry_run,                       "Only print commands, do not execute (CAUTION: be aware that this may have bugs)",
                                                      short: "n"
      p.opt :no_package_name_check,         "Do not check if package names in package_config are valid",
                                                      short: "N"
      p.opt :download_sources,              "Also download sources of the packages to be installed. Currently only supported for Debian"
      p.opt :max_packages,                  "Only install this many packages at once.",
          type: :integer,                             short: "m"
    end
  end

  def check_options
    options = @options

    parse_common_options options

    @dry_run = options[:dry_run]
    @download_sources = options[:download_sources]

    if ENV["classes"].nil?
      error "classes environment variable is not set"
    end

    if @fai_config_dir
      @config = FaiConfig.new @fai_config_dir
      @config.load_fai_conf
      @config.load_nfsroot_conf
      @config_space = @config.config_space
      @FAI_ROOT = @config.nfsroot
    end

    unless @fai_config_dir
      if ENV["FAI"].nil?
        error "No FAI config dir found. Neither FAI environment variable is set nor --fai-config-dir is given."
      end
      @config_space = ENV["FAI"]
    end

    unless File.directory? "#{@config_space}/package_config"
      error "Config space #{@config_space} is not valid FAI config space."
    end

    unless @FAI_ROOT
      if ENV["FAI_ROOT"].nil?
        error "No nfsroot found. Neither FAI_ROOT environment variable is set nor --fai-config-dir is given."
      end
      @FAI_ROOT = ENV["FAI_ROOT"]
    end

    @list_packages = options[:list_packages]
    @list_packages_with_version = options[:list_packages_with_version]
    @listonly = @list_packages || @list_packages_with_version

    @only_download = options[:download_only]

    # TODO: Optimist doesn’t parse arguments like "--foo --bar" where --bar is value of --foo correctly.
    @install_command_options = options[:install_command_options]
    @install_command_options = ENV["INSTALL_COMMAND_OPTIONS"] if @install_command_options.nil?

    @pm_command_options = ENV["PM_COMMAND_OPTIONS"]

    check_target_distribution options[:target_distribution] do |distro|
      case distro
      when "debian"
        #raise "apt-get is missing. Are you sure you are on Debian?" unless command_available? "apt-get"
        @impl = Packaging::PackagesDebian.new options
      when "suse"
        #raise "zypper is missing. Are you sure you are on SUSE?" unless command_available? "zypper"
        @impl = Packaging::PackagesSuse.new options
      when "opensuse"
        #raise "zypper is missing. Are you sure you are on OpenSUSE?" unless command_available? "zypper"
        @impl = Packaging::PackagesOpenSuse.new options
      else
        raise "Invalid target distribution: #{distro.inspect}"
      end

      @impl.FAI_ROOT = @FAI_ROOT
      @impl.only_download = @only_download
      @impl.list_packages = @list_packages
      @impl.download_sources = @download_sources
    end

    # Generates variables after we have enough data to generate them correctly.
    generate_variables

    list_install_targets = options[:list_install_targets]
    return showcommands if list_install_targets

    @max_package_list_size = options[:max_packages] if options[:max_packages]
    @max_package_list_size = ENV["MAXPACKAGES"] if ENV["MAXPACKAGES"]
    verb "Maximum number of packages installed at a time set to #{@max_package_list_size}";

    @no_package_name_check = options[:no_package_name_check]
    unless @no_package_name_check
      @no_package_name_check = true if ENV["FAI_DISABLE_PACKAGE_NAME_CHECK"]
    end

    set_download_only if @only_download

    package_files_dir = "#{@config_space}/package_config"

    @rootcmd = (@FAI_ROOT == "/" ) ? '' : "chroot #{@FAI_ROOT}"
    @impl.rootcmd = @rootcmd

    # TODO: how to specify classes for testing?
    classes = ENV["classes"].split(/[\s\n]+/)
    @classes = classes.grep_v /^#|^\s*$/

    @class_is_defined = {}
    @classes.each do |klass|
      @class_is_defined[klass] = 1
    end

    verb "[ft packages] reading config files from directory #{package_files_dir}"
    @classes.each do |klass|
      # read all package config files
      if File.exist? "#{package_files_dir}/#{klass}"
        readconfig(package_files_dir, klass)
      end
    end

    # check if any Debian related commands/types are used in package_config
    debiantypes = %w[ taskinst cupt cupt-r apt aptitude aptitude-r install install-norec remove dselect-upgrade smarti ];
    debiantypes.each do |debian_type|
      @hasdebian = true if @install_command_types[debian_type]
    end

    # smarti is used in a packages_config file
    if @install_command_types['smarti']
      @COMMANDS['clean-internal'] = @COMMANDS['smartc-internal'];
      @COMMANDS['pending-internal'] = "true";
      @COMMANDS['dpkgc-internal'] = "true";
    end

    set_opts options
  end

  def perform
    startup_checks
    real_perform
  end

  private

  def generate_variables
    aptopt = '-y -o Dpkg::Options::=--force-confdef -o Dpkg::Options::=--force-confnew'
    @download_directory = "/var/cache/apt/archives/partial/" # where to download packages that gets only unpacked

    @COMMANDS = {
         "taskinst" => "tasksel install #{@install_command_options}",
           "taskrm" => "tasksel remove #{@install_command_options}",
            "urpmi" => "urpmi --auto --foce --allow-force --keep #{@install_command_options}",
            "urpme" => "urpme --auto --foce --allow-force --keep #{@install_command_options}",
             "yumi" => "yum -y #{@pm_command_options} install #{@install_command_options}",
             "yumr" => "yum -y #{@pm_command_options} remove #{@install_command_options}",
        "yumgroup"  => "yum -y #{@pm_command_options} groupinstall #{@install_command_options}",
             "dnfi" => "dnf -y #{@pm_command_options} install #{@install_command_options}",
             "dnfr" => "dnf -y #{@pm_command_options} remove #{@install_command_options}",
        "dnfgroup"  => "dnf -y #{@pm_command_options} group install #{@install_command_options}",
             "y2i"  => "y2pmsh isc #{@install_command_options}",
             "y2r"  => "y2pmsh remove #{@install_command_options}",
             "rpmr" => "rpm #{@pm_command_options} -e #{@install_command_options}",
           "smarti" => "smart install -y #{@install_command_options}",
           "smartr" => "smart remove -y #{@install_command_options}",
  "smartc-internal" => "smart clean",
    }

    case @target_distribution
    when "debian"
      @COMMANDS.merge!({
          "install" => "apt-get #{aptopt} --fix-missing install #{@install_command_options} #{@install_command_options}",
    "install-norec" => "apt-get #{aptopt} --fix-missing install --no-install-recommends #{@install_command_options}",
           "remove" => "apt-get #{aptopt} --purge remove #{@install_command_options}",
  "dselect-upgrade" => "apt-get #{aptopt} dselect-upgrade #{@install_command_options}",
             "hold" => "dpkg --set-selections #{@install_command_options}",
             "cupt" => "cupt -R #{aptopt} install #{@install_command_options}",
           "cupt-r" => "cupt #{aptopt} install #{@install_command_options}",
   "clean-internal" => "apt-get clean",
         "aptitude" => "aptitude -R #{aptopt} install #{@install_command_options}",
       "aptitude-r" => "aptitude -r #{aptopt} install #{@install_command_options}",
              "apt" => "apt #{aptopt} install #{@install_command_options}",
           "unpack" => "cd #{@download_directory}; aptitude download #{@install_command_options}",
  "unpack-internal" => "dpkg --unpack --recursive #{@download_directory}; rm #{@download_directory}/*.deb",
 "pending-internal" => "dpkg --configure --pending",
   "dpkgc-internal" => "dpkg -C",
      })
    when "opensuse", "suse"
      @COMMANDS.merge!({
             "yast" => "yast #{@pm_command_options} -i #{@install_command_options}",
           "zypper" => "zypper --non-interactive #{@pm_command_options} install -l #{@install_command_options}",
   "zypper-pattern" => "zypper --non-interactive #{@pm_command_options} install -l -t pattern #{@install_command_options}",
   "zypper-product" => "zypper --non-interactive #{@pm_command_options} install -l -t product #{@install_command_options}",
        "zypper-rm" => "zypper --non-interactive #{@pm_command_options} remove #{@install_command_options}",
      })
    end
  end

  def startup_checks
    unless File.directory? @FAI_ROOT
      error "#{@FAI_ROOT} does not exist. Are you using correct nfsroot?"
    end

    # get files which must exist before installing packages
    unless @listonly
      (@preloadlist + @preloadrmlist).each do |entry|
        (url, directory) = entry
        matches = url.match %r{^file:/(.+)}m
        if (matches)
          file = matches[1]
          if @dry_run
            puts "Copying #{@FAI_ROOT}/#{file} to #{@FAI_ROOT}/#{directory}"
          else
            FileUtils.cp "#{@FAI_ROOT}/#{file}", "#{@FAI_ROOT}/#{directory}"
          end
        else
          execute("cd #{@FAI_ROOT}/#{directory}; curl -# -O #{url}")
        end
      end
    end

    @impl.startup_checks
  end

  def real_perform
    debug "Running installers"

    # call apt-get or tasksel for each type of command whith the list of packages
    @EXECUTION_ORDER.each do |command|
      trace "Running #{command}"

      # Trying to install something we don’t have defined (wrong distribution most likely).
      if @COMMANDS[command].nil?
        next
      end

      if command == "clean-internal" and @hasdebian
        execute("#{@rootcmd} #{@COMMANDS['clean-internal']}") unless @listonly
        next
      end

      # skip if empty list
      next unless @package_list[command]

      if command == "dselect-upgrade"
        @impl.dselectupgrade(command, @COMMANDS[command], @package_list)
        next
      end

      packlist = @package_list[command].join " "

      trace "Operating with package list: #{packlist.inspect}"

      if command == "hold"
        hold = @package_list["hold"].join " hold\n"
        execute("echo \"#{hold}\" | #{@rootcmd} #{@COMMANDS["hold"]}")
        next
      end

      execute_here = %w[ install install-norec smarti cupt cupt-r apt aptitude aptitude-r unpack ]
      if execute_here.include? command
        execute_here command
        next
      end

      if command == "taskinst" or command == "taskrm"
        @package_list[command].each do |task|
          execute("#{@rootcmd} #{@COMMANDS[command]} #{task}")
        end
        next
      end

      # other types
      env = { "LC_ALL" => "C" }
      install_packlist = packlist
      if packlist
        if command == "zypper"
          install_packlist = "-- #{install_packlist}"
        end

        execute("#{@rootcmd} #{@COMMANDS[command]} #{install_packlist}", env: env)
      end
    end

    rest_of_stuff
  end

  # TODO: This probably belongs to packages-debian.rb.
  def execute_here command
    known = @impl.create_package_list @package_list[command]
    @known_packages.concat known
    @impl.get_sources known

    return if @list_packages

    if (@list_packages_with_version)
      # only print the package list, only works for install and install-norec
      execute("#{@rootcmd} #{@COMMANDS[command]} -s #{known.join(" ")} | grep -E ^Inst")
      return
    end

    # pass only maxpl packages to apt-get
    while (known.size > 0)
      shortlist = known.slice!(0, @max_package_list_size).join(" ")

      # We can skip a package if it is already processed.
      if (@max_package_list_size == 1 && shortlist)
        return if Dir.glob "#{ENV["FAI_ARCHIVE_DIR"]}/#{shortlist}_*deb"
      end

      execute("#{@rootcmd} #{@COMMANDS[command]} #{shortlist}") if shortlist
      execute("#{@rootcmd} #{@COMMANDS['clean-internal']}") if @hasdebian
      execute("#{@rootcmd} #{@COMMANDS['unpack-internal']}") if (command == "unpack") # unpack and rm deb files
    end
  end

  def rest_of_stuff
    exit 0 if @list_packages_with_version

    if @list_packages
      trace "Following files are to be installed:"
      puts @known_packages.join(" ")
      exit 0
    end

    # remove preloaded files
    @preloadrmlist.each do |entry|
      (url, directory) = entry
      matches = url.match %r{/([^/]+$)}m
      file = "#{directory}/#{matches[1]}";
      verb "Removing #{file}"
      File.rm file
    end

    if @hasdebian
      # in case of unconfigured packages because of apt errors
      # retry configuration
      execute("#{@rootcmd} #{@COMMANDS['pending-internal']}");
      # check if all went right
      execute("#{@rootcmd} #{@COMMANDS['dpkgc-internal']}");
      # clean apt cache
      execute("#{@rootcmd} #{@COMMANDS['clean-internal']}");
    end
  end

  def set_download_only
    # Definitions for install_packages(8) for download only mode
    # Used by fai-mirror(1)

    # we are using a variable from %ENV
    # for debugging remove >/dev/null and -q

    @max_package_list_size = 9999

    @COMMANDS = {
        "clean-internal" => 'true',
      "pending-internal" => 'true',
        "dpkgc-internal" => 'true'
    }

    @COMMANDS.merge!(@impl.download_commands(@pm_command_options, @install_command_options))
  end

  def readconfig path, file
    type = nil
    do_it = false

    f = File.open "#{path}/#{file}"
    verb "[ft packages] reading config file #{path}/#{file}\n"

    f.readlines.each do |line|
      next if line.match? /^#/ # Skip fully comment lines.
      line.sub! /#.*$/, "" # Delete end of line comments.
      next if line.match? /^\s*$/ # Skip empty lines.
      line.chomp!

      matches = line.match /^PRELOAD\s+(\S+)\s+(\S+)/
      if matches
        @preloadlist << [matches[1], matches[2]]
        next
      end
      matches = line.match /^PRELOADRM\s+(\S+)\s+(\S+)/
      if matches
        @preloadrmlist << [matches[1], matches[2]]
        next
      end

      matches = line.match /^PACKAGES\s+(\S+)\s*/
      if matches
        cllist = $'
        type = matches[1]
        do_it = handle_readconfig_match cllist, type, line, do_it
        next
      end

      @impl.readconfig_checks type, file, line

      @package_list[type] ||= []
      if do_it
        @package_list[type].concat(line.split(" ").reject(&:empty?))
        @install_command_types[type] = 1 # remember which types are used in package_config
      end
    end

    f.close
  end

  def handle_readconfig_match cllist, type, line, do_it
    unless @COMMANDS[type]
      if not @unkown_action_msg
        @unkown_action_msg = "Unknown action #{type} after PACKAGES. This probably means that you have packages defined for other distribution than your --target-distribution."
        verb @unkown_action_msg
      end
      trace "Skipping #{type} with line #{line}"

      return do_it
    end

    # by default no classes are listed after this command so doit
    do_it = true
    return do_it if cllist.nil? or cllist.empty?

    # No classes specified after PACKAGES command, so add all packages listed.
    # Use packages on for a list of classes.

    do_it = false # assume no class is defined

    oclasses = cllist.split(/\s+/)

    # If a listed class is defined, add the packages, otherwise skip these packages.
    oclasses.each do |oclass|
      do_it = true if @class_is_defined[oclass]
    end

    do_it
  end
end