
require_relative "packaging"
require_relative "../common"

# TODO: check for sudo available for qemu process.
class Packaging::Qemu
  include Common

  @@qemu_pids = []

  attr_accessor :mac, :tap_device, :bridge_device, :build_machine
  attr_reader :vm_ip

  def initialize opts
    unless command_available? "qemu-system-x86_64"
      raise "qemu for amd64 needs to be installed"
    end

    unless command_available? "qemu-img"
      raise "qemu-img needs to be installed"
    end

    @opts = opts
  end

  def startup_checks
    execute %{qemu-img check #{@build_machine}}
    raise "Validating #{@build_machine} failed." unless $?.success?

    at_exit do
      terminate_qemus
    end
  end

  def start_vm build_machine
    #graphics_params="-display none -nographic"
    graphics_params=""

    network_device = "-device virtio-net-pci,netdev=hostnet0,id=net0,mac=#{@mac}"
    netconfig = "-netdev tap,id=hostnet0,ifname=#{@tap_device},vhost=on,script=no,downscript=no #{network_device}"
    # cd_image: UPDATE_IMAGE_NAME can be used for debugging.
    qemu_pid = start_qemu_vm "Build machine VM", disk_image_name: build_machine, netconfig: netconfig,
        additional_params: "-no-reboot", graphics_params: graphics_params
  end

  def close_hypervisors
    terminate_qemus
  end

  # Wait for SSH connection to be available.
  def wait_for_connection
    # Connection may not yet be up, so this may block for a bit.
    @vm_ip = find_vm_ip @mac

    # Then wait for SSH to be available
    wait_for_ssh

    @vm_ip
  end

  private

  # Waits for approximately maximum of 20 seconds.
  def wait_for_ssh counter: 0
    remote_ssh_port = 22
    begin
      Socket.tcp(@vm_ip, remote_ssh_port, connect_timeout: 3) {}
    rescue Errno::ETIMEDOUT => e
      if counter > 19
        puts.STDERR "Connecting to host #{@vm_ip}:#{remote_ssh_port} failed. Please check if there is a connection problem."
        raise e
      else
        wait_for_ssh counter: (counter + 3)
      end
    rescue Errno::ECONNREFUSED => e
      raise "Failed to connect to #{@vm_ip}:22: Connection refused." if counter > 19
      sleep 1
      wait_for_ssh counter: (counter + 1)
    end

    nil
  end

  # May block for a bit.
  def find_vm_ip mac, counter: 0
    vm_ip = execvalue %{ip neigh | grep -i #{mac} | cut -d" " -f1}
    begin
      IPAddr.new vm_ip
    rescue IPAddr::InvalidAddressError
      raise "Were not able to find out test VM’s IP address." if counter > 10
      sleep 1
      vm_ip = find_vm_ip mac, counter: (counter + 1)
    end
    #debug "Found ip #{vm_ip}"
    vm_ip
  end

  def terminate_qemus
    @@qemu_pids.delete_if do |pid|
      silent_kill pid
      true
    end
  end

  def start_qemu_vm vm_name, disk_image_name:, memory: 1000, netconfig:, cd_boot: false, cd_image: nil, additional_params: "", graphics_params: ""
    graphics_stack = "-display egl-headless -spice addr=127.0.0.1,port=18000,disable-ticketing,seamless-migration=on,playback-compression=off,image-compression=off"
    #monitor_params = "-monitor unix:/tmp/qemu_monitor"
    # mode=readline == “human” control mode, mode=control == QMP (json protocol)
    monitor_params = "-chardev socket,id=mon0,path=/tmp/qemu_monitor,server,nowait -mon chardev=mon0,mode=readline,pretty=on"
    disk_params = "-drive file=#{disk_image_name},driver=qcow2,if=virtio,index=1,aio=native,media=disk,cache=unsafe,cache.direct=on,cache.no-flush=on,discard=unmap,detect-zeroes=unmap"

    cd_boot_params = cd_boot ? "-boot d" : ""
    cd_params = cd_image ? "-cdrom #{cd_image}" : ""
    params = "-smp #{(Etc.nprocessors / 2.0).round} -accel kvm -enable-kvm #{disk_params} #{cd_boot_params} #{cd_params} -m #{memory} #{monitor_params} #{additional_params}"
    command = %[sudo qemu-system-x86_64 -global isa-fdc.driveA= -name "#{vm_name}" #{params} #{netconfig} #{graphics_stack} #{graphics_params}]
    pid = execute command, wait: false

    sleep 0.1
    if not running? pid
      raise "Starting qemu failed."
    end

    @@qemu_pids << pid
    pid
  end

  def send_to_qemu_monitor command
    # TODO: use ruby for this
    #execute "echo #{command} | socat - UNIX-CONNECT:/tmp/qemu_monitor"
    execute "echo #{command} | sudo nc -N -U /tmp/qemu_monitor", fail: false
  end

  def shutdown_qemu qemu_pid
    send_to_qemu_monitor "system_powerdown"

    begin
      Timeout.timeout(10) do
        Process.wait qemu_pid
      end
    rescue Timeout::Error
      raise "Failed to cleanly close qemu with pid #{qemu_pid}"
    end
  end

  # First checks if the process even is running.
  def silent_kill pid
    return unless running? pid
    # Just killing the sudo process sometimes seems to fail, so let’s kill actual qemu process.
    sub_pids = execvalue "pgrep -P #{pid}"
    unless sub_pids.empty?
      execute "sudo kill #{sub_pids}", fail: false
      unless $?.success?
        debug "Killing subpids #{sub_pids} didn’t work... If there is a warning following, something is wrong. If not, the process probably already was killed. Or following code is broken."
        if running? sub_pids.to_i
          sleep 0.2
          puts "[WARNING] Had to kill #{pid} with force"
          execute "sudo kill -9 #{pid}"
        end
      end
    else
      execute "sudo kill #{pid}", fail: false
      unless $?.success?
        sleep 0.2
        puts "[WARNING] Had to kill #{pid} with force"
        execute "sudo kill -9 #{pid}"
      end
    end
  end

  def running? pid
    begin
      Process.getpgid pid
      true
    rescue Errno::ESRCH
      false
    end
  end
end