
require_relative "qemu-module"

require_relative "../options"
require_relative "../fai-config"

# Main class.
class Qemu::Qemu
  include Options

  def options parser
    options = common_options(parser: parser, log_file: "create-nfsroot.log") do |p|
      p.opt :name,          "Name of the virtual machine. There can be a config file {--qemu-config-dir}/configs/{--name}.cfg that is passed to QEMU",
          type: :string,  short: "n"
      p.opt :image,         "Path to a image that will be attached to the VM",
          type: :string,  short: "i"
      p.opt :boot_image,    "Path to a image that QEMU will boot to",
          type: :string,  short: "b"
      p.opt :disk,          "Path to a QEMU disk. It will be created if it does not exist",
          type: :string,  short: "D",     required: true
      p.opt :disk_type,     "Type of the disk. Possible types: host_device, file",
          type: :string,  short: "t",     default: "host_device"
      p.opt :disk_size,     "If a disk needs to be created, this can be used to pass size of the disk",
          type: :string,  short: :none,   default: "10G"
      p.opt :memory,        "Memory for the VM",
          type: :string,                  default: "1500M"
      p.opt :qemu_config_dir,     "Path to a directory containing VM definitions as this script expects",
          type: :string,  short: "Q",     default: "/etc/qemu"
    end
  end

  def check_options
    options = @options

    parse_common_options options

    @vm_name = options[:name]
    @image_path = options[:image]
    @boot_image_path = options[:boot_image]
    @disk_path = options[:disk]
    @disk_type = options[:disk_type]
    @disk_size = options[:disk_size]
    @memory = options[:memory]
    @qemu_directory = options[:qemu_config_dir]

    if @image_path and not File.file? @image_path
      error "--image was given, but #{@image_path} does not exist."
    end

    if @boot_image_path and not File.file? @boot_image_path
      error "--boot-image was given, but #{@boot_image_path} does not exist."
    end

    unless File.directory? @qemu_directory
      error "Directory #{@qemu_directory} does not exist. You have invalid installation, or specified invalid --qemu-config-dir."
    end
    unless File.file? "#{@qemu_directory}/lib/global.cfg"
      error "QEMU config file #{@qemu_directory}/lib/global.cfg does not exist. You have invalid installation, or specified invalid --qemu-config-dir. You can add the file manually if you need to."
    end

    if File.directory? "#{@qemu_directory}/configs"
      @qemu_config_dir = "#{@qemu_directory}/configs"
    end

    # NOTE: When booting to SUSE graphical installer, it seems to need quite a bit of memory. I tested it successfully with 3 GB memory.
    @install_image_args = ""
    if opts[:boot_image]
        @boot_image_args = "-cdrom #{opts[:install_image]} -boot once=d"
    end

    case @disk_type
    when "host_device"
    when "file"
    else
      error "Invalid --host-type. See --help for allowed values."
    end
  end

  def perform
    startup_checks
    create_qemu_image @disk_path unless File.file? @disk_path
    run_qemu
  end

  private

  def startup_checks
    # TODO: This is kind of dummy, but I think create_qemu_image may be useful outside of this class, so that maybe needs to resolve it too. Or make it enforced argument?
    @disk_format = solve_disk_format @disk_path, nil
  end

  def run_qemu
    spiceconfig = []
    if File.file? "#{@qemu_config_dir}/#{@vm_name}-spice.cfg"
      spiceconfig = [ "-readconfig", "#{@qemu_config_dir}/#{@vm_name}-spice.cfg" ]
    end

    vm_config = []
    if File.file? "#{@qemu_config_dir}/#{@vm_name}.cfg"
      vm_config = [ "-readconfig", "#{@qemu_config_dir}/#{@vm_name}.cfg" ]
    end

    case @disk_type
    when "host_device"
      # TODO: ensure that given/parsed format matches raw.
      diskconfig = [ "-blockdev", "node-name=system,driver=raw,file.driver=host_device,file.filename=#{@disk_path},discard=unmap,file.aio=native,file.cache.direct=on" ]
    when "file"
      diskconfig = [ "-blockdev", "node-name=system,driver=raw,file.driver=host_device,file.filename=#{@disk_path},discard=unmap,file.aio=native,file.cache.direct=on" ]
      diskconfig = [ "-blockdev", "node-name=rootdisk,driver=#{@disk_format},file.driver=file,file.filename=#{@disk_path}," +
                                  "file.aio=native,file.cache.direct=on,file.cache.no-flush=on,discard=unmap,detect-zeroes=unmap",
                     "-device", "scsi-hd,id=rootdisk,bus=scsi.0,drive=rootdisk,bootindex=3,channel=0,lun=1,scsi-id=1" ]
    end

    command = [ which("qemu-system-x86_64"),
                "-nodefaults",
                "-readconfig", "#{@qemu_directory}/lib/global.cfg",
                vm_config,
                spiceconfig,
                diskconfig,
                @boot_image_args,
                "-display", "egl-headless",
                "-m", @memory
    ]
    command.flatten!

    execute command
  end

  def create_qemu_image image_name, format: nil
    #FileUtils.rm image_name if File.exist? image_name

    format = solve_disk_format image_name, format

    cluster_size = format == "qcow2" && ",cluster_size=512" || ""

    execute [ "qemu-img", "create",
              "-f", format,
              "-o", "nocow=on#{cluster_size}",
              image_name,
              @disk_size ]
  end

  def solve_disk_format image_name, format
    return format unless format.nil?

    splitted = image_name.split "."
    case splitted[1]
    when "qcow2"
      format = "qcow2"
    when "raw"
      format = "raw"
    else
      raise "Not implemented disk image type."
    end

    format
  end

end
